!> @file RA15Mod.f90
!! Radau15 integrator

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Module  RA15Mod
!> Module for integration using radau method
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com from a code of V. Lainey (IMCCE)
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
module ra15mod
   use kindtype
   implicit none
 
public :: setupra15,ra15m
integer(id), private, save       :: nclass !< 1= fist order ODE y'=f(y,t) ; 2=2nd order ODE y"=f(y,y',t) ; -2=2nd order equation y"=f(y,t)
integer(id), private, save       :: neq !< number of equations
real(rd), private, save          :: ss !< if >0 : tolerance if <0: use of a fixed step
integer(id), private, save       :: neqcontrol !< if step size control, error is controlled on neqcontrol first equations

contains

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine setupra15
!> Subroutine to setup the radau integrator properties
!!
!! @param[in] ne number of equations to be integrated
!! @param[in] nc integer- 1=1st order equations (y'=f(y,t)) - 2=2nd order equation y''=f(y,y',t)   -2=2nd order equation y''=f(y,t)
!! @param[in] tol integer tolerance of integration given in 10**-tol, if tol<0, use of a fixed step size
!! @param[in] neqc optional integer - control of step-size is done on the neqc first equations
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine setupra15(ne,nc,tol,neqc)
   use kindtype
   implicit none
   integer(id), intent(in)             :: ne
   integer(id), intent(in)             :: nc
   real(rd), intent(in)             :: tol
   integer(id), intent(in), optional   :: neqc
   neq=ne
   nclass=nc
   ss=tol
   neqcontrol=neq
   if(present(neqc)) neqcontrol=neqc
  ! write(*,*)ne,nc,neqcontrol

end subroutine setupra15

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine ra15m
!> Subroutine to perform integration
!!
!! @param[in] f external function to be integrated (y'=f or y"=f)
!! @param[in,out] x real array of dimension neq - in: initial conditions - out: positionf after integration
!! @param[in,out] v optional real array of dimension neq - initial velocities if nclass has been steup to 2 or -2 out: velocities after integration
!! @param[in] td real starting time
!! @param[in] tf0 real end time
!! @param[in] xl real representing the step size wanted
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!!
!!  Integrator RADAU by E. Everhart, Physics Department, University of Denver 
!!  This 15th-order version, called RA15, is written out for faster execution. 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
SUBROUTINE RA15M(f,X,V,TD,TF0,XL) 

  USE kindtype


  IMPLICIT NONE

  external   f
  REAL(wp), DIMENSION(neq), INTENT(INout):: X,V
  REAL(wp), INTENT(IN):: TD, TF0, XL


  REAL(wp):: XLrd, SR, DIR, TF, PW, WW, W1,  TP, TM, Trd, T2, TVAL, S, TEMP, HV, GKrd, VIP
  REAL(wp), DIMENSION(size(X)):: F1, FJ, Y, Z
  REAL(wp), DIMENSION(21):: Crd, D, R
  REAL(wp), DIMENSION(7,NEQ):: Brd, G, E, BD
  REAL(wp), DIMENSION(8):: H
  REAL(wp), DIMENSION(7):: W, U, Qrd

  INTEGER(iwp):: J, K, L, M, N, LA, LB, LC, LD, LE, NF, NI, JD, JDM, NCOUNT, NS
  INTEGER(iwp), DIMENSION(8):: NW

  LOGICAL:: NPQ, NSF, NPER, NCL, NES 
  

  IF(ABS(TF0-TD)>1.E-15_wp)THEN

     Y=X 
     Z=V 

     XLrd=XL


     NW=(/ 0,0,1,3,6,10,15,21 /) 
     SR=1.4E0_wp 
     !  These H values are the Gauss-Radau spacings, scaled to the range 0 to 1, for integrating to order 15. 
     H=(/ 0.E0_wp, 0.05626256053692215E0_wp, 0.18024069173689236E0_wp, 0.35262471711316964E0_wp, 0.54715362633055538E0_wp, &
          0.73421017721541053E0_wp, 0.88532094683909577E0_wp, 0.97752061356128750E0_wp /) 
     !  The sum of the H-values should be 3.73333333333333333 
     NPER=.FALSE. 
     NSF=.FALSE. 
     NCL=(NCLASS==1) 
     NPQ=(NCLASS<2) 
     !  y'=F(y,t)  NCL=.TRUE.    y"=F(y,t)  NCL=.FALSE.   y"=F(y',y,t) NCL=.FALSE. 
     !  NCLASS=1   NPQ=.TRUE.    NCLASS= -2 NPQ=.TRUE.    NCLASS= 2    NPQ=.FALSE. 
     !  NSF is .FALSE. on starting sequence, otherwise .TRUE. 
     !  NPER is .TRUE. only on last sequence of the integration. 
     !  NES is .TRUE. only if LL is negative. Then the sequence size is XLrd. 
     DIR=1.E0_wp 
     TF=TF0-TD                !!! NOUVELLE INSTRUCTION  !!!!!!!!!!!!!!!!!!!!!! 
     IF(TF<0.E0_wp) DIR=-1.E0_wp   
     NES=ss<0 
     XLrd=DIR*ABS(XLrd) 
     PW=1.E0_wp/9.E0_wp 
     !  Evaluate the constants in the W-, U-, C-, D-, and R-vectors 
     DO N=2,8 
        WW=N+N*N 
        IF(NCL) WW=N 
        W(N-1)=1.E0_wp/WW 
        WW=N 
        U(N-1)=1.E0_wp/WW 
     ENDDO
     IF(NCL) V(1:NEQ)=0.E0_wp  
     BD=0.E0_wp 
     Brd=0.E0_wp 
     W1=0.5E0_wp 
     IF(NCL) W1=1.E0_wp 
     Crd(1)=-H(2) 
     D(1)=H(2) 
     R(1)=1.E0_wp/(H(3)-H(2)) 
     LA=1 
     LC=1 
     DO K=3,7 
        LB=LA 
        LA=LC+1 
        LC=NW(K+1) 
        Crd(LA)=-H(K)*Crd(LB) 
        Crd(LC)=Crd(LA-1)-H(K) 
        D(LA)=H(2)*D(LB) 
        D(LC)=-Crd(LC) 
        R(LA)=1.E0_wp/(H(K+1)-H(2)) 
        R(LC)=1.E0_wp/(H(K+1)-H(K)) 
        IF(K/=3)THEN 
           DO L=4,K 
              LD=LA+L-3 
              LE=LB+L-4 
              Crd(LD)=Crd(LE)-H(K)*Crd(LE+1) 
              D(LD)=D(LE)+H(L-1)*D(LE+1) 
              R(LD)=1.E0_wp/(H(K+1)-H(L-1)) 
           ENDDO
        ENDIF
     ENDDO
     !  The statements above are used only once in an integration to set up the 
     !  constants. They use less than a second of execution time.  Next set in 
     !  a reasonable estimate to TP based on experience. Same sign as DIR. 
     !  An initial first sequence size can be set with XLrd even with LL positive. 
     TP=0.1E0_wp*DIR 
     IF(XLrd.NE.0.E0_wp) TP=XLrd 
     IF(NES) TP=XLrd 
     IF(TP/TF>0.5E0_wp) TP=0.5E0_wp*TF 
     NCOUNT=0 

     !  An * is the symbol for writing on the monitor. The printer is unit 4. 
     !  Line 4000 is the starting place of the first sequence. 
4000 NS=0 
     NF=0 
     NI=6 
     TM=TD     !0.E0_wp               !!! CORRECTION  *****************************

     if (NCL) then   
!   write(*,*)'ra15 - 1'
        call f(TM,X,F1)
!     else
        !CALL f (TM,X, V,  F1) 
      endif
     NF=NF+1 
     ! Line 722 is begins every sequence after the first. First find new beta- 
     !  values from the predicted B-values, following Eq. (2.7) in text. 
722  CONTINUE 
     G(1,:)=Brd(1,:)+D(1)*Brd(2,:)+D(2)*Brd(3,:)+D(4)*Brd(4,:)+D( 7)*Brd(5,:)+D(11)*Brd(6,:)+D(16)*Brd(7,:) 
     G(2,:)=              Brd(2,:)+D(3)*Brd(3,:)+D(5)*Brd(4,:)+D( 8)*Brd(5,:)+D(12)*Brd(6,:)+D(17)*Brd(7,:) 
     G(3,:)=                            Brd(3,:)+D(6)*Brd(4,:)+D( 9)*Brd(5,:)+D(13)*Brd(6,:)+D(18)*Brd(7,:) 
     G(4,:)=                                          Brd(4,:)+D(10)*Brd(5,:)+D(14)*Brd(6,:)+D(19)*Brd(7,:) 
     G(5,:)=                                                         Brd(5,:)+D(15)*Brd(6,:)+D(20)*Brd(7,:) 
     G(6,:)=                                                                        Brd(6,:)+D(21)*Brd(7,:) 
     G(7,:)=                                                                                       Brd(7,:) 

     Trd=TP 
     T2=Trd*Trd 
     IF(NCL) T2=Trd 
     TVAL=ABS(Trd) 

     !  Loop 175 is 6 iterations on first sequence and two iterations therafter. 
     DO M=1,NI 
        !  Loop 174 is for each substep within a sequence. 
        DO J=2,8 
           JD=J-1 
           JDM=J-2 
           S=H(J) 
           Qrd(1)=S 
           IF(NCL) Qrd(1)=1.E0_wp 
           !  Use Eqs. (2.9) and (2.10) of text to predict positions at each aubstep. 
           DO K=1,NEQ
             
                 Y(K)=X(K)+Qrd(1)*(Trd*V(K)+T2*S*(F1(K)*W1+S*(W(1)*Brd(1,K)+S*(W(2)*Brd(2,K)+S*(W(3)*Brd(3,K)  &
                      +S*(W(4)*Brd(4,K)+S*(W(5)*Brd(5,K)+S*(W(6)*Brd(6,K)+S*W(7)*Brd(7,K))))))))) 
                 IF(.NOT.(NPQ))THEN
                    !  Next are calculated the velocity predictors need for general class II. 
                    Z(K)=V(K)+S*Trd*(F1(K)+S*(U(1)*Brd(1,K)+S*(U(2)*Brd(2,K)+S*(U(3)*Brd(3,K)+S*(U(4)*Brd(4,K)    &
                         +S*(U(5)*Brd(5,K)+S*(U(6)*Brd(6,K)+S*U(7)*Brd(7,K)))))))) 
                 ENDIF
            
           ENDDO !K
           !  Find s at each substep. 
           if (NCL) then
!		write(*,*)'ra15 - 2'
             CALL F(TM+S*Trd,Y,FJ)
!           else 
             ! CALL F(TM+S*Trd,Y,Z,FJ) 
           endif


           NF=NF+1 
           DO K=1,NEQ
            
                 !  Find G-value for the force FJ found at the current substep. This 
                 !  section, including the many-branched GOTO, uses Eq. (2.4) of text. 
                 TEMP=G(JD,K) 
                 GKrd=(FJ(K)-F1(K))/S 
                 GO TO (102,102,103,104,105,106,107,108),J 
102              G(1,K)=GKrd 
                 GO TO 160 
103              G(2,K)=(GKrd-G(1,K))*R(1) 
                 GO TO 160 
104              G(3,K)=((GKrd-G(1,K))*R(2)-G(2,K))*R(3) 
                 GO TO 160 
105              G(4,K)=(((GKrd-G(1,K))*R(4)-G(2,K))*R(5)-G(3,K))*R(6) 
                 GO TO 160 
106              G(5,K)=((((GKrd-G(1,K))*R(7)-G(2,K))*R(8)-G(3,K))*R(9)-G(4,K))*R(10) 
                 GO TO 160 
107              G(6,K)=(((((GKrd-G(1,K))*R(11)-G(2,K))*R(12)-G(3,K))*R(13)-G(4,K))*R(14)-G(5,K))*R(15) 
                 GO TO 160 
108              G(7,K)=((((((GKrd-G(1,K))*R(16)-G(2,K))*R(17)-G(3,K))*R(18)-G(4,K))*R(19)-G(5,K))*R(20)-G(6,K))*R(21) 
                 !  Upgrade all B-values. 
160              TEMP=G(JD,K)-TEMP 
                 Brd(JD,K)=Brd(JD,K)+TEMP 
                 !  TEMP is now the improvement on G(JD,K) over its former value. 
                 !  Now we upgrade the B-value using this dfference in the one term. 
                 !  This section is based on Eq. (2.5). 
                 GO TO (171,171,203,204,205,206,207,208),J 
203              Brd(1,K)=Brd(1,K)+Crd(1)*TEMP 
                 GO TO 171 
204              Brd(1:2,K)=Brd(1:2,K)+Crd(2:3)*TEMP 
                 GO TO 171 
205              Brd(1:3,K)=Brd(1:3,K)+Crd(4:6)*TEMP 
                 GO TO 171 
206              Brd(1:4,K)=Brd(1:4,K)+Crd(7:10)*TEMP 
                 GO TO 171 
207              Brd(1:5,K)=Brd(1:5,K)+Crd(11:15)*TEMP 
                 GO TO 171 
208              Brd(1:6,K)=Brd(1:6,K)+Crd(16:21)*TEMP 

171           continue
          ENDDO !K
174     ENDDO !J

        IF(.NOT.(NES.OR.M<NI))THEN 
           !  Integration of sequence is over. Next is sequence size control. 
           HV=MAXVAL(ABS(Brd(7,1:neqcontrol))) 
   !        write(*,*) hv,neqcontrol
 !          write(*,*) hv,'ici'
          ! write(*,*)'neq',neq
 !         do k=1,neq
 !             if (ABS(Brd(7,k)).gt.1.0_rd)              write(*,*)'   ',k,ABS(Brd(7,k))
!write(*,*)k,ABS(Brd(7,k))
!enddo
      
           HV=HV*W(7)/TVAL**7 
        ENDIF
     ENDDO !M
     IF (.NOT.(NSF))THEN 
        IF(.NOT.NES) then
           TP=(SS/HV)**PW*DIR 
!write(*,*)tp,hv
 !          write(*,*) ss,hv,pw,dir,tp
        endif
        IF(NES) TP=XLrd 
        IF(.NOT.(NES).AND.(.NOT.(TP/Trd>1.E0_wp)))THEN
           TP=0.8E0_wp*TP 
           NCOUNT=NCOUNT+1 
           IF(NCOUNT>10) RETURN 
           !  Restart with 0.8x sequence size if new size called for is smaller than 
           !  originally chosen starting sequence size on first sequence. 
           GO TO 4000 
        ENDIF
        NSF=.TRUE. 
     ENDIF
     ! Loop 35 finds new X and V values at end of sequence using Eqs. (2.11),(2.12) 
     DO K=1,NEQ
       
 X(K)=X(K)+V(K)*Trd+T2*(F1(K)*W1+Brd(1,K)*W(1)+Brd(2,K)*W(2)+Brd(3,K)*W(3)+Brd(4,K)*W(4)+Brd(5,K)*W(5)+Brd(6,K)*W(6)+Brd(7,K)*W(7)) 
           IF(.NOT.(NCL)) then
 V(K)=V(K)+Trd*(F1(K)+Brd(1,K)*U(1)+Brd(2,K)*U(2)+Brd(3,K)*U(3)+Brd(4,K)*U(4)+Brd(5,K)*U(5)+Brd(6,K)*U(6)+Brd(7,K)*U(7)) 
        endif
      
     ENDDO !K
     NS=NS+1 
     IF(NES)THEN 
        TM=TD+NS*Trd   
     ELSE
        TM=TM+Trd 
     ENDIF
     !  Return if done. 
     IF(NPER)THEN
!        CALL SORTIE(X,V,TD+NS*Trd,VIP) 
!        WRITE(16,*) TM,VIP 
        RETURN 
     ENDIF
     ! Control on size of next sequence and adjust last sequence to exactly cover the integration span. NPER=.TRUE. set on last sequence. 
     if (NCL) then
        call f(TM,X,F1)  

!     else
        !CALL f (TM,X,V,F1) 
     endif
     NF=NF+1 
     IF(.NOT.(NES))THEN
        TP=DIR*(SS/HV)**PW 
        IF(TP/Trd>SR) TP=Trd*SR
     ENDIF
     IF(NES) TP=XLrd 
     IF(.NOT.(DIR*(TM+TP)<DIR*(TF0)-1.E-08_wp))THEN  
        TP=TF0-TM 
        NPER=.TRUE. 
     ENDIF
     !  Now predict B-values for next step. The predicted values from the preceding sequence were saved in the E-matrix. 
     !  The correction BD between the actual B-values found and these predicted values is applied in advance to the 
     !  next sequence. The gain in accuracy is significant. Using Eqs. (2.13): 
     Qrd(1)=TP/Trd 
     DO K=2,7
        Qrd(K)=Qrd(K-1)*Qrd(1)
     ENDDO !K

     IF(NS/=1) BD=Brd-E

  E(1,:)=Qrd(1)*(Brd(1,:)+2.E0_wp*Brd(2,:)+3.E0_wp*Brd(3,:)+4.E0_wp*Brd(4,:)+ 5.E0_wp*Brd(5,:)+ 6.E0_wp*Brd(6,:)+ 7.E0_wp*Brd(7,:)) 
  E(2,:)=                 Qrd(2)*(Brd(2,:)+3.E0_wp*Brd(3,:)+6.E0_wp*Brd(4,:)+10.E0_wp*Brd(5,:)+15.E0_wp*Brd(6,:)+21.E0_wp*Brd(7,:)) 
  E(3,:)=                                  Qrd(3)*(Brd(3,:)+4.E0_wp*Brd(4,:)+10.E0_wp*Brd(5,:)+20.E0_wp*Brd(6,:)+35.E0_wp*Brd(7,:)) 
  E(4,:)=                                                   Qrd(4)*(Brd(4,:)+ 5.E0_wp*Brd(5,:)+15.E0_wp*Brd(6,:)+35.E0_wp*Brd(7,:)) 
  E(5,:)=                                                                     Qrd(5)*(Brd(5,:)+ 6.E0_wp*Brd(6,:)+21.E0_wp*Brd(7,:)) 
  E(6,:)=                                                                                       Qrd(6)*(Brd(6,:)+ 7.E0_wp*Brd(7,:)) 
  E(7,:)=                                                                                                          Qrd(7)*Brd(7,:) 

     Brd=E+BD

     !  Two iterations for every sequence after the first. 
     NI=2 
     IF(.TRUE.) GO TO 722 

  ENDIF !diff t > 1e-15

END SUBROUTINE RA15M

end module ra15mod
