!> @file constantsMod.f90
!! Definition of all the constants used in the simulations (some of them may be loaded from INPOP)

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Module  constantsMod
!> Module where the important constants used in the whole simulations are defined
!>
!> Compilation requirred preprocessor variables INPOP defined. If INPOP=1, the compilation is done with a link to 
!> INPOP planetary ephemerides. This case is requirred for including perturbations from planets. In this case,
!> the compilation needs to be done with a link to CALCEPH library. Moreover, the preprocessor variable INPOPFILE needs to
!> be defined and needs to point towards an INPOP inut file.
!!
!! @author Aurelien Hees, Rhodes University, aurelien.hees@gmail.com
!! @date 01/2015
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
module constantsMod
  use kindtype
  implicit none

  public :: norm, vect, countLines, fromMJD2year, solvekepler



!@ variables public
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	type, public :: prior ! Gaussian prior
		integer(id), public :: idx
		real(rd), public    :: sig
	end type

!@@@@@@@@@ CONSTANTS @@@@@@@@@@@@
  real(rd), parameter, public                        :: GMsun=1.3271244004193942e11_rd !< km^3/s^2
  real(rd), public, parameter                        :: day=24.0_rd*3600.0_rd !< seconds in a day
  real(rd), public, parameter                        :: yr2day=365.25_rd !< seconds in a year
  real(rd), public, parameter                        :: yr=yr2day*day !< seconds in a year
  real(rd), public, parameter                        :: PI=3.1415926535897932384626433832_rd !< pi...
  real(rd), public, parameter                        :: AU=1.495978707E8_rd !< astronomical unit in km
  real(rd), public, parameter                        :: AUyr2kms=AU/yr
  real(rd), public, parameter                        :: GMs_yr2_au3=GMsun*yr**2/AU**3 ! no unit
!  real(rd), parameter, public                        :: as=pi/6.48e5_rd     !< arcsecond in rad
  real(rd), parameter, public                        :: mas=pi/6.48e8_rd     !< milliarcsecond in rad

  real(rd), public, parameter                        :: pc=648000._rd/pi*AU !< in km
  real(rd), public, parameter                        :: cmks=299792.458_rd    !< speed of light in km/s
  real(rd), public, parameter                        :: c=cmks/AUyr2kms !< speed of light in AU/yr
  real(rd), public, parameter                        :: c1=1._rd/c !< 1/c 
  real(rd), public, parameter                        :: c2=c1*c1 !< 1/c2
  real(rd), public, parameter                        :: c4=c2*c2 !< 1/c4			

	integer(id), public, save                             :: npriors
	type(prior), dimension(:), allocatable, public, save  :: priors
	
	integer(id), public, save                             :: nparamLoc !< number of local parameters (to be adjusted)
  integer(id), public, save                             :: nparamGlob!< number of global parameters integrated in the variational equations
  logical, dimension(:), public, save, allocatable      :: toFitGlobal
  integer(id), public, save                             :: nglobtofit=0
  integer(id), public, dimension(:), allocatable        :: idx_glob !< gives the list of index in the order to which parameters are fitted
  integer(id), public, save                             :: nparam !< number of parameters (to be adjusted) =#nglobdyn	+ #nparamloc #nstars

	logical, public, save                                 :: sigNext=.false.
	real(rd), public, save                                :: timeNext
  
  character(len=50),public,save                              :: dirout !< directory for output files	
	character(len=50),public,save                              :: name !< name of the simulation

		
!  character(len=4), public, parameter                        :: extOrb='.orb' !< file extension for the orbitography
	character(len=11), public, parameter                       :: extAng='_sim.points' !< file extension for the signal (angles or radar)
	character(len=11), public, parameter                       :: extrv='_sim.rv' !< file extension for the signal (angles or radar)
!  character(len=4), public, parameter                        :: extpder='.opd' !< file extension for the partials (orbito)
 ! character(len=4), public, parameter                        :: extapder='.spd' !< file extension for the partials (signal)
 ! character(len=4), public, parameter                        :: extpar='.par' !< file extension for the parameters of the simulation
 ! character(len=4), public, parameter                        :: extfit='.fit' !< file extension for the output of the fit
	character(len=4), public, parameter                        :: extres='.res' !< file extension for the residuals

real(rd), private, parameter            :: tolN=1.e-12_rd
integer, parameter, private             :: nitermax=20

  


contains




!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Function norm
!
!> Function used to compute the norm of a 3-vector
!> @param[in] x 3-vector
!> @return the value of the norm of x
!>
!! @author Aurelien Hees, Rhodes University, aurelien.hees@gmail.com
!! @date 01/2015
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
real(rd) function norm(x)
	real(rd), dimension(3), intent(in) :: x
	norm = sqrt(x(1)**2+x(2)**2+x(3)**2)
end function norm

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Function vect
!
!> Function used to compute the vectorial product of 2 vectors
!> @param[in] a 3-vector
!> @param[in] b 3-vector
!> @return a 3 vector which is the cross product of a and b (a x b)
!>
!! @author Aurelien Hees, Rhodes University, aurelien.hees@gmail.com
!! @date 01/2015
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
function vect(a,b)
	real(rd), dimension(3), intent(in) :: a,b
	real(rd), dimension(3) :: vect
	vect(1)=a(2)*b(3)-a(3)*b(2)
	vect(2)=a(3)*b(1)-a(1)*b(3)
	vect(3)=a(1)*b(2)-a(2)*b(1)
end function vect



!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Function countlines
!> Function used to count the number of lines of a file
!> @param[in] name name of the file
!> @return the number of line of the file name
!!
!! @author Aurelien Hees, Rhodes University, aurelien.hees@gmail.com
!! @date 01/2015
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
integer(id) function countLines(name)
  CHARACTER(LEN=*), INTENT(IN) :: name
  chaRACTER(LEN=1) :: TMP
	INTEGER :: ok
	logical::exist
   
  countLines=0
	inquire(file=trim(name), EXIST=exist) 
	if (.NOT.exist) then
    	WRITE(*,*) " XXXX - ERROR : File ", trim(name)," does not exist in countlines (constantsMod.f90)"
    	STOP
	end if
	
  open ( unit =12 ,form='formatted' , file=trim(name) ,iostat=ok)
	if(ok.ne.0) then
    	WRITE(*,*) " XXXX - ERROR : File ", trim(name)," does not exist in countlines (constantsMod.f90)"
		stop
	endif
  do while(ok == 0) !boucle pour compter le nombre de ligne du fichier
   	read(12,*,iostat = ok) 
   	if (ok == 0) countLines = countLines+ 1
  end do
   close(unit=12)
!   WRITE(*,'(I8,A,A)') countlines, ' lines in file ', name

   return
end function countLines

subroutine fromMJD2year(tab)
	real(rd), dimension(:), intent(inout) :: tab
	tab=2000.0_rd + (tab - 51544.5_rd)/yr2day
end subroutine fromMJD2year

real(rd) function fromJY2MJD(yr)
	real(rd), intent(in) :: yr
	fromJY2MJD =(yr -2000.0_rd)*yr2day  +  51544.5_rd
	return 
end function fromJY2MJD

subroutine invMat_addI(in,out)
	real(rd), intent(in), dimension(2,2) :: in
	real(rd), intent(out), dimension(2,2) :: out
	real(rd) :: det
	det=(in(1,1)+1._rd)*(in(2,2)+1._rd)-in(1,2)*in(2,1)
	out(1,1)=(in(2,2)+1._rd)/det
	out(1,2)=-in(1,2)/det
	out(2,1)=-in(2,1)/det
	out(2,2)=(in(1,1)+1._rd)/det
end subroutine invMat_addI


subroutine solveKepler(m,ec,E)
	real(rd), intent(in)   ::  m
	real(rd), intent(in)   ::  ec
	real(rd), intent(out)  ::  E
	
	integer :: i
	real(rd) ::err

	E=m
	err=E-ec*sin(E)-m
	i=0
	do while (abs(err)>tolN) 
		i=i+1
		E=E-err/(1-ec*cos(E))
			err=E-ec*sin(E)-m
		if (i==Nitermax) then
			write(*,*) " WARNING -- Kepler solution not converged for m= ",m, " and e= ",ec, "  E=",E
			exit 
		end if
	end do !
end subroutine solveKepler


end module constantsMod
