!> @file fitMod.f90
!> Module doing a least-square fit 

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Module fitMod
!> Module doing the computation of a least-square fits by different methods: inversion of the normal matrix, SVD decomposition
!!
!> @warning compilation options SVDFIT and LAPACK have to be specified SVDFIT=0 -> standard inversion of the normal matrix ; SVDFIT=1: Standard Value Decomposition
!! inversion of the normal matrix. If SVDFIT=1: use of LAPACK95 module. LAPACK variables has then to be specified in the compilation option. LAPACK=1: use of the MKL library (available with INTEL compiler);
!! if LAPACK !=1 use standard LAPACK95 library
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 06/2010
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
module fitMod
	use kindtype
  implicit none


  public :: fitSVD, computeSig,inverse_SVD
  private :: fitsvdall,covsrt ,swap
#if LAPACK==0
	private :: svdcmp, pythag, gelss_nolapack
#endif
contains
 
subroutine computeSig(A,sigma,maskpar,sig,idx) 
  use kindtype

#if LAPACK==1
	  USE mkl95_LAPACK, ONLY: GELSS !MKL
#endif
  implicit none
  real(rd), dimension(:,:), intent(in)           :: A
  real(rd), dimension(:), intent(in)             :: sigma
  logical, dimension(:), intent(in)              :: maskpar
  real(rd), intent(out)                          :: sig
	integer, intent(in) :: idx

 	integer(id) :: m,n,i,nfit,info,k,lwork, idxR
  real(rd), dimension(:,:), allocatable :: atemp,V,U
	real(rd), dimension(:), allocatable ::sv
	real(8), dimension(:), allocatable :: work

	idxR=0
	if (.not.maskpar(idx)) then
			write(*,*) ' ERROR in computeSig. Parmeters requested not fitted. Idx is ', idx
			stop
	end if
	do i=1,idx
		if(maskpar(i)) idxR=idxR+1
	end do
	
  m=size(A(:,1))
  n=size(A(1,:))
  nfit=count(maskpar)

  if (size(sigma).ne.m) stop "Error in fitSVD size of vector B does not match with vector sigma"
  if (size(maskpar).ne.n) stop "Error in fitSVD size of vector maskpar does not match with matrix A"

  allocate(atemp(m,nfit),sv(nfit),V(nfit,nfit),U(m,nfit))
  k=0
  do i=1,n
     if (maskpar(i))     then
        k=k+1
        atemp(:,k)=a(:,i)
     endif ! maskpar i
  end do ! i

	do i=1,m
   atemp(i,:)=atemp(i,:)/sigma(i)
	enddo !i

#if LAPACK==1   
	  write(*,*) "TO BE IMPLEMENTED"
		stop
		
#elif LAPACK==2
		LWORK = 		MAX(1,3*MIN(M,N) + MAX(M,N),5*MIN(M,N))
		ALLOCATE( WORK(LWORK))


		call dgesvd('N','A',m,nfit,atemp,m,sv,U,1,V,nfit,work,lwork,info)
		deallocate(work)
		if (INFO.ne.0) write(*,*) 'WARNING: problem in SVD decomposition - INFO value ',INFO
		sig=0._rd
		do i=1,nfit
			sig=sig+(V(i,idxR)/sv(i))**2
		end do!i

#else
		call svdcmp(atemp,sv,v)
	sig=0._rd
	do i=1,nfit
		sig=sig+(V(idxR,i)/sv(i))**2
		
	end do!i
#endif

	  

  deallocate(atemp,sv,V,U)
end subroutine computeSig
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine fitSVD 
!> Subroutine that makes a least-squares fit of ||Ax-B|| using a SVD decomposition 
!!                 
!> @param[in] A: matrix of the partial derivative
!> @param[in] B: vector (observations - simulateions)
!> @param[in] sigma : standard deviation of the observations
!> @param[inout] par: in: nominal value for parameters - return the estimated values for parameters 
!> @param[in] maskpar: logical array to specify which parameters has to be estimated
!> @param[out] covar: variance/covariance matrix
!> @param[out] chisq: estimated chi square
!> @param[out] res - optional: estimated residuals
!!
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 11/2011 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
SUBROUTINE fitSVD(A,B,sigma,par,maskpar,covar,chisq,res) 
  use kindtype
#if LAPACK==1
	  USE mkl95_LAPACK, ONLY: GELSS !MKL
#endif

  implicit none
  real(rd), dimension(:,:), intent(in)           :: A
  real(rd), dimension(:), intent(in)             :: B
  real(rd), dimension(:), intent(in)             :: sigma
  real(rd), dimension(:), intent(inout)            :: par
  logical, dimension(:), intent(in)         :: maskpar
  real(rd), intent(out)                          :: chisq
  real(rd),  dimension(:,:),intent(out) :: covar
  real(rd), optional, dimension(:),intent(out)   :: res

  integer(id) :: m,n,i,nfit,k
  real(rd), dimension(:,:), allocatable :: atemp
  real(rd), dimension(:), allocatable :: partemp,btemp
  m=size(A(:,1))
  n=size(A(1,:))
  nfit=count(maskpar)

  if (size(b).ne.m) stop "Error in fitSVD size of vector B does not match with matrix A"
  if (size(sigma).ne.m) stop "Error in fitSVD size of vector B does not match with vector sigma"
  if (size(par).ne.n) stop "Error in fitSVD size of vector par does not match with matrix A"
  if (size(maskpar).ne.n) stop "Error in fitSVD size of vector maskpar does not match with matrix A"
  if (present(res).and.(size(res).ne.m)) stop "Error in fitSVD size of vector res does not match with matrix A"
  if ((size(covar(:,1)).ne.n).and.(size(covar(1,:)).ne.n)) &
       &stop "Error in fitSVD size of matrix corr does not match with matrix A"

  allocate(atemp(m,nfit),partemp(nfit),btemp(m))

  k=0
  do i=1,n
     if (maskpar(i))     then
        k=k+1
        atemp(:,k)=a(:,i)
     endif ! maskpar i
  end do ! i
  do i=1,m
     btemp(i)=b(i)
     if (nfit<n)  btemp(i)=btemp(i)-sum(par(1:n)*a(i,:), mask=.not. maskpar)
  enddo
  call fitsvdall(atemp,btemp,sigma,partemp,chisq,covar(1:nfit,1:nfit),res)

  par(:)=unpack(partemp,maskpar,par(:))
  call covsrt(covar,maskpar)


  deallocate(atemp,partemp,btemp)
end subroutine fitSVD

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine fitSVDAll 
!> Subroutine that makes a least-squares fit of ||Ax-B|| using a SVD decomposition 
!!                 
!> @param[in] A: matrix of the partial derivative
!> @param[in] B: vector (observations - simulateions)
!> @param[in] sigma : standard deviation of the observations
!> @param[out] par:  estimated values for parameters 
!> @param[out] chisq: estimated chi square
!> @param[out] corr: variance/covariance matrix
!> @param[out] res - optional: estimated residuals
!!
!> @warning method using LAPACK95 (gelss method)
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 11/2011 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine fitSVDall(A,B,sigma,par,chisq,corr,res)

#if LAPACK==1
  USE mkl95_LAPACK, ONLY: GELSS !MKL
#endif

  use kindtype
  implicit none
  real(rd), dimension(:,:), intent(in)        :: A
  real(rd), dimension(:), intent(in)          :: B
  real(rd), dimension(:), intent(in)          :: sigma
  real(rd), dimension(:), intent(out)         :: par
  real(rd), intent(out)                       :: chisq
  real(rd), optional, dimension(:,:),intent(out):: corr
  real(rd), optional, dimension(:),intent(out):: res


	integer(id) ::lwork
  integer(id) :: m,n,info,rank,i,k
  real(8), dimension(:), allocatable :: s
  real(8), dimension(:,:), allocatable :: btemp,atemp
	real(8), dimension(:), allocatable :: work
	real(rd), dimension(:,:), allocatable ::  InvS
	real(rd), dimension(:), allocatable :: restmp
	real(rd), dimension(:,:),allocatable :: Smat
  m=size(A(:,1))
  n=size(A(1,:))

  if (size(b).ne.m) stop "Error in fitSVDall size of vector B does not match with matrix A"
  if (size(sigma).ne.m) stop "Error in fitSVDall size of vector B does not match with vector sigma"
  if (size(par).ne.n) stop "Error in fitSVDall size of vector par does not match with matrix A"
  if (present(res).and.(size(res).ne.m)) stop "Error in fitSVDall size of vector res does not match with matrix A"
  if (present(corr).and.(size(corr(:,1)).ne.n).and.(size(corr(1,:)).ne.n))&
       &stop "Error in fitSVDall size of matrix corr does not match with matrix A"

  allocate(s(n),btemp(max(m,n),1),atemp(max(m,n),n),smat(m,m))
	s=0._rd	
  do i=1,m
     atemp(i,:)=a(i,:)/sigma(i)
  enddo !i
  btemp(1:m,1)=b/sigma


#if LAPACK==1   
  CALL GELSS( Atemp, Btemp(1:m,1), RANK, S,1.0e-8_8, INFO=INFO )  
#elif LAPACK==2
	LWORK = MAX( 1, 3*MIN(M,N) + MAX( 2*MIN(M,N), MAX(M,N)) )
	ALLOCATE( WORK(LWORK))
	

	call DGELSS(M,N,1,Atemp,M,Btemp,max(M,N),S,1.0e-8_8,RANK,WORK,LWORK,INFO)

	deallocate(work)

#else
	call GELSS_nolapack(Atemp,Btemp,RANK,s,1.0e-8_rd)
	INFO=0
#endif

  if (INFO.ne.0) write(*,*) 'WARNING: problem in SVD decomposition - INFO value ',INFO
  par(:)=Btemp(1:n,1)
  chisq=(sum(Btemp(n+1:m,1)**2))


  if (present(corr)) then
	!		write(*,*) '   START COMPUTATION OF CORRELATION '
     corr=0.0_rd
!     do i=1,n
!        do k=1,n
!           if(s(k).ne.0.0_rd) then
!						 !corr(i,1:n)=corr(i,1:n)+Atemp(k,i)*Atemp(k,1:n)/s(k)**2
!              corr(1:n,i)=corr(1:n,i)+Atemp(i,i)*Atemp(1:n,k)/s(k)**2
!           endif
!        enddo !k
!     enddo !i
 
 allocate(InvS(n,n))
 invs=0.0_rd 

! write(*,*)'Rank',rank

 do k=1,rank !
	 if (s(k).ne.0._rd) InvS(k,k)=1._rd/s(k)**2
 end do
 corr(1:n,1:n)=matmul(transpose(Atemp(1:n,1:n)),matmul(InvS,Atemp(1:n,1:n)))
 deallocate(invs)


!		write(*,*) '   CORRELATION COMPUTED '
  endif ! present corr

  if (present(res)) then
			write(*,*) '   START COMPUTATION OF RESIDUALS '
      res=b-matmul(A,par)
			allocate(restmp(size(res)))
			restmp=res/sigma
		 	write(*,*) '   RESIDUALS COMPUTED '
     write(*,*)'chi2corr',sum(restmp**2)
		 deallocate(restmp)
  end if
     
  deallocate(s,btemp,atemp)
end subroutine fitSVDall

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine inverse_SVD
!> Subroutine that inverse a matrix by using a SVD decomposition
!> This method used LAPACK subroutines.
!!                 
!> @param[in] A: input squared matrix
!> @param[out]  Ai: output matrix: the inverse of A
!> @param[in] C - optional  : input matrix
!> @param[out]  AiC: output matrix: the inverse of A times C
!> @param[in] minsvd - optional: regularisation: the singular values lower than minsvd are set to 0
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 04/2015 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine inverse_SVD(A,Ai,minsvd)

#if LAPACK==1
  USE mkl95_LAPACK, ONLY: GELSS
!#else
!  use f95_lapack, only:la_gelss
#endif


	implicit none
	real(rd), dimension(:,:), intent(in)           :: a
	real(rd), dimension(:,:), intent(out)          :: ai

	real(rd), optional :: minsvd
	
	real(rd) :: epsilon
	real(rd), dimension(:,:), allocatable :: Atemp,Btemp
	real(rd), dimension(:), allocatable :: S
	integer :: i,m,n,nb,info,lwork,rank
	real(8), dimension(:), allocatable :: work
	epsilon=1.0e-8_8
	if (present(minsvd)) epsilon=minsvd

  m=size(A(:,1))
  n=size(A(1,:))
	if (m.ne.n) stop "ERRROR in INVERSION OF MATRIX: MATRIX NOT SQUARED"
	nb=n

	
  allocate(btemp(n,nb),atemp(n,n),s(n))
	atemp=a
	btemp=0.0_rd
	do i=1,n
		btemp(i,i)=1.0_rd
	end do

!	write(*,*) '   ENTERING LAPACK '
#if LAPACK==1   
  CALL GELSS( Atemp, Btemp, S=S,RCOND=epsilon, INFO=INFO )  
#elif LAPACK==2
	LWORK = MAX( 1, 3*MIN(M,N) + MAX( 2*MIN(M,N), MAX(M,N)),size(Btemp(1,:)) )
	ALLOCATE( WORK(LWORK))
	

	call DGELSS(M,N,size(Btemp(1,:)),Atemp,M,Btemp,max(M,N),S,epsilon,RANK,WORK,LWORK,INFO)
	deallocate(work)

#elif LAPACK!=0
  CALL la_GELSS( Atemp, Btemp,S=S,RCOND=epsilon, INFO=INFO )
#endif
!	write(*,*) '   END OF LAPACK '
  if (INFO.ne.0) write(*,*) 'WARNING: problem in SVD decomposition - INFO value ',INFO

	Ai=Btemp(1:n,1:n)
end subroutine inverse_SVD

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine covsrt
!> Subroutine that put the covariance matrix in the right order (see Numerical Recipes)
!!                 
!> @param[inout] covar
!> @param[in]  maska: logical array to specify which parameters have been estimated
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 11/2011 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
SUBROUTINE covsrt(covar,maska) 
  USE kindtype 
  IMPLICIT NONE 
  REAL(RD), DIMENSION(:,:), INTENT(INOUT) :: covar
  LOGICAL(LGT), DIMENSION(:), INTENT(IN) :: maska

  integer(id) :: ma,mfit,j,k 

  if (size(covar,1)==size(covar,2) .and. size(covar,2)==size(maska)  ) then
     ma=size(maska)
  else
     write(*,*) 'error in lfit :  size(covar,1), size(covar,2) and size(maska) not equal'
     stop
  endif

 
  mfit=count(maska) 
  covar(mfit+1:ma,1:ma)=0.0_rd 
  covar(1:ma,mfit+1:ma)=0.0_rd 
  k=mfit 

  do j=ma,1,-1
     if (maska(j)) then 
        call swap(covar(1:ma,k),covar(1:ma,j)) 
        call swap(covar(k,1:ma),covar(j,1:ma)) 
        k=k-1
     end if
  end do
END SUBROUTINE covsrt  


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine swap
!> Subroutine that exchange two reals (in a tab for example)
!!                 
!> @param[inout] a: real
!> @param[inout]  b: real
!> @post the values of a and b have been exchanged
!> @details see numerical recipes
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 11/2011 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
SUBROUTINE swap(a,b)
  use kindtype
  implicit none
  REAL(rd), DIMENSION(:), INTENT(INOUT) :: a,b
  REAL(rd), DIMENSION(SIZE(a)) :: dum
  dum=a
  a=b
  b=dum
END SUBROUTINE swap


#if LAPACK==0
subroutine GELSS_nolapack( A, B, RANK, S,epsilon) 
		implicit none
		real(8), dimension(:,:), intent(inout) ::a
		real(8), dimension(:,:), intent(inout) :: b
		integer, intent(out) :: rank
		real(8), intent(in)   :: epsilon
		real(8), intent(out),dimension(:) :: s
		
			
		real(rd), dimension(:,:), allocatable :: V, Winv,Acop
		integer :: i,np,mp
		real(rd) :: maxsvd
		np=size(A(1,:))
		mp=size(A(:,1))
	
		allocate(Winv(np,np))
		allocate(V(np,np))

		winv=0._rd
	call	svdcmp(A,s,v)

		maxsvd=0.0_rd
		do i=1,size(S)
			if (s(i)>maxsvd) maxsvd=s(i)
		end do
		rank=0
		do i=1,size(s)
			if(S(i)>maxsvd*epsilon) then
				rank=rank+1
				Winv(i,i)=s(i)
			end if

		end do
		

		do i=1,size(B(1,:))
			B(1:np,i)=matmul(V,matmul(Winv,matmul(transpose(A),B(:,i))))
		end do!i
		A(1:np,1:np)=transpose(V)


end subroutine GELSS_nolapack


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine svdcmp
!> Subroutine that provides a SVD decomposition of a matrix
!!                 
!> @param[inout] A: in: matrix thta we want to decompose ; out: left singular vectors metrix
!> @param[out]  W: singular values
!> @param[out] V  : right singular vector
!> @details numerical recipes
!!
!! @date 04/2015 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
SUBROUTINE svdcmp(a,w,v)
	IMPLICIT NONE
	REAL(rd), DIMENSION(:,:), INTENT(INOUT) :: a
	REAL(rd), DIMENSION(:), INTENT(OUT) :: w
	REAL(rd), DIMENSION(:,:), INTENT(OUT) :: v
	INTEGER :: i,its,j,k,l,m,n,nm
	REAL(rd) :: anorm,c,f,g,h,s,scale,x,y,z
	REAL(rd), DIMENSION(size(a,1)) :: tempm
	REAL(rd), DIMENSION(size(a,2)) :: rv1,tempn
	m=size(a,1)
	n=size(a,2)!,size(v,1),size(v,2),size(w),'svdcmp_dp')
	g=0.0_rd
	scale=0.0_rd
	do i=1,n
		l=i+1
		rv1(i)=scale*g
		g=0.0_rd
		scale=0.0_rd
		if (i <= m) then
			scale=sum(abs(a(i:m,i)))
			if (scale /= 0.0_rd) then
				a(i:m,i)=a(i:m,i)/scale
				s=dot_product(a(i:m,i),a(i:m,i))
				f=a(i,i)
				g=-sign(sqrt(s),f)
				h=f*g-s
				a(i,i)=f-g
				tempn(l:n)=matmul(a(i:m,i),a(i:m,l:n))/h
				a(i:m,l:n)=a(i:m,l:n)+spread(a(i:m,i),dim=2,ncopies=size(tempn(l:n))) * spread(tempn(l:n),dim=1,ncopies=size(a(i:m,i))) ! outerprod(a(i:m,i),tempn(l:n))
				a(i:m,i)=scale*a(i:m,i)
			end if
		end if
		w(i)=scale*g
		g=0.0_rd
		scale=0.0_rd
		if ((i <= m) .and. (i /= n)) then
			scale=sum(abs(a(i,l:n)))
			if (scale /= 0.0_rd) then
				a(i,l:n)=a(i,l:n)/scale
				s=dot_product(a(i,l:n),a(i,l:n))
				f=a(i,l)
				g=-sign(sqrt(s),f)
				h=f*g-s
				a(i,l)=f-g
				rv1(l:n)=a(i,l:n)/h
				tempm(l:m)=matmul(a(l:m,l:n),a(i,l:n))
				a(l:m,l:n)=a(l:m,l:n)+spread(tempm(l:m),dim=2,ncopies=size(rv1(l:n))) * spread(rv1(l:n),dim=1,ncopies=size(tempm(l:m))) 
				a(i,l:n)=scale*a(i,l:n)
			end if
		end if
	end do
	anorm=maxval(abs(w)+abs(rv1))
	do i=n,1,-1
		if (i < n) then
			if (g /= 0.0) then
				v(l:n,i)=(a(i,l:n)/a(i,l))/g
				tempn(l:n)=matmul(a(i,l:n),v(l:n,l:n))
				v(l:n,l:n)=v(l:n,l:n)+spread(v(l:n,i),dim=2,ncopies=size(tempn(l:n))) * spread(tempn(l:n),dim=1,ncopies=size(v(l:n,i)))
			end if
			v(i,l:n)=0.0_rd
			v(l:n,i)=0.0_rd
		end if
		v(i,i)=1.0_rd
		g=rv1(i)
		l=i
	end do
	do i=min(m,n),1,-1
		l=i+1
		g=w(i)
		a(i,l:n)=0.0_rd
		if (g /= 0.0) then
			g=1.0_rd/g
			tempn(l:n)=(matmul(a(l:m,i),a(l:m,l:n))/a(i,i))*g
			a(i:m,l:n)=a(i:m,l:n)+spread(a(i:m,i),dim=2,ncopies=size(tempn(l:n))) * spread(tempn(l:n),dim=1,ncopies=size(a(i:m,i)))
			a(i:m,i)=a(i:m,i)*g
		else
			a(i:m,i)=0.0_rd
		end if
		a(i,i)=a(i,i)+1.0_rd
	end do
	do k=n,1,-1
		do its=1,30
			do l=k,1,-1
				nm=l-1
				if ((abs(rv1(l))+anorm) == anorm) exit
				if ((abs(w(nm))+anorm) == anorm) then
					c=0.0_rd
					s=1.0_rd
					do i=l,k
						f=s*rv1(i)
						rv1(i)=c*rv1(i)
						if ((abs(f)+anorm) == anorm) exit
						g=w(i)
						h=pythag(f,g)
						w(i)=h
						h=1.0_rd/h
						c= (g*h)
						s=-(f*h)
						tempm(1:m)=a(1:m,nm)
						a(1:m,nm)=a(1:m,nm)*c+a(1:m,i)*s
						a(1:m,i)=-tempm(1:m)*s+a(1:m,i)*c
					end do
					exit
				end if
			end do
			z=w(k)
			if (l == k) then
				if (z < 0.0_rd) then
					w(k)=-z
					v(1:n,k)=-v(1:n,k)
				end if
				exit
			end if
			if (its == 30) stop 'svdcmp_dp: no convergence in svdcmp'
			x=w(l)
			nm=k-1
			y=w(nm)
			g=rv1(nm)
			h=rv1(k)
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0_rd*h*y)
			g=pythag(f,1.0_rd)
			f=((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x
			c=1.0_rd
			s=1.0_rd
			do j=l,nm
				i=j+1
				g=rv1(i)
				y=w(i)
				h=s*g
				g=c*g
				z=pythag(f,h)
				rv1(j)=z
				c=f/z
				s=h/z
				f= (x*c)+(g*s)
				g=-(x*s)+(g*c)
				h=y*s
				y=y*c
				tempn(1:n)=v(1:n,j)
				v(1:n,j)=v(1:n,j)*c+v(1:n,i)*s
				v(1:n,i)=-tempn(1:n)*s+v(1:n,i)*c
				z=pythag(f,h)
				w(j)=z
				if (z /= 0.0) then
					z=1.0_rd/z
					c=f*z
					s=h*z
				end if
				f= (c*g)+(s*y)
				x=-(s*g)+(c*y)
				tempm(1:m)=a(1:m,j)
				a(1:m,j)=a(1:m,j)*c+a(1:m,i)*s
				a(1:m,i)=-tempm(1:m)*s+a(1:m,i)*c
			end do
			rv1(l)=0.0_rd
			rv1(k)=f
			w(k)=x
		end do
	end do
END SUBROUTINE svdcmp

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Function pythag
!> Function that computes the sqrt of the sum of the square of two numbers efficiently
!!                 
!> @param[in] a: a real
!> @param[in]  b: a real
!> @return  sqrt(a**2+b**2)
!> @details numerical recipes
!!  
!! @date 04/2015 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
FUNCTION pythag(a,b)
	IMPLICIT NONE
	REAL(rd), INTENT(IN) :: a,b
	REAL(rd) :: pythag
	REAL(rd) :: absa,absb
	absa=abs(a)
	absb=abs(b)
	if (absa > absb) then
		pythag=absa*sqrt(1.0_rd+(absb/absa)**2)
	else
		if (absb == 0.0) then
			pythag=0.0_rd
		else
			pythag=absb*sqrt(1.0_rd+(absa/absb)**2)
		end if
	end if
END FUNCTION pythag
#endif
end module fitMod
