!> @file fitParMod.f90
!> Module doing the least square fit and doing the estimation of the parameters

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Module fitParMod
!> Module doing the estimation of the parameters by using a lesat-square fit. The least-squre fit in itself is in the \link fitmod fitMod module\endlink.
!!
!> @warning compilation option SVDFIT and LAPACK have to be specified SVDFIT=0 -> standard inversion of the normal matrix ; SVDFIT=1: Standard Value Decomposition
!! inversion of the normal matrix. If SVDFIT=1: use of LAPACK95 module. LAPACK variables has then to be specified in the compilation option. LAPACK=1: use of the MKL library (available with INTEL compiler);
!! if LAPACK !=1 use standard LAPACK95 library
!!
!! @author Aurelien Hees, Rhodes University, aurelien.hees@gmail.com
!! @date 01/2015
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

module fitParMod

	use kindtype
	use starsMod, only : star, stars, nstars

implicit none	
	integer(id), private, save :: nobs
	integer(id), private, save :: nparamtofit
	integer(id), private, save :: nglobtofit
	real(rd), private, allocatable, dimension(:,:) :: pder
	real(rd), private, allocatable, dimension(:,:) :: cov
	real(rd), private, allocatable, dimension(:)   :: OMC
	real(rd), private, allocatable, dimension(:)   :: sig
	real(rd), private, allocatable, dimension(:)   :: par
	logical, private, allocatable, dimension(:)    :: parmask
contains


subroutine fit()
	use fitMod, only: fitsvd
	use constantsMod, only:nparamglob,nparamloc,tofitglobal,signext,nparam,timenext
	use modelinterface, only:  nameglob,nameloc

	real(rd) :: chi2
	integer(id) :: i,idx,j
	real(rd),dimension(nparamglob+nparamloc,nparamglob+nparamloc) :: covLoc
!	write(*,*) "   Inversion of the normal matrix"
	
	allocate(cov(nparam,nparam))
	
	
	call fitsvd(pder,OMC,sig,par,parmask,cov,chi2)
	
	

	write(*,*)""

			
	write(*,*)" ================================== "
	write(*,*)" ================================== "
	write(*,*)" Resulting uncertainties: "
	write(*,*)"  == Global parameters: =="
	
	
	do i=1,nparamglob
		if (tofitglobal(i)) then
		    write(*,'(2x,A5,2x,G10.5)') nameGlob(i), sqrt(cov(i,i))
	  else
	      write(*,'(2x,A5,2x,A)') nameGlob(i), ' -NF- '

    end if
	end do
	write(*,*)
	write(*,*)"  == Local parameters: =="
	idx=nparamglob+1
	
	do i=1,nstars
		write(*,*)"  Star: ",stars(i)%name
		do j=1,nparamloc
			if (stars(i)%tofit(j)) then
			    write(*,'(2x,A5,2x,G10.5)') nameLoc(j),sqrt(cov(idx,idx))
		  else
		      write(*,'(2x,A5,2x,A)') nameLoc(j), ' -NF- '
	    end if
			idx=idx+1

		end do!j
		

				
		! compute uncertainty on next point
		if(signext) then
			covLoc=0.0_rd
			do j=1, nparamglob
				if (tofitglobal(i)) then
					covloc(j,1:nparamglob)=cov(j,1:nparamglob)
					covloc(j,nparamglob+1:nparamglob+nparamloc)=cov(j,nparamglob+1+(i-1)*nparamloc:nparamglob+i*nparamloc)
				end if
			end do
			do j=1, nparamloc
				if (stars(i)%tofit(j)) then

					covloc(nparamglob+j,1:nparamglob)=cov(nparamglob+j+(i-1)*nparamloc,1:nparamglob)

					covloc(nparamglob+j,nparamglob+1:nparamglob+nparamloc)=&
							&cov(nparamglob+j+(i-1)*nparamloc,nparamglob+1+(i-1)*nparamloc:nparamglob+i*nparamloc)
				end if
			end do

		
		stars(i)%sigAstronext(1)=sqrt(dot_product(stars(i)%dradpnext,matmul(covloc,stars(i)%dradpnext))) ! uncertainty for RA for one future point
		stars(i)%sigAstronext(2)=sqrt(dot_product(stars(i)%ddecdpnext,matmul(covloc,stars(i)%ddecdpnext)))
		stars(i)%sigRVnext=sqrt(dot_product(stars(i)%drvdpnext,matmul(covloc,stars(i)%drvdpnext)))
		write(*,*) ""
		if(stars(i)%astrob)then
		write(*,'(A,F7.2,A,F7.2)') "   Last astro measurement: ",stars(i)%tastro(stars(i)%na), '. Next astro: ',timenext

		write(*,'(A,G9.4,A,G10.4,A,G9.4)') '  Mean of RA obs.   [mas] : ',stars(i)%meansigastro(1),'  Delta RA for next obs:  ',&
					& stars(i)%diffastronext(1), '  RA uncertainty for next obs.  : ' , stars(i)%sigastronext(1)
					
		write(*,'(A,G9.4,A,G10.4,A,G9.4)') '  Mean of DEC obs.  [mas] : ',stars(i)%meansigastro(2),'  Delta DEC for next obs: ',&
					& stars(i)%diffastronext(2), '  DEC uncertainty for next obs. : ' , stars(i)%sigastronext(2)					
		write(*,*)
	else
		write(*,'(A,F7.2)') '   NO astro measurement. Next astro: ',timenext

		write(*,'(A,G9.4)')  '  RA uncertainty for next obs.  : ' , stars(i)%sigastronext(1)

		write(*,'(A,G9.4)') '  DEC uncertainty for next obs. : ' , stars(i)%sigastronext(2)
		write(*,*)
	end if
	if (stars(i)%rvb)then
		write(*,'(A,F7.2,A,F7.2)') "  Last RV measurement: ",stars(i)%trv(stars(i)%nrv), '. Next RV: ',timenext
		write(*,'(A,G9.4,A,G10.4,A,G9.4)') '  Mean of RV obs.  [km/s] : ',stars(i)%meansigrv,'  Delta RV for next obs:  ',&
			& stars(i)%diffrvnext, '  RV uncertainty for next obs.  : ' , stars(i)%sigrvnext
			write(*,*)
		else
			write(*,'(A,F7.2)') '   NO RV measurement. Next RV: ',timenext

			write(*,'(A,G9.4)') '  RV uncertainty for next obs.  : ' , stars(i)%sigrvnext		
		end if
	end if!sigNext
				write(*,*)
	end do!i

end subroutine fit


subroutine initFit()
	use constantsMod, only :  nparamloc, nparamglob,tofitglobal,nparam,npriors,priors

	integer(id) :: i,idx,j
	nobs=0
	nparamtofit=0
	
	write(*,*)"  =====  INITIALIZATION OF THE FIT ===== "
	do i=1,nstars
		if (stars(i)%astrob) nobs=nobs+2*stars(i)%na
		if (stars(i)%rvb)    nobs=nobs+stars(i)%nrv

		do j=1,nparamloc
			if (stars(i)%tofit(j)) nparamtofit=nparamtofit+1
		end do!j
	end do
	do j=1,nparamglob
		if (toFitGlobal(j))  nparamtofit=nparamtofit+1
	end do !j

	allocate(pder(nobs+npriors,nparam),omc(nobs+npriors),sig(nobs+npriors),par(nparam),parmask(nparam))
	pder=0._rd;omc=0.0_rd;par=0._rd
	idx=1

	parmask(1:nparamglob)=toFitGlobal
	do i=1,nstars

		parmask(nparamglob+1+(i-1)*nparamloc:nparamglob+i*nparamloc)=stars(i)%tofit
		if (stars(i)%astrob) then
			pder(idx:idx+stars(i)%na-1,1:nparamglob)=stars(i)%dradp(:,1:nparamglob)
			pder(idx:idx+stars(i)%na-1,nparamglob+(i-1)*nparamloc+1:nparamglob+i*nparamloc)=stars(i)%dradp(:,nparamglob+1:)
			sig(idx:idx+stars(i)%na-1)=stars(i)%sigra
			idx=idx+stars(i)%na

			pder(idx:idx+stars(i)%na-1,1:nparamglob)=stars(i)%ddecdp(:,1:nparamglob)
			pder(idx:idx+stars(i)%na-1,nparamglob+(i-1)*nparamloc+1:nparamglob+i*nparamloc)=stars(i)%ddecdp(:,nparamglob+1:)
			sig(idx:idx+stars(i)%na-1)=stars(i)%sigdec		
			idx=idx+stars(i)%na	
		end if
		if (stars(i)%rvb)   then
			pder(idx:idx+stars(i)%nrv-1,1:nparamglob)=stars(i)%drvdp(:,1:nparamglob)
			pder(idx:idx+stars(i)%nrv-1,nparamglob+(i-1)*nparamloc+1:nparamglob+i*nparamloc)=stars(i)%drvdp(:,nparamglob+1:)
			sig(idx:idx+stars(i)%nrv-1)=stars(i)%sigrv	
			idx=idx+stars(i)%nrv
		end if
	end do
	if (npriors>0)then
		do i=1,npriors
			pder(idx,priors(i)%idx)=1._rd
			sig(idx)=priors(i)%sig
			idx=idx+1
		end do
	end if!priors

		write(*,*)"  =====  FIT INITIALIZED ===== "
end subroutine initFit


subroutine writeResults(fout)
	use constantsMod
	use modelinterface
	integer(id), intent(in) :: fout
	integer :: isLast,i,idx,j
	logical :: exist
	write(fout,*)  "  ====================================================  "
	write(fout,*)  "         RESULTS    "
	write(fout,*)  "  ====================================================  "
	write(fout,*)" Resulting uncertainties: "
	write(fout,*)"  == Global parameters: =="
	do i=1,nparamglob
		if (tofitglobal(i)) then
			write(fout,'(2x,A5,2x,G10.5)') nameGlob(i), sqrt(cov(i,i))
		else
			write(fout,'(2x,A5,2x,A)') nameGlob(i), '  -NF- '
		end if
	end do

	write(fout,*)
	write(fout,*)"  == Local parameters: =="
	idx=nparamglob+1
	do i=1,nstars
		write(fout,*)"  Star: ",stars(i)%name
		do j=1,nparamloc
			if (stars(i)%tofit(j)) then
				write(fout,'(2x,A5,2x,G10.5)') nameLoc(j),sqrt(cov(idx,idx))
			else
				write(fout,'(2x,A5,2x,A)') nameLoc(j),'  -NF-  '
			end if
			idx=idx+1
		end do!j
		write(fout,*)""
		if (sigNext) then
				write(fout,*) ""			
				if (stars(i)%astrob) then	
				write(fout,'(A,F7.2,A,F7.2)') "   Last astro measurement: ",stars(i)%tastro(stars(i)%na), '. Next astro: ',timenext

				write(fout,'(A,G9.4,A,G10.4,A,G9.4)') '  Mean of RA obs.   [mas] : ',stars(i)%meansigastro(1),'  Delta RA for next obs:  ',&
							& stars(i)%diffastronext(1), '  RA uncertainty for next obs.  : ' , stars(i)%sigastronext(1)

				write(fout,'(A,G9.4,A,G10.4,A,G9.4)') '  Mean of DEC obs.  [mas] : ',stars(i)%meansigastro(2),'  Delta DEC for next obs: ',&
							& stars(i)%diffastronext(2), '  DEC uncertainty for next obs. : ' , stars(i)%sigastronext(2)					
				else
					write(fout,'(A,F7.2)') '   NO astro measurement. Next astro: ',timenext

					write(fout,'(A,G9.4)')  '  RA uncertainty for next obs.  : ' , stars(i)%sigastronext(1)

					write(fout,'(A,G9.4)') '  DEC uncertainty for next obs. : ' , stars(i)%sigastronext(2)
				end if
				write(fout,*)
				if (stars(i)%rvb) then
				write(fout,'(A,F7.2,A,F7.2)') "   Last RV measurement: ",stars(i)%trv(stars(i)%nrv), ' 	 Next RV: ',timenext
				
				write(fout,'(A,G9.4,A,G10.4,A,G9.4)') '  Mean of RV obs.  [km/s] : ',stars(i)%meansigrv,'  Delta RV for next obs:  ',&
					& stars(i)%diffrvnext, '  RV uncertainty for next obs.  : ' , stars(i)%sigrvnext
				else
					write(fout,'(A,F7.2)') '   NO RV measurement. Next RV: ',timenext

					write(fout,'(A,G9.4)') '  RV uncertainty for next obs.  : ' , stars(i)%sigrvnext
					
				end if
				write(fout,*)
		end if
	end do!i
		write(fout,*)""
		write(fout,*)  "  == Correlation matrix == "
		do i=1,nparam
			do j=1,nparam
				if(parmask(i).and.parmask(j)) write(fout,'(F12.6,2x)',ADVANCE='NO') cov(i,j)/sqrt(cov(i,i)*cov(j,j))
!				write(fout,'(i2)',ADVANCE='no')j
				if ((j==nparamglob).and.parmask(i)) write(fout,'(A)',ADVANCE='no')"    "
				if ((mod(j-nparamglob,nparamloc)==0).and.(j>nparamglob).and.parmask(i))write(fout,'(A)',ADVANCE='no')"    "
			end do!j
			if (parmask(i)) write(fout,'(A)',ADVANCE='YES')
			if (i==nparamglob)write(fout,*)" "
			if ((mod(i-nparamglob,nparamloc)==0).and.(i>nparamglob))write(fout,*)" "
		end do!i
	
		   
		   
	     		!==========================================================
       		! OUTPUT DATA file ======================================== 
        ! PARAMETESRS
       	

! FIND THE LAST TIME
       
        open (unit = 34, file = trim(dirout)//'/covout_'//trim(dirout)//'.dat', position='append') !output file for the global parameters
        write(34,'(A,4x)', advance='no') trim(name)
				
		write(34,'(2x,G14.9)',advance='no')  findLastTime()
       	do i=1,nparamglob
     		if (tofitglobal(i)) then
         			write(34,'(2x,G10.5)',advance='no')  sqrt(cov(i,i))
       		else
       			write(34,'(2x,A)',advance='no')'  -NF-       '
       		end if
       	end do!nparam glob

		if (nparamglob==9) write(34,'(2x,G10.5)',advance='no')  lambda5

			!correlations
       	if (nparamglob==8) then
			write(34,'(2x,G11.5,4x,G11.5)', advance='no') cov(1,7)/sqrt(cov(1,1)*cov(7,7)),cov(6,8)/sqrt(cov(6,6)*cov(8,8))
       	else if (nparamglob==10) then
			write(34,'(2x,G11.5,4x,G11.5)', advance='no') cov(1,7)/sqrt(cov(1,1)*cov(7,7)),cov(8,10)/sqrt(cov(8,8)*cov(10,10))
        else
             write(34,'(2x,G11.5)', advance='no') cov(1,7)/sqrt(cov(1,1)*cov(7,7)) 
        end if

       	idx=nparamglob+1
		do i=1,nstars		!!!stars
		write(34,'(A,4x)', advance='no') trim(stars(i)%name)
			do j=1,nparamloc
				if (stars(i)%tofit(j)) then
				if (j==nparamloc) then
					write(34,'(2x,G10.5,4x)',advance='no') sqrt(cov(idx,idx))
				else
					write(34,'(2x,G10.5,4x)',advance='no') sqrt(cov(idx,idx))
				end if
				else
				if ((i==nstars).and.(j==nparamloc)) then
					write(34,'(2x,A,4x)',advance='no')'  -NF-   '
				else
					write(34,'(2x,A,4x)',advance='no')'  -NF-    '
				end if
				end if
				idx=idx+1
			end do!j each star
		end do !i nstar
       	close(34)
         write(fout,*)
        write(fout,*)  
end subroutine writeResults

real(rd) function findLastTime()	
	integer(id) ::i 
	findLastTime = 0._rd
	
	do i=1,nstars
		if (stars(i)%astrob) findLastTime=max(findLastTime,stars(i)%tastro(stars(i)%na))
		if (stars(i)%rvb) findLastTime=max(findLastTime,stars(i)%trv(stars(i)%nrv))		
	end do !
	return 
end function findLastTime
end module fitParMod	
