!> @file odeMod.f90
!! Interface for doing ODE integration with rksuite or Radau

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Module  odeMod
!> Module for interfacing integration of ODE with rksuite, Radau or ODEX methods
!!
!! Different possibilities to use this module: \n
!!   1)  call setup to setup the integrator properties, then integration with integ with initial conditions \n
!!   2)  call setupStart to setup the integrator properties and the initial conditions,
!!   then integration with integ without initial conditions (and each step begin at the end of the previous one)    \n
!!   3)  call setup to setup the integrator properties, call start to setup the initial conditions,
!!   then integration with integ without initial conditions (and each step begin at the end of the previous one)
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com  
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
module odeMod

use kindtype
implicit none


!@@@@@@@@@@@@@@@@
!@ Declarations @
!@@@@@@@@@@@@@@@@


!@ subroutines, modules, functions 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
interface integ
   module procedure integ1start,integ2start,integ1
end interface integ


public  :: integ, setup, setupstart, start
private :: integ1start, integ2start, integ1


integer(id), private, save                             :: typeOde  !< type of solver 1=ra15 1st order, 2=ra15 2nd order, 3=ode23, 4=ode45, 5=ode78, 6=odex
integer(id), private, save                             :: neq      !< number of equations
integer(id), private, save                             :: lenwrk=0  !< needed for rksuite
integer(id), private, save                             :: liwork=0   !< needed for odex
real(rd),    private, save, dimension(:), allocatable  :: work     !< needed for rksuite
integer(id),    private, save, dimension(:), allocatable  :: iwork     !<needed for odex
real(rd), private, save                                :: tol      !< tolerance
real(rd),    private, save                             :: threshI  !< thrsehold for rksuite
real(rd),    private, save                             :: xl       !< step for ra15
real(rd),    private, save, dimension(:), allocatable  :: xcurrent !< current position
real(rd),    private, save, dimension(:), allocatable  :: vcurrent !< current velocity
real(rd),    private, save                             :: tcurrent  !< current time
integer(id), private, save                             :: neqcontrol !< number of equations on which error is estimated
logical(lgt), private, save                            :: errControl !< true if error is controlled



contains

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine setup
!> Subroutine to setup integrator properties
!!
!! @param[in] typeO type of solver 1=RA15 first order, 2=RA15 2nd order, 3=ode23, 4=ode45, 5=ode78, 6=odex
!! @param[in] ne number of equations
!! @param[in] tolp  tolerance of integration, if tolp<0, use of a fixed step size  (for radau)
!! @param[in] thresh1 real treshold for rksuite
!! @param[in] hl optional real representing fixed step size (used for Radau with tolp<0)  
!! @param[in] neqc optional integer: if variable step size, error control is performed on the neqc first equations
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine setup(typeO,NE,ToLp,THRESH1,hl,neqc)
   use kindtype
   use ra15mod
   use rksuitemod
   implicit none
   integer(id), intent(in)             :: typeO
   integer(id), intent(in)             :: ne
   real(rd), intent(in)                :: tolp
   real(rd), intent(in)                :: thresh1
   real(rd), intent(in), optional      :: hl
   integer(id), intent(in), optional   :: neqc
	 integer :: km
   typeode=typeo
   neq=ne
   tol=tolp
   if(present(neqc)) then
      errControl=.true.
      neqcontrol=neqc
   else
      errControl=.false.
   endif

   if (typeode.le.2) then
      if (present(hl)) then
         xl=hl
         if ((TOL.le.0).and.(hl.le.0.0_wp)) then
            write(*,*) 'Error in odeMod.f90: TOL < 0 and xl < 0'
            stop
         endif
      else
         xl=0.0_wp
         if (tol.le.0) then
            write(*,*) 'Error in odeMod: TOL < 0 and xl not furnished'
            stop
         endif
      endif 

      call setupra15(neq,typeode,tolp,neqc)
 
   else if (typeode.le.5) then !RK SUITE
      lenwrk=32_id*neq
      threshI=thresh1
		else if (typeode==6) then !ODEX
			threshi=thresh1
			km=9
			lenwrk=neq*(KM+5)+5*KM+20
			LIWORK=2*km+21

   endif
      if (allocated(work)) deallocate(work)
      if (lenwrk>0) then
				allocate(work(lenwrk))
				work=0.0_rd
			end if
			
      if (allocated(iwork)) deallocate(iwork)
      if (liwork>0) then
				allocate(iwork(liwork))
				iwork=0
			end if
		 
     
      if (allocated(xcurrent)) deallocate(xcurrent)
      allocate (xcurrent(neq))

      if (allocated(vcurrent)) deallocate(vcurrent)
      allocate (vcurrent(neq))
end subroutine setup



!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine setupStart
!> Subroutine to setup integrator properties and the initial conditions
!!
!! @param[in] typeO type of solver 1=RA15 first order, 2=RA15 2nd order, 3=ode23, 4=ode45, 5=ode78, 6=odex
!! @param[in] ne number of equations
!! @param[in] tstart starting time
!! @param[in] ystart array of reals: initial conditions
!! @param[in] tend end time
!! @param[in] tolp  tolerance of integration given, if tolp<0, use of a fixed step size  (for radau)
!! @param[in] thresh1 real treshold for rksuite
!! @param[in] vstart optional array of reals: initial velocities (only with type)=2
!! @param[in] hl optional real representing fixed step size (used for Radau with tolp<0)  
!! @param[in] neqc optional integer: if variable step size, error control is performed on the neqc first equations
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine setupStart(typeO,NE,TSTART,YSTART,TEND,ToLp,THRESH1,vstart,hl,neqc)
   use kindtype
   use ra15mod
   use rksuitemod
   implicit none
   integer(id), intent(in)             :: typeO
   integer(id), intent(in)             :: ne
   real(rd), intent(in)                :: tstart
   real(rd), intent(in), dimension(:) :: ystart
   real(rd), intent(in)                :: tend
   real(rd), intent(in)                :: tolp
   real(rd), intent(in)                :: thresh1
   real(rd),intent(in),dimension(ne), optional:: vstart
   real(rd), intent(in), optional      :: hl
   integer(id), intent(in), optional   :: neqc

   real(rd), dimension(ne)             :: thresh

   call setup(typeO,NE,ToLp,THRESH1,hl,neqc)
   call start(tstart,YSTART,TEND,vstart)
end subroutine setupStart

!!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine start
!> Subroutine to setup the initial conditions for the integration
!!
!! @param[in] tstart starting time
!! @param[in] ystart array of reals: initial conditions
!! @param[in] tend end time
!! @param[in] vstart optional array of reals: initial velocities (only with type)=2
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine start(tstart,YSTART,TEND,vstart)
  use kindtype
  use rksuitemod
  implicit none
  real(rd), intent(in)                :: tstart
  real(rd), intent(in), dimension(neq) :: ystart
  real(rd), intent(in)                :: tend
  real(rd),intent(in),dimension(neq), optional:: vstart
  
  real(rd), dimension(neq)            :: thresh

  thresh(:)=threshI

  xcurrent=ystart

  if (typeOde.eq.2) then
     if (.not.(present(vstart))) then
        write(*,*) 'Error: using ra15 with 2nd order equations and initial velocity not given'
        stop
     endif
     vcurrent=vstart
  endif
  tcurrent=tstart
   
  if (typeOde.ge.3) then
   ! write(*,*) 'INTEG', neq, tstart, tend, tol, thresh, typeode-2!, work, lenwrk
     if(errControl) then
        call setuprk(NEQ,TSTART,YSTART,TEND,abs(TOL),THRESh,METHOD=typeode-2,TASK='U' ,&
             & ERRASS=.FALSE.,HSTART=0.0_rd,WORK=work,LENWRK=lenwrk,MESAGE=.TRUE.,neqc=neqcontrol)
	elseif ((typeOde.ge.3).and.(typeode.le.5)) then
        call setuprk(NEQ,TSTART,YSTART,TEND,abs(TOL),THRESh,METHOD=typeode-2,TASK='U' ,&
             & ERRASS=.FALSE.,HSTART=0.0_rd,WORK=work,LENWRK=lenwrk,MESAGE=.TRUE.)
     endif
  endif
end subroutine start

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine integ1
!> Subroutine to perform integration from current time (end of previous integration) 
!!
!! @param[in] f external function to be integrated
!! @param[in] tf real: final time
!! @param[out] x array of reals: results of integration
!! @param[out] v optional array of reals: velocity after integration
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine integ1(f,tf,x,v)
  use kindtype
  use ra15mod
  use rksuitemod
	use odexmod
  implicit none
  external f

  real(rd), intent(in) :: tf
  real(rd), intent(out), dimension(neq) :: x
  real(rd), intent(out), dimension(neq),optional :: v

  
  real(rd) :: tgot,h
  real(rd), dimension(neq) :: temp, temp2
  integer(id)               :: UFLAG,idid

  if (typeode.le.2) then    
     call RA15M(f,Xcurrent,Vcurrent,Tcurrent,TF,XL) 
     tcurrent=tf
   else if (typeode.le.5) then    

      call UT(F,tf,tgot,xcurrent,YPGOT=temp,YMAX=temp2,WORK=work,UFLAG=uflag)
      if(uflag==3)  call UT(F,tf,tgot,xcurrent,YPGOT=temp,YMAX=temp2,WORK=work,UFLAG=uflag)

      tcurrent=tgot

      if(tcurrent.ne.tgot) write(*,*)'WARNING in odeMod.f90: tcurrent diff from t wanted: ',tcurrent,tf
      if (UFLAG.eq.1_id) then

      else if ((UFLAG.eq.2_id).or.(UFLAG.eq.3_id).or.(UFLAG.eq.4_id)) then
         WRITE(*,'(A/,I1,A)') 'Warning appearing during the integration - see rksuite for more details (UFLAG=',UFLAG,')'
      else 
         WRITE(*,'(A/,I1,A)') 'Integration failed - see rksuite for more details (UFLAG=',UFLAG,')'
      ENDIF
			
		else if (typeode==6) then
			h=0.0_rd
			iwork(1)=100000
			call ODEX(Neq,f,tcurrent,xcurrent,XEND=tf,H=h, RTOL=(/TOL/),ATOL=(/threshI/),ITOL=0,SOLOUT=SOLOUT,IOUT=0, &
						WORK=WoRK,LWORK=LenWRK,IWORK=IWORK,LIWORK=LIWORK,IDID=IDID)
						if(IDID==-1) then
							WRITE(*,*) ' INTEGRATION WITH ODEX FAILED AT t= ',tcurrent
							stop
						end if
   endif

   x=xcurrent
   if (present(v))  v=vcurrent
   
end subroutine integ1

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine integ1Start
!> Subroutine to perform integration between two given time (2nd order ODE)
!!
!! @param[in] f external function to be integrated
!! @param[in] td real: starting time  
!! @param[in] tf real: final time  
!! @param[in,out] x array of reals: initial conditions and results of integration
!! @param[in,out] v array of reals: initial and final velocity
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine integ1start(f,td,tf,x,v)
  use kindtype
  implicit none
  external f
  real(rd), intent(in)    :: td
  real(rd), intent(inout)    :: tf
  real(rd), intent(inout), dimension(neq) :: x
  real(rd), intent(inout), dimension(neq) :: v
  !real(rd), intent(out), dimension(neq), optional :: xp

  call start(td,x,tf,v)
  call integ1(f,tf,x,v)
end subroutine integ1start

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Subroutine integ2Start
!> Subroutine to perform integration between two given time (1std order ODE)
!!
!! @param[in] f external function to be integrated
!! @param[in] td real: starting time  
!! @param[in] tf real: final time  
!! @param[in,out] x array of reals: initial conditions and results of integration
!!
!! @author Aurelien Hees, ORB, aurelien.hees@gmail.com
!! @date 05/2010 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine integ2start(f,td,tf,x)
  use kindtype
  use ra15mod
  use rksuitemod
  implicit none
  external f
  real(rd), intent(in)    :: td
  real(rd), intent(inout)    :: tf
  real(rd), intent(inout), dimension(neq) :: x
 ! real(rd), intent(out), dimension(neq), optional ::xp

  real(rd), dimension(neq) :: temp
  temp=0.0_rd
  
  call integ1start(f,td,tf,x,v=temp)
end subroutine integ2start

subroutine solout
end subroutine solout
end module odeMod
