!> @file constantsMod.f90
!! Definition of all the constants used in the simulations (some of them may be loaded from INPOP)

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Module  constantsMod
!> Module where the important constants used in the whole simulations are defined
!>
!> Compilation requirred preprocessor variables INPOP defined. If INPOP=1, the compilation is done with a link to 
!> INPOP planetary ephemerides. This case is requirred for including perturbations from planets. In this case,
!> the compilation needs to be done with a link to CALCEPH library. Moreover, the preprocessor variable INPOPFILE needs to
!> be defined and needs to point towards an INPOP inut file.
!!
!! @author Aurelien Hees, Rhodes University, aurelien.hees@gmail.com
!! @date 01/2015
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
module starsMod
  use kindtype
  implicit none


type, public :: star
		character(len=20), public  :: name
		real(rd), allocatable, dimension(:), public  :: par
		logical, allocatable, dimension(:), public   :: tofit

		integer(id), public                          :: na=0
		integer(id), public                          :: nrv=0
		character(len=100), public                   :: astroFile
		character(len=100), public                   :: rvFile

		real(rd), allocatable, dimension(:), public  :: tastro
		real(rd), allocatable, dimension(:), public  :: trv
		logical, public                              :: astrob=.false.
		logical, public                              :: rvb=.false.
		real(rd), allocatable, dimension(:), public  :: sigra
		real(rd), allocatable, dimension(:), public  :: sigdec
		real(rd), allocatable, dimension(:), public  :: sigrv			
		real(rd), allocatable, dimension(:), public  :: ra
		real(rd), allocatable, dimension(:), public  :: dec
		real(rd), allocatable, dimension(:), public  :: rv
		real(rd), public                             :: sigFutRa
		real(rd), public                             :: sigFutDec
		real(rd), public                             :: sigFutRV			
		
		real(rd), allocatable, dimension(:,:),public :: dradp			
		real(rd), allocatable, dimension(:,:),public :: ddecdp
		real(rd), allocatable, dimension(:,:),public :: drvdp
		real(rd), public                             :: sigRVnext
		real(rd), public, dimension(2)               :: sigAstronext
		real(rd), public                             :: diffRVnext
		real(rd), public, dimension(2)               :: diffAstronext
		real(rd), public                             :: meansigRV
		real(rd), public, dimension(2)               :: meanSigAstro
		
		real(rd), public                             :: RVnext
		real(rd), public                             :: ranext
		real(rd), public                             :: decnext			
		real(rd), allocatable, dimension(:),public   :: dradpnext			
		real(rd), allocatable, dimension(:),public   :: ddecdpnext
		real(rd), allocatable, dimension(:),public   :: drvdpnext
		
		! for sensitivity and weather
		integer(id)           :: npastastro=0
    integer(id)           :: npastrv=0
    integer(id)           :: dnrv=0
    integer(id)           :: dnastro=0
    
    logical               :: included=.false.
    integer(id)           :: ntofit=-1 !number of parameters (local) to be fitted
    integer(id), dimension(:),allocatable :: idx_fit !array with the index of parameters to fit
     
    real(rd), dimension(:,:),allocatable  :: D !< matrix D - partial derivatives wrt the local parameters; D_{ij}=partial O_i / partial x^j
!       real(rd), dimension(:,:),allocatable  :: B!< matrix B - partial derivatives wrt the global parameters ; B_{ij}=partial O_i / partial p_glob^j
		real(rd), dimension(:,:),allocatable  :: BtB !< product of transpose B and B
		real(rd), dimension(:,:),allocatable  :: BtD!< product of transpose B and D
		real(rd), dimension(:,:), allocatable :: Ui !< 
		real(rd), dimension(:,:), allocatable :: invDD !< inverse matrix of transpose(D) times D
end type

type(star), dimension(:), public, allocatable, save :: stars
integer(id), public, save                             :: NStars !< number of stars integrated
!@ variables public
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

contains

subroutine loadStar(s,nfutastro,nfutrv)
	use constantsMod, only : countLines,nparamloc,nparamglob,signext,fromMJD2year
	type(star), intent(inout) :: s
	integer(id), intent(in), optional :: nfutastro! for cadence 
	integer(id), intent(in), optional :: nfutrv   ! for cadence 
	real(rd),dimension(2) :: tmp
	real(rd) :: tmpyr
	integer(id) ::i , nlines
	write(*,*)" ============ LOADING STAR :",trim(s%name)
	
	! ASTROMETRIC OBSERVATIONS
	if (trim(s%astroFile)=="None") then
		s%astrob=.False.
		s%na = 0
	else
		s%astrob=.True.
		nlines=countLines(s%astroFile)
		s%na=nlines
		s%npastastro=s%na
		if (present(nfutastro)) then
		  
		  s%na=s%na+nfutastro
	  end if
		allocate(s%tastro(s%na));		allocate(s%sigra(s%na));allocate(s%sigdec(s%na));allocate(s%ra(s%na));allocate(s%dec(s%na));
		allocate(s%dradp(s%na,nparamLoc+nparamGlob))
		allocate(s%ddecdp(s%na,nparamLoc+nparamGlob))
		
		write(*,*) "     -- loading astro data --"
		open ( unit = 21, form='formatted', file=trim(s%astroFile))

		do i=1,nlines
			read(21,*)tmpyr,tmp,s%sigra(i),s%sigdec(i),s%tastro(i)
!					read(21,*)s%tastro(i),tmp,s%sigra(i),s%sigdec(i),tmpyr
		end do!i
		close(21)
		call fromMJD2year(s%tastro)
		
		s%sigra=s%sigra*1000.0_rd;s%sigdec=s%sigdec*1000.0_rd !going from as to mas

		s%meansigastro(1)=sum(s%sigra)/s%na; 		s%meansigastro(2)=sum(s%sigdec)/s%na;
		write(*,*) "     -- astro data loaded --"		
	end if!astro
	
	! RV OBSERVATIONS
	if (trim(s%rvFile)=="None") then
		s%rvb=.False.
		s%nrv=0
	else
		s%rvb=.True.
		nlines=countLines(s%rvFile)
		s%nrv=nlines
		s%npastrv=s%nrv
		if (present(nfutrv)) then
		  s%npastrv=s%nrv
		  s%nrv=s%nrv+nfutrv
	  end if
	  
		allocate(s%trv(s%nrv));		allocate(s%sigrv(s%nrv));	allocate(s%rv(s%nrv));
		allocate(s%drvdp(s%nrv,nparamLoc+nparamGlob))
		write(*,*) "     -- loading RV data --"
		open ( unit = 22, form='formatted', file=trim(s%rvFile))
		do i=1,nlines
			read(22,*)tmpyr,tmp(1),s%sigrv(i),s%trv(i)
!			read(22,*)s%trv(i),tmp(1),s%sigrv(i),tmpyr
		end do!i
		call fromMJD2year(s%trv)
		close(22)
		s%meansigrv=sum(s%sigrv)/s%nrv
		write(*,*) "     -- RV data loaded --"		
	end if! RV

	if(sigNext) then
		allocate(s%dradpnext(nparamLoc+nparamGlob));		allocate(s%ddecdpnext(nparamLoc+nparamGlob));
		allocate(s%drvdpnext(nparamLoc+nparamGlob));
	end if
end subroutine loadStar

subroutine upgrade_block(s,type,idx)
    use constantsMod, only: NPARAMGLOB,nglobtofit, idx_glob
    type(star), intent(inout)             :: s
    integer, intent(in)                   :: type !1=astro, 2=RV
    integer(id), intent(in)               :: idx
    
    real(rd), dimension(:,:), allocatable :: b,d
    integer(id) :: npastobs
    npastobs = 2*s%npastastro+s%npastrv
    if (type==1) then
      allocate(b(nglobtofit,2),d(s%ntofit,2))
      b(:,1)=s%dradp(idx,idx_glob)/s%sigra(idx)
      b(:,2)=s%ddecdp(idx,idx_glob)/s%sigdec(idx)      
      d(:,1)=s%dradp(idx,NPARAMGLOB+s%idx_fit)/s%sigra(idx)
      d(:,2)=s%ddecdp(idx,NPARAMGLOB+s%idx_fit)/s%sigdec(idx)
      
!      s%D(npastobs+1,:) = d(:,1)
 !     s%D(npastobs+2,:) = d(:,2)
      
      s%npastastro=s%npastastro+1
      s%dnastro=s%dnastro-1
      
    else if (type==2) then 
      allocate(b(nglobtofit,1),d(s%ntofit,1))
      b(:,1)=s%drvdp(idx,idx_glob)/s%sigrv(idx)
      d(:,1)=s%drvdp(idx,NPARAMGLOB+s%idx_fit)/s%sigrv(idx)
!      s%D(npastobs+1,:) = d(:,1)
      s%npastrv=s%npastrv+1
      s%dnrv=s%dnrv-1
      
    end if
      s%btb =  matmul(b,transpose(b)) +  s%btb
      s%btD =  matmul(b,transpose(D)) +  s%btD


    deallocate(b,d)
end subroutine upgrade_block

subroutine init_block_star(s,past)
  use constantsMod, only: NPARAMGLOB,nglobtofit,idx_glob!,NPARAMLOC
  type(star), intent(inout) :: s
  logical, optional, intent(in) :: past
  logical :: ispast
  integer(id) :: npastastro,npastrv
  real(rd), dimension(:,:), allocatable    :: btmp,dtmp

  integer(id) :: j

  ispast=.false.
  npastastro=s%na  
  npastrv=s%nrv
  if (present(past)) then
    if(past) then
      npastastro=s%npastastro
       npastrv=s%npastrv
       ispast=.true.
    end if
  end if
  if (allocated(s%D)) deallocate(s%D)
  allocate(s%D(2*s%na+s%nrv,s%ntofit))
  
  if (allocated(s%BtD)) deallocate(s%BtD)
  allocate(s%BtD(nglobtofit,s%ntofit))
  
  if (allocated(s%invDD)) deallocate(s%invDD)
  allocate(s%invDD(s%ntofit,s%ntofit))
  
  if (allocated(s%BtB)) deallocate(s%BTB)
  allocate(s%BTB(nglobtofit,nglobtofit))
  
  if (allocated(s%Ui)) deallocate(s%Ui)
  allocate(s%Ui(nglobtofit,nglobtofit))
  
  allocate(btmp(2*npastastro+npastrv,nglobtofit))
  allocate(dtmp(2*npastastro+npastrv,s%ntofit))

  if (s%astrob) then
    do j=1,s%na
      s%D(2*j-1,:) = s%dradp(j,NPARAMGLOB+s%idx_fit)/s%sigra(j)
      s%D(2*j,:) = s%ddecdp(j,NPARAMGLOB+s%idx_fit)/s%sigdec(j)
      if (j<=npastastro)      then 
        btmp(2*j-1,:) = s%dradp(j,idx_glob)/s%sigra(j)
        btmp(2*j,:)   = s%ddecdp(j,idx_glob)/s%sigdec(j)
        dtmp(2*j-1,:) = s%D(2*j-1,:)
        dtmp(2*j,:) =  s%D(2*j,:)
      end if
    end do!j
  end if
  if (s%rvb) then 
    do j=1,s%nrv 
      s%D(2*s%na+j,:) = s%drvdp(j,NPARAMGLOB+s%idx_fit)/s%sigrv(j)
      if (j<=npastrv) then
        btmp(2*npastastro+j,:) = s%drvdp(j,idx_glob)/s%sigrv(j)
        dtmp(2*npastastro+j,:) =  s%D(2*s%na+j,:)
      end if
    end do
  end if

  s%btd= matmul(transpose(btmp),dtmp) 
  s%btb= matmul(transpose(btmp),btmp)
   
  deallocate(btmp,dtmp)
  call inverse_star(s,ispast)
end subroutine init_block_star

subroutine inverse_star(s,past)
  use fitmod, only : fitsvd
  type(star), intent(inout) :: s
  logical, optional, intent(in) :: past
  real(rd), dimension(:), allocatable    :: omc,sig,par
  real(rd) :: chi2
  logical, dimension(:), allocatable :: parmask
  real(rd), dimension(:,:), allocatable :: Dtmp
  integer(id):: npastastro,npastrv
  npastastro=s%na  
  npastrv=s%nrv
  if (present(past)) then
    if(past) then
      npastastro=s%npastastro
      npastrv=s%npastrv
    end if
  end if
  allocate(OMC(2*npastastro+npastrv),sig(2*npastastro+npastrv),par(s%ntofit),parmask(s%ntofit))
  OMC=0._rd;sig=1._rd;par=0._rd;parmask=.true.

  allocate(Dtmp(2*npastastro+npastrv,s%ntofit))
  Dtmp(1:2*npastastro,:) = s%D(1:2*npastastro,:)
  Dtmp(2*npastastro+1:2*npastastro+npastrv,:) = s%D(2*s%na+1:2*s%na+npastrv,:)
  call fitsvd(dtmp,OMC,sig,par,parmask,s%invDD,chi2)
  s%ui=s%btb -  matmul(s%btd,matmul(s%invDD,transpose(s%btd)))
   deallocate(omc,sig,par,parmask,dtmp)
end subroutine inverse_star

function star_included(s)
  use constantsMod, only: NPARAMGLOB, NPARAMLOC
  logical :: star_included
  type(star), intent(in) :: s
  
  real(rd) :: time
  if ((s%npastastro > 0).and.(s%npastrv>0)) then
    time = max(s%tastro(s%npastastro),s%trv(s%npastrv))- min(s%tastro(1),s%trv(1))
  else if (s%npastastro > 0) then
    time = s%tastro(s%npastastro)-s%tastro(1)
  else if (s%npastrv>0) then 
    time = s%trv(s%npastrv) - s%trv(1)
  else 
    time = 0._rd
  end if
    
  star_included = ( ( 2*s%npastastro+s%npastrv > (NPARAMLOC + NPARAMGLOB) ) .AND. time > 1.0_rd*s%par(1) ) &
        & .or.( 2*s%npastastro+s%npastrv > 5*(NPARAMLOC + NPARAMGLOB) )
  return
end function star_included

end module starsMod
