!> @file rksuiteMod.f90
!! RKsuite for integration ODE

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!  Module  rksuiteMod
!> Module  used for the integrator (implements an ode23, ode45 and ode78 integrator
!!
!!  Informations: more informations on http://www.netlib.org/ode/rksuite/
!! Transcription from fortran 77 to 90 by A. Hees but useless since it can be found online (which is probably better)
!! I also add a variable (neqc) to control the error on only some of the equation
!!
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

module rksuitemod


 use kindtype
 implicit none
 public :: setuprk, ut, ct, stat, GLBERR,INTRP,reset
 private ::  MCONST, CONST, FORMI, EVALI, rkmsg, rksit, truerr, step, stepa, stepb, stiff, stiffa, stiffb, stiffc, stiffd,&
softfl, chkfl, dotprd

 integer(id), private, save :: neqcontrol
contains

     SUBROUTINE SETUPrk(NEQ,TSTART,YSTART,TEND,TOL,THRES,METHOD,TASK,&
     &                 ERRASS,HSTART,WORK,LENWRK,MESAGE,neqc)
!C$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!C
!C  If you are not familiar with the code SETUP and how it is used in
!C  conjunction with UT or CT to solve initial value problems, you should study
!C  the document file rksuite.doc carefully before attempting to use the code.
!C  The following "Brief Reminder" is intended only to remind you of the
!C  meaning, type, and size requirements of the arguments.
!C
!C  The environmental parameters OUTCH, MCHEPS, and DWARF are used in the
!C  following description.  To find out their values
!C
!C       CALL ENVIRN(OUTCH,MCHEPS,DWARF)
!C
!C  INPUT VARIABLES
!C
!C     NEQ       - INTEGER(id)
!C                 The number of differential equations in the system.
!C                 Constraint: NEQ >= 1
!C     TSTART    - REAL(rd)
!C                 The initial value of the independent variable.
!C     YSTART(*) - REAL(rd) array of length NEQ
!C                 The vector of initial values of the solution components.
!C     TEND      - REAL(rd)
!C                 The integration proceeds from TSTART in the direction of
!C                 TEND. You cannot go past TEND.
!C                 Constraint: TEND must be clearly distinguishable from TSTART
!C                 in the precision available.
!C     TOL       - REAL(rd)
!C                 The relative error tolerance.
!C                 Constraint: 0.01_rd >= TOL >= 10*MCHEPS
!C     THRES(*)  - REAL(rd) array of length NEQ
!C                 THRES(L) is the threshold for the Ith solution component.
!C                 Constraint: THRES(L) >= SQRT(DWARF)
!C     METHOD    - INTEGER(id)
!C                 Specifies which Runge-Kutta pair is to be used.
!C                  = 1 - use the (2,3) pair
!C                  = 2 - use the (4,5) pair
!C                  = 3 - use the (7,8) pair
!C     TASK      - CHARACTER*(*)
!C                 Only the first character of TASK is significant.
!C                 TASK(1:1) = `U' or `u' - UT is to be used
!C                           = `C' or `c' - CT is to be used
!C                 Constraint: TASK(1:1) = `U'or `u' or`C' or `c'
!C     ERRASS    - LOGICAL
!C                 = .FALSE. - do not attempt to assess the true error.
!C                 = .TRUE.  - assess the true error. Costs roughly twice
!C                             as much as the integration with METHODs 2 and
!C                             3, and three times with METHOD = 1.
!C     HSTART    - REAL(rd)
!C                 0.0_rd     - select automatically the first step size.
!C                 non-zero  - try HSTART for the first step.
!C
!C  WORKSPACE
!C
!C     WORK(*) - REAL(rd) array of length LENWRK
!C               Do not alter the contents of this array after calling SETUP.
!C
!C  INPUT VARIABLES
!C
!C     LENWRK  - INTEGER(id)
!C               Length of WORK(*): How big LENWRK must be depends
!C               on the task and how it is to be solved.
!C
!C               LENWRK = 32*NEQ is sufficient for all cases. 
!C
!C               If storage is a problem, the least storage possible 
!C               in the various cases is:
!C
!C                 If TASK = `U' or `u', then
!C                   if ERRASS = .FALSE. and
!C                     METHOD = 1, LENWRK must be at least 10*NEQ
!C                            = 2                          20*NEQ
!C                            = 3                          16*NEQ
!C                   if ERRASS = .TRUE. and
!C                     METHOD = 1, LENWRK must be at least 15*NEQ
!C                            = 2                          32*NEQ
!C                            = 3                          21*NEQ
!C
!C                 If TASK = `C' or `c', then
!C                   if ERRASS = .FALSE. and
!C                     METHOD = 1, LENWRK must be at least 10*NEQ
!C                            = 2                          14*NEQ
!C                            = 3                          16*NEQ
!C                   if ERRASS = .TRUE. and
!C                     METHOD = 1, LENWRK must be at least 15*NEQ
!C                            = 2                          26*NEQ
!C                            = 3                          21*NEQ
!C
!C                 Warning:  To exploit the interpolation capability
!C                 of METHODs 1 and 2, you have to call INTRP.  This
!C                 subroutine requires working storage in addition to
!C                 that specified here.
!C
!C     MESAGE    - LOGICAL
!C                 Specifies whether you want informative messages written to
!C                 the standard output channel OUTCH.
!C                 = .TRUE.   - provide messages
!C                 = .FALSE.  - do not provide messages
!C
!C  In the event of a "catastrophic" failure to call SETUP correctly, the
!C  nature of the catastrophe is reported on the standard output channel,
!C  regardless of the value of MESAGE.  Unless special provision was made
!C  in advance (see rksuite.doc), the computation then comes to a STOP.
!C
!C$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
  USE kindtype

  implicit none
!     .. Scalar Arguments ..
      REAL(rd)  HSTART, TEND, TOL, TSTART
      INTEGER(id)          LENWRK, METHOD, NEQ
      LOGICAL(lgt)           ERRASS, MESAGE
      integer(id), intent(in), optional :: neqc
      CHARACTER*(*)     TASK
!   .. Array Arguments ..
      REAL(rd)  THRES(*), WORK(*), YSTART(*)
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!C     .. Common Block for General Workspace Pointers ..
      INTEGER(id)           PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      COMMON /RKCOM3/   PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      SAVE   /RKCOM3/
!C     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!C     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!C     .. Common Block for Global Error Assessment ..
      REAL(rd)  MAXERR, LOCMAX
      INTEGER(id)           GNFCN, PRZSTG, PRZY, PRZYP, PRZERS, PRZERR,&
     &                  PRZYNU
      LOGICAL(lgt)           ERASON, ERASFL
      COMMON /RKCOM6/   MAXERR, LOCMAX, GNFCN, PRZSTG, PRZY, PRZYP,&
     &                  PRZERS, PRZERR, PRZYNU, ERASON, ERASFL
      SAVE   /RKCOM6/
!C     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!C     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!C     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!C     .. Parameters ..
      CHARACTER*6       SRNAME
      PARAMETER         (SRNAME='SETUP')
      INTEGER(id)           MINUS1
      LOGICAL(lgt)           TELL
      PARAMETER         (MINUS1=-1_id,TELL=.FALSE.)
      REAL(rd)  ONE, ZERO, PT01
      PARAMETER         (ONE=1.0_rd,ZERO=0.0_rd,PT01=0.01_rd)
!C     .. Local Scalars ..
      REAL(rd)  HMIN
      INTEGER(id)           FLAG, FREEPR, IER, L, LINTPL, LREQ, NREC, VECSTG
      LOGICAL(lgt)           LEGALT, REQSTG
      CHARACTER         TASK1
!C     .. External Subroutines ..
      !EXTERNAL          CONST, MCONST, RKMSG, RKSIT
!C     .. Intrinsic Functions ..
      INTRINSIC         ABS, MAX, SIGN
!C     .. Executable Statements ..
!C
!C!  Clear previous flag values of subprograms in the suite.
!C 
      neqcontrol=neq
      if(present(neqc))   neqcontrol=neqc


      IER = MINUS1
      CALL RKSIT(TELL,SRNAME,IER)
!C
      IER = 1_id
      NREC = 0_id
!C
!C  Fetch output channel and machine constants; initialise common
!C  block /RKCOM7/
!C
      CALL MCONST(METHOD)
!C
!C  Check for valid input of trivial arguments
      TASK1 = TASK(1:1)
      LEGALT = TASK1 .EQ. 'U' .OR. TASK1 .EQ. 'u' .OR.&
     &         TASK1 .EQ. 'C' .OR. TASK1 .EQ. 'c'
      IF (.NOT.LEGALT) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A/A,A,A/A)') &
     &' ** You have set the first character of ',&
     &' ** TASK to be ''',TASK1,'''. It must be one of ',&
     &' ** ''U'',''u'',''C'' or ''c''.'
      ELSE IF (NEQ.LT.1) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A,I6,A)')&
     &' ** You have set NEQ = ',NEQ,' which is less than 1.'
      ELSE IF (METHOD.LT.1 .OR. METHOD.GT.3) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A,I6,A)')&
     &' ** You have set METHOD = ',METHOD,' which is not 1, 2, or 3.'
      ELSE IF (TSTART.EQ.TEND) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A,D13.5,A)')&
     &' ** You have set TSTART = TEND = ',TSTART,'.'
      ELSE IF ((TOL.GT.PT01) .OR. (TOL.LT.RNDOFF)) THEN
         IER = 911_id
         NREC = 2_id
         WRITE (REC,'(A,D13.5,A/A,D13.5,A)')&
     &' ** You have set TOL = ',TOL,' which is not permitted. The',&
     &' ** range of permitted values is (',RNDOFF,',0.01_rd).'
      ELSE
         L = 1_id
   20    CONTINUE
         IF (THRES(L).LT.TINY) THEN
            IER = 911_id
            NREC = 2_id
            WRITE (REC,'(A,I6,A,D13.5,A/A,D13.5,A)')&
     &' ** You have set THRES(',L,') to be ',THRES(L),' which is ',&
     &' ** less than the permitted minimum,',TINY,'.'
         END IF
         L = L + 1_id
         IF (IER.NE.911 .AND. L.LE.NEQ) GO TO 20
      END IF
!C
!C  Return if error detected
!C
      IF (IER.NE.1) GO TO 80
!C
!C  Set formula definitions and characteristics by means of arguments
!C  in the call list and COMMON blocks /RKCOM4/ and /RKCOM5/
!C
      CALL CONST(METHOD,VECSTG,REQSTG,LINTPL)
!C
!C  Set options in /RKCOM8/
      UTASK = TASK1 .EQ. 'U' .OR. TASK1 .EQ. 'u'
      MSG = MESAGE
!C
!  Initialise problem status in /RKCOM1/ and /RKCOM2/
      NEQN = NEQ
      TSTRT = TSTART
      TND = TEND
      T = TSTART
      TOLD = TSTART
      DIR = SIGN(ONE,TEND-TSTART)
!C
!  In CT the first step taken will have magnitude H.  If HSTRT = ABS(HSTART)
!  is not equal to zero, H = HSTRT.  If HSTRT is equal to zero, the code is
!  to find an on-scale initial step size H.  To start this process, H is set
!  here to an upper bound on the first step size that reflects the scale of
!  the independent variable.  UT has some additional information, namely the
!  first output point, that is used to refine this bound in UT when UTASK
!  is .TRUE..  If HSTRT is not zero, but it is either too big or too small,
!  the input HSTART is ignored and HSTRT is set to zero to activate the
!  automatic determination of an on-scale initial step size.
!C
      HSTRT = ABS(HSTART)
      HMIN = MAX(TINY,TOOSML*MAX(ABS(TSTART),ABS(TEND)))
      IF (HSTRT.GT.ABS(TEND-TSTART) .OR. HSTRT.LT.HMIN) HSTRT = ZERO
      IF (HSTRT.EQ.ZERO) THEN
         H = MAX(ABS(TEND-TSTART)/RS3,HMIN)
      ELSE
         H = HSTRT
      END IF
      HOLD = ZERO
      TOLR = TOL
      NFCN = 0_id
      SVNFCN = 0_id
      OKSTP = 0_id
      FLSTP = 0_id
      FIRST = .TRUE.
      LAST = .FALSE.
!C
!  WORK(*) is partioned into a number of arrays using pointers. These
!  pointers are set in /RKCOM3/.
      PRTHRS = 1_id
!                           the threshold values
      PRERST = PRTHRS + NEQ
!                           the error estimates
      PRWT = PRERST + NEQ
!                           the weights used in the local error test
      PRYOLD = PRWT + NEQ
!                           the previous value of the solution
      PRSCR = PRYOLD + NEQ
!                           scratch array used for the higher order
!                           approximate solution and for the previous 
!                           value of the derivative of the solution
      PRY = PRSCR + NEQ
!                           the dependent variables
      PRYP = PRY + NEQ
!                           the derivatives
      PRSTGS = PRYP + NEQ
!                           intermediate stages held in an internal
!                           array STAGES(NEQ,VECSTG)
!C
      FREEPR = PRSTGS + VECSTG*NEQ
!C
!  Allocate storage for interpolation if the TASK = `U' or `u' was
!  specified. INTP and LINTPL returned by CONST indicate whether there
!  is an interpolation scheme associated with the pair and how much
!  storage is required.
!C
      PRINTP = 1_id
      LNINTP = 1_id
      IF (UTASK) THEN
         IF (INTP) THEN
            LNINTP = LINTPL*NEQ
            IF (REQSTG) THEN
               PRINTP = FREEPR
               FREEPR = PRINTP + LNINTP
            ELSE
               PRINTP = PRSTGS
               FREEPR = MAX(PRINTP+VECSTG*NEQ,PRINTP+LNINTP)
            END IF
         END IF
      END IF
!C
!  Initialise state and allocate storage for global error assessment
!  using /RKCOM6/
      GNFCN = 0_id
      MAXERR = ZERO
      LOCMAX = TSTART
      ERASON = ERRASS
      ERASFL = .FALSE.
      IF (ERRASS) THEN
!C
!  Storage is required for the stages of a secondary integration. The
!  stages of the primary intergration can only be overwritten in the
!  cases where there is no interpolant or the interpolant does not
!  require information about the stages (e.g. METHOD 3 and METHOD 1,
!  respectively).
         IF (.NOT.REQSTG) THEN
            PRZSTG = PRSTGS
         ELSE
            PRZSTG = FREEPR
            FREEPR = PRZSTG + VECSTG*NEQ
         END IF
         PRZY = FREEPR
         PRZYP = PRZY + NEQ
         PRZERS = PRZYP + NEQ
         PRZERR = PRZERS + NEQ
         PRZYNU = PRZERR + NEQ
         FREEPR = PRZYNU + NEQ
      ELSE
         PRZSTG = 1_id
         PRZY = 1_id
         PRZYP = 1_id
         PRZERS = 1_id
         PRZERR = 1_id
         PRZYNU = 1_id
      END IF
!C
      LREQ = FREEPR - 1_id
!C
!!  Check for enough workspace and suitable range of integration
!C
      IF (LENWRK.LT.LREQ) THEN
         IER = 911_id
         NREC = 2_id
         WRITE (REC,'(A/A,I6,A,I6,A)')&
     &' ** You have not supplied enough workspace. You gave LENWRK ',&
     &' ** as', LENWRK, ', but it must be at least ',LREQ,'.'
      ELSE
         HMIN = MAX(TINY,TOOSML*MAX(ABS(TSTART),ABS(TEND)))
         IF (ABS(TEND-TSTART).LT.HMIN) THEN
            IER = 911_id
            NREC = 4_id
            WRITE (REC,'(A/A/A,D13.5/A,D13.5,A)')&
     &' ** You have set values for TEND and TSTART that are not ',&
     &' ** clearly distinguishable for the method and the precision ',&
     &' ** of the computer being used. ABS(TEND-TSTART) is ',&
     &ABS(TEND-TSTART),&
     &' ** but should be at least ',HMIN,'.'
         END IF
      END IF
!C
!  Return if error detected
!C
      IF (IER.NE.1) GO TO 80
!C
!  Initialize elements of the workspace
      DO 40 L = 1_id, NEQ
         WORK(PRTHRS-1_id+L) = THRES(L)
         WORK(PRY-1_id+L) = YSTART(L)
   40 CONTINUE
!C
!  Initialize the global error to zero when ERRASS = .TRUE.
      IF (ERRASS) THEN
         DO 60 L = 1_id, NEQ
            WORK(PRZERR-1_id+L) = ZERO
   60    CONTINUE
      END IF
!C
   80 CONTINUE
!C
      CALL RKMSG(IER,SRNAME,NREC,FLAG)
!C
      RETURN
      END SUBROUTINE
     

 SUBROUTINE UT(F,TWANT,TGOT,YGOT,YPGOT,YMAX,WORK,UFLAG)
!C$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!C
!  If you are not familiar with the code UT and how it is used in
!  conjunction with SETUP to solve initial value problems, you should study
!  the document file rksuite.doc carefully before proceeding further.  The
!  following "Brief Reminder" is intended only to remind you of the meaning,
!  type, and size requirements of the arguments.
!C
!  NAME DECLARED IN AN EXTERNAL STATEMENT IN THE CALLING PROGRAM:
!C
!     F         - name of the subroutine for evaluating the differential
!                 equations.
!C
!  The subroutine F must have the form
!C
!  SUBROUTINE F(T,Y,YP)
!  REAL(rd) T,Y(*),YP(*)
!     Given input values of the independent variable T and the solution
!     components Y(*), for each L = 1,2,...,NEQ evaluate the differential
!     equation for the derivative of the Ith solution component and place the
!     value in YP(L).  Do not alter the input values of T and Y(*).
!  RETURN
!  END
!C
!  INPUT VARIABLE
!C
!     TWANT     - REAL(rd)
!                 The next value of the independent variable where a
!                 solution is desired.
!C
!                 Constraints: TWANT must lie between the previous value
!                 of TGOT (TSTART on the first call) and TEND. TWANT can be
!                 equal to TEND, but it must be clearly distinguishable from
!                 the previous value of TGOT (TSTART on the first call) in 
!                 the precision available.
!C
!  OUTPUT VARIABLES
!C
!     TGOT      - REAL(rd)
!                 A solution has been computed at this value of the
!                 independent variable.
!     YGOT(*)   - REAL(rd) array of length NEQ
!                 Approximation to the true solution at TGOT. Do not alter
!                 the contents of this array
!     YPGOT(*)  - REAL(rd) array of length NEQ
!                 Approximation to the first derivative of the true
!                 solution at TGOT.
!     YMAX(*)   - REAL(rd) array of length NEQ
!                 YMAX(L) is the largest magnitude of YGOT(L) computed at any
!                 time in the integration from TSTART to TGOT. Do not alter
!                 the contents of this array.
!C
!  WORKSPACE
!C
!     WORK(*)   - REAL(rd) array as used in SETUP
!                 Do not alter the contents of this array.
!C
!  OUTPUT VARIABLE
!C
!     UFLAG     - INTEGER(id)
!C
!                       SUCCESS.  TGOT = TWANT.
!                 = 1 - Complete success.
!C
!                       "SOFT" FAILURES
!                 = 2 - Warning:  You are using METHOD = 3 inefficiently
!                       by computing answers at many values of TWANT.  If
!                       you really need answers at so many specific points,
!                       it would be more efficient to compute them with
!                       METHOD = 2.  To do this you would need to restart
!                       from TGOT, YGOT(*) by a call to SETUP.  If you wish
!                       to continue as you are, you may.
!                 = 3 - Warning:  A considerable amount of work has been
!                       expended.  If you wish to continue on to TWANT, just
!                       call UT again.
!                 = 4 - Warning:  It appears that this problem is "stiff".
!                       You really should change to another code that is
!                       intended for such problems, but if you insist, you can
!                       continue with UT by calling it again.
!C
!                       "HARD" FAILURES
!                 = 5 - You are asking for too much accuracy. You cannot
!                       continue integrating this problem.
!                 = 6 - The global error assessment may not be reliable beyond
!                       the current point in the integration.  You cannot
!                       continue integrating this problem.
!C
!                       "CATASTROPHIC" FAILURES
!                 = 911 - The nature of the catastrophe is reported on
!                         the standard output channel. Unless special
!                         provision was made in advance (see rksuite.doc),
!                         the computation then comes to a STOP.
!C
!C$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  TGOT, TWANT
      INTEGER(id)           UFLAG
!     .. Array Arguments ..
      REAL(rd)  WORK(*), YGOT(*), YMAX(*), YPGOT(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block for General Workspace Pointers ..
      INTEGER(id)           PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      COMMON /RKCOM3/   PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      SAVE   /RKCOM3/
!     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      CHARACTER*6       SRNAME
      PARAMETER         (SRNAME='UT')
      LOGICAL(lgt)           ASK, TELL
      PARAMETER         (ASK=.TRUE.,TELL=.FALSE.)
      INTEGER(id)           MINUS1, MINUS2
      PARAMETER         (MINUS1=-1_id,MINUS2=-2_id)
      REAL(rd)  ZERO
      PARAMETER         (ZERO=0.0_rd)
!     .. Local Scalars ..
      REAL(rd)  HMIN, TLAST, TNOW, UTEND
      INTEGER(id)           CFLAG, IER, L, NREC, STATE
      LOGICAL(lgt)           BADERR, GOBACK
!     .. External Subroutines ..
 !     EXTERNAL          CHKFL, CT, INTRP, RESET, RKMSG, RKSIT
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, MAX, MIN
!     .. Save statement ..
      SAVE              UTEND, TLAST
!     .. Executable Statements ..
      IER = 1_id
      NREC = 0_id
      GOBACK = .FALSE.
      BADERR = .FALSE.
!C
!  Is it permissible to call UT?
!C
      CALL RKSIT(ASK,'SETUP',STATE)
      IF (STATE.EQ.911) THEN
         IER = 912_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** A catastrophic error has already been detected elsewhere.'
         GO TO 100
      END IF
      IF (STATE.EQ.MINUS1) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** You have not called SETUP, so you cannot use UT.'
         GO TO 100
      END IF
      IF (.NOT.UTASK) THEN
         IER = 911_id
         NREC = 2_id
         WRITE (REC,'(A/A)')&
     &' ** You have called UT after you specified in SETUP that ',&
     &' ** you were going to use CT. This is not permitted.'
         GO TO 100
      END IF
      CALL RKSIT(ASK,SRNAME,STATE)
      IF (STATE.EQ.5 .OR. STATE.EQ.6) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A/A)')&
     &' ** This routine has already returned with a hard failure.',&
     &' ** You must call SETUP to start another problem.'
         GO TO 100
      END IF
      STATE = MINUS2
      CALL RKSIT(TELL,SRNAME,STATE)

      IF (FIRST) THEN

!  First call.

!  A value of TND is specified in SETUP. When INTP = .FALSE., as with
!  METHD = 3, output is obtained at the specified TWANT by resetting TND
!  to TWANT.  At this point, before the integration gets started, this can
!  be done with a simple assignment.  Later it is done with a call to RESET.
!  The original TND is SAVEd as a local variable UTEND.

         UTEND = TND
         IF (.NOT.INTP) TND = TWANT

!  The last TGOT returned is SAVEd in the variable TLAST.  T (a variable
!  passed through the common block RKCOM2) records how far the integration
!  has advanced towards the specified TND.  When output is obtained by
!  interpolation, the integration goes past the TGOT returned (T is closer
!  to the specified TND than TGOT).  Initialize these variables and YMAX(*).
         TLAST = TSTRT
         TGOT = TSTRT
         DO 20 L = 1_id, NEQN
            YMAX(L) = ABS(WORK(PRY-1_id+L))
   20    CONTINUE

!  If the code is to find an on-scale initial step size H, a bound was placed
!  on H in SETUP.  Here the first output point is used to refine this bound.
         IF (HSTRT.EQ.ZERO) THEN
            H = MIN(ABS(H),ABS(TWANT-TSTRT))
            HMIN = MAX(TINY,TOOSML*MAX(ABS(TSTRT),ABS(TND)))
            H = MAX(H,HMIN)
         END IF

      ELSE

!  Subsequent call.

         IF (TLAST.EQ.UTEND) THEN
            IER = 911_id
            NREC = 3_id
            WRITE (REC,'(A/A/A)')&
     &' ** You have called UT after reaching TEND. (Your last    ',&
     &' ** call to UT resulted in TGOT = TEND.)  To start a new ',&
     &' ** problem, you will need to call SETUP.'
            GO TO 100
         END IF

      END IF

!  Check for valid TWANT.

      IF (DIR*(TWANT-TLAST).LE.ZERO) THEN
         IER = 911_id
         NREC = 4_id
         WRITE (REC,'(A/A/A/A)')&
     &' ** You have made a call to UT with a TWANT that does   ',&
     &' ** not lie between the previous value of TGOT (TSTART  ',&
     &' ** on the first call) and TEND. This is not permitted. ',&
     &' ** Check your program carefully.'
         GO TO 100
      END IF
      IF (DIR*(TWANT-UTEND).GT.ZERO) THEN
         HMIN = MAX(TINY,TOOSML*MAX(ABS(TWANT),ABS(UTEND)))
         IF (ABS(TWANT-UTEND).LT.HMIN) THEN
            IER = 911_id
            NREC = 5_id
            WRITE (REC,'(A/A/A/A)')&
     &' ** You have made a call to UT with a TWANT that does      ',&
     &' ** not lie between the previous value of TGOT (TSTART on  ',&
     &' ** the first call) and TEND. This is not permitted. TWANT ',&
     &' ** is very close to TEND, so you may have meant to set    ',&
     &' ** it to be TEND exactly.  Check your program carefully.  '
         ELSE
            IER = 911_id
            NREC = 4_id
            WRITE (REC,'(A/A/A/A)')&
     &' ** You have made a call to UT with a TWANT that does   ',&
     &' ** not lie between the previous value of TGOT (TSTART  ',&
     &' ** on the first call) and TEND. This is not permitted. ',&
     &' ** Check your program carefully.'
         END IF
         GO TO 100
      END IF
      IF (.NOT.INTP) THEN
         HMIN = MAX(TINY,TOOSML*MAX(ABS(TLAST),ABS(TWANT)))
         IF (ABS(TWANT-TLAST).LT.HMIN) THEN
            IER = 911_id
            NREC = 4_id
            WRITE (REC,'(A/A/A/A,D13.5,A)')&
     &' ** You have made a call to UT with a TWANT that is not ',&
     &' ** sufficiently different from the last value of TGOT  ',&
     &' ** (TSTART on the first call).  When using METHOD = 3, ',&
     &' ** it must differ by at least ',HMIN,'.'
            GO TO 100
         END IF

!  We have a valid TWANT. There is no interpolation with this METHD and
!  therefore we step to TWANT exactly by resetting TND with a call to RESET.
!  On the first step this matter is handled differently as explained above.

         IF (.NOT.FIRST) THEN
            CALL RESET(TWANT)
            CALL CHKFL(ASK,BADERR)
            IF (BADERR) GO TO 100
         END IF
      END IF

!  Process output, decide whether to take another step.

   40 CONTINUE

      IF (INTP) THEN

!  Interpolation is possible with this METHD.  The integration has
!  already reached T. If this is past TWANT, GOBACK is set .TRUE. and
!  the answers are obtained by interpolation.

         GOBACK = DIR*(T-TWANT) .GE. ZERO
         IF (GOBACK) THEN
            CALL INTRP(TWANT,'Both solution and derivative',NEQN,YGOT,&
     &                 YPGOT,F,WORK,WORK(PRINTP),LNINTP)
            CALL CHKFL(ASK,BADERR)
            IF (BADERR) GO TO 100
            TGOT = TWANT
         END IF
      ELSE

!  Interpolation is not possible with this METHD, so output is obtained
!  by integrating to TWANT = TND.  Both YGOT(*) and YPGOT(*) are then 
!  already loaded with the solution at TWANT by CT.

         GOBACK = T .EQ. TWANT
         IF (GOBACK) TGOT = TWANT
      END IF

!  Updating of YMAX(*) is done here to account for the fact that when
!  interpolation is done, the integration goes past TGOT.  Note that YGOT(*)
!  is not defined until CT is called.  YMAX(*) was initialized at TSTRT
!  from values stored in WORK(*), so only needs to be updated for T
!  different from TSTRT.
      IF (T.NE.TSTRT) THEN
         DO 60 L = 1, NEQN
            YMAX(L) = MAX(YMAX(L),ABS(YGOT(L)))
   60    CONTINUE
      END IF

!  If done, go to the exit point.
      IF (GOBACK) GO TO 100

!  Take a step with CT in the direction of TND.  On exit, the solution is
!  advanced to TNOW.  The way CT is written, the approximate solution at
!  TNOW is available in both YGOT(*) and in WORK(*).  If output is obtained by
!  stepping to the end (TNOW = TWANT = TND), YGOT(*) can be returned directly.
!  If output is obtained by interpolation, the subroutine INTRP that does this
!  uses the values in WORK(*) for its computations and places the approximate
!  solution at TWANT in the array YGOT(*) for return to the calling program.
!  The approximate derivative is handled in the same way. TNOW is output from
!  CT and is actually a copy of T declared above in a common block.

      CALL CT(F,TNOW,YGOT,YPGOT,WORK,CFLAG)
      IER = CFLAG

!  A successful step by CT is indicated by CFLAG = 1 or = 2.
      IF (CFLAG.EQ.1) THEN
         GO TO 40
      ELSE IF (CFLAG.EQ.2) THEN

!  Supplement the warning message written in CT.
         NREC = 3_id
         WRITE (REC,'(A/A/A)')&
     &' ** The last message was produced on a call to CT from UT.  ',&
     &' ** In UT the appropriate action is to change to METHOD = 2,',&
     &' ** or, if insufficient memory is available, to METHOD = 1. '
      ELSE IF (CFLAG.LE.6) THEN
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** The last message was produced on a call to CT from UT.'
      ELSE
         BADERR = .TRUE.
      END IF
      TGOT = T

!  Update YMAX(*) before the return.
      DO 80 L = 1_id, NEQN
         YMAX(L) = MAX(YMAX(L),ABS(YGOT(L)))
   80 CONTINUE

!  Exit point for UT.

  100 CONTINUE

      IF (BADERR) THEN
         IER = 911_id
         NREC = 4_id
         WRITE (REC,'(A/A/A/A)')&
     &' ** An internal call by UT to a subroutine resulted in an  ',&
     &' ** error that should not happen.  Check your program      ',&
     &' ** carefully for array sizes, correct number of arguments,',&
     &' ** type mismatches ... .'
      END IF

      TLAST = TGOT

!  All exits are done here after a call to RKMSG to report
!  what happened and set UFLAG.

      CALL RKMSG(IER,SRNAME,NREC,UFLAG)

      RETURN
    END  SUBROUTINE
     

SUBROUTINE STAT(TOTFCN,STPCST,WASTE,STPSOK,HNEXT)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

!  If you are not familiar with the code STAT and how it is used in
!  conjunction with the integrators CT and UT, you should study the 
!  document file rksuite.doc carefully before attempting to use the code. 
!  The following "Brief Reminder" is intended only to remind you of the 
!  meaning, type, and size requirements of the arguments.
!
!  STAT is called to obtain some details about the integration.
!
!  OUTPUT VARIABLES
!
!     TOTFCN    - INTEGER(id)
!                 Total number of calls to F in the integration so far --
!                 a measure of the cost of the integration.
!     STPCST    - INTEGER(id)
!                 Cost of a typical step with this METHOD measured in
!                 calls to F.
!     WASTE     - REAL(rd)
!                 The number of attempted steps that failed to meet the 
!                 local error requirement divided by the total number of 
!                 stattempted so far in the integration.
!     STPSOK    - INTEGER(id)
!                 The number of accepted steps.
!     HNEXT     - REAL(rd)
!                 The step size the integrator plans to use for the next step.
!
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  HNEXT, WASTE
      INTEGER(id)           STPCST, STPSOK, TOTFCN
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Global Error Assessment ..
      REAL(rd)  MAXERR, LOCMAX
      INTEGER(id)           GNFCN, PRZSTG, PRZY, PRZYP, PRZERS, PRZERR,&
     &                  PRZYNU
      LOGICAL(lgt)           ERASON, ERASFL
      COMMON /RKCOM6/   MAXERR, LOCMAX, GNFCN, PRZSTG, PRZY, PRZYP,&
     &                  PRZERS, PRZERR, PRZYNU, ERASON, ERASFL
      SAVE   /RKCOM6/
!     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      CHARACTER*6       SRNAME
      PARAMETER         (SRNAME='STAT')
      LOGICAL(lgt)           ASK
      INTEGER(id)           MINUS1, MINUS2
      PARAMETER         (ASK=.TRUE.,MINUS1=-1_id,MINUS2=-2_id)
      REAL(rd)  ZERO
      PARAMETER         (ZERO=0.0_rd)
!     .. Local Scalars ..
      INTEGER(id)           FLAG, IER, NREC, STATE
!     .. External Subroutines ..
 !     EXTERNAL          RKMSG, RKSIT
!     .. Intrinsic Functions ..
      INTRINSIC         DBLE
!     .. Executable Statements ..
!
      IER = 1_id
      NREC = 0_id
!
!  Is it permissible to call STAT?
!
      CALL RKSIT(ASK,SRNAME,STATE)
      IF (STATE.EQ.911) THEN
         IER = 912_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** A catastrophic error has already been detected elsewhere.'
         GO TO 20
      END IF
      IF (STATE.EQ.MINUS2) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A/A/A)')&
     &' ** You have already made a call to STAT after a hard   ',&
     &' ** failure was reported from the integrator. You cannot',&
     &' ** call STAT again.'
         GO TO 20
      END IF
      CALL RKSIT(ASK,'CT',STATE)
      IF (STATE.EQ.MINUS1) THEN
         IER = 911_id
         NREC = 1_id
         IF (UTASK) THEN
            WRITE (REC,'(A)')&
     &' ** You have not called UT, so you cannot use STAT.'
         ELSE
            WRITE (REC,'(A)')&
     &' ** You have not called CT, so you cannot use STAT.'
         END IF
         GO TO 20
      END IF
!
!  Set flag so that the routine can only be called once after a hard 
!  failure from the integrator.
      IF (STATE.EQ.5_id .OR. STATE.EQ.6_id) IER = MINUS2
!
      TOTFCN = SVNFCN + NFCN
      IF (ERASON) TOTFCN = TOTFCN + GNFCN
      STPCST = COST
      STPSOK = OKSTP
      IF (OKSTP.LE.1_id) THEN
         WASTE = ZERO
      ELSE
         WASTE = DBLE(FLSTP)/DBLE(FLSTP+OKSTP)
      END IF
      HNEXT = H
!
   20 CONTINUE
!
      CALL RKMSG(IER,SRNAME,NREC,FLAG)
!
      RETURN
    END  SUBROUTINE STAT
       

SUBROUTINE GLBERR(RMSERR,ERRMAX,TERRMX,WORK)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!
!  If you are not familiar with the code GLBERR and how it is used in
!  conjunction with UT and CT to solve initial value problems, you should
!  study the document file rksuite.doc carefully before attempting to use 
!  the code.  The following "Brief Reminder" is intended only to remind you 
!  of the meaning, type, and size requirements of the arguments.
!
!  If ERRASS was set .TRUE. in the call to SETUP, then after any call to UT
!  or CT to advance the integration to TNOW or TWANT, the subroutine GLBERR
!  may be called to obtain an assessment of the true error of the integration.
!  At each step and for each solution component Y(L), a more accurate "true"
!  solution YT(L), an average magnitude "size(L)" of its size, and its error
!                abs(Y(L) - YT(L))/max("size(L)",THRES(L))
!  are computed.  The assessment returned in RMSERR(L) is the RMS (root-mean-
!  square) average of the error in the Lth solution component over all steps
!  of the integration from TSTART through TNOW.
!
!  OUTPUT VARIABLES
!
!     RMSERR(*) - REAL(rd) array of length NEQ
!                 RMSERR(L) approximates the RMS average of the true error 
!                 of the numerical solution for the Ith solution component,
!                 L = 1,2,...,NEQ.  The average is taken over all steps from
!                 TSTART to TNOW.
!     ERRMAX    - REAL(rd)
!                 The maximum (approximate) true error taken over all
!                 solution components and all steps from TSTART to TNOW.
!     TERRMX    - REAL(rd)
!                 First value of the independent variable where the
!                 (approximate) true error attains the maximum value ERRMAX.
!
!  WORKSPACE
!
!     WORK(*)   - REAL(rd) array as used in SETUP and UT or CT
!                 Do not alter the contents of this array.
!
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  ERRMAX, TERRMX
!     .. Array Arguments ..
      REAL(rd)  RMSERR(*), WORK(*)
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block for Global Error Assessment ..
      REAL(rd)  MAXERR, LOCMAX
      INTEGER(id)           GNFCN, PRZSTG, PRZY, PRZYP, PRZERS, PRZERR,&
     &                  PRZYNU
      LOGICAL(lgt)           ERASON, ERASFL
      COMMON /RKCOM6/   MAXERR, LOCMAX, GNFCN, PRZSTG, PRZY, PRZYP,&
     &                  PRZERS, PRZERR, PRZYNU, ERASON, ERASFL
      SAVE   /RKCOM6/
!     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      CHARACTER*6       SRNAME
      PARAMETER         (SRNAME='GLBERR')
      LOGICAL(lgt)           ASK
      PARAMETER         (ASK=.TRUE.)
      INTEGER(id)           MINUS1, MINUS2
      PARAMETER         (MINUS1=-1_id,MINUS2=-2_id)
!     .. Local Scalars ..
      INTEGER(id)           FLAG, IER, L, NREC, STATE
!     .. External Subroutines ..
 !     EXTERNAL          RKMSG, RKSIT
!     .. Intrinsic Functions ..
      INTRINSIC         SQRT
!     .. Executable Statements ..
!
      IER = 1_id
      NREC = 0_id
!
!  Is it permissible to call GLBERR?
!
      CALL RKSIT(ASK,SRNAME,STATE)
      IF (STATE.EQ.911_id) THEN
         IER = 912_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** A catastrophic error has already been detected elsewhere.'
         GO TO 40
      END IF
      IF (STATE.EQ.MINUS2) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A/A/A)')&
     &' ** You have already made a call to GLBERR after a hard ',&
     &' ** failure was reported from the integrator. You cannot',&
     &' ** call GLBERR again.'
         GO TO 40
      END IF
      CALL RKSIT(ASK,'CT',STATE)
      IF (STATE.EQ.MINUS1) THEN
         IER = 911_id
         NREC = 1_id
         IF (UTASK) THEN
            WRITE (REC,'(A)')&
     &' ** You have not yet called UT, so you cannot call GLBERR.'
         ELSE
            WRITE (REC,'(A)')&
     &' ** You have not yet called CT, so you cannot call GLBERR.'
         END IF
         GO TO 40
      END IF
!
!  Set flag so that the routine can only be called once after a hard 
!  failure from the integrator.
      IF (STATE.EQ.5_id .OR. STATE.EQ.6_id) IER = MINUS2
!
!  Check that ERRASS was set properly for error assessment in SETUP.
!
      IF (.NOT.ERASON) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A/A/A)')&
     &' ** No error assessment is available since you did not ',&
     &' ** ask for it in your call to the routine SETUP.',&
     &' ** Check your program carefully.'
         GO TO 40
      END IF
!
!  Check to see if the integrator has not actually taken a step.
!
      IF (OKSTP.EQ.0) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A/A/A)')&
     &' ** The integrator has not actually taken any successful ',&
     &' ** steps.  You cannot call GLBERR in this circumstance. ',&
     &' ** Check your program carefully.'
         GO TO 40
      END IF
!
!  Compute RMS error and set output variables.
!
      ERRMAX = MAXERR
      TERRMX = LOCMAX
      DO 20 L = 1_id, NEQN
         RMSERR(L) = SQRT(WORK(PRZERR-1_id+L)/OKSTP)
   20 CONTINUE
!
   40 CONTINUE
!
      CALL RKMSG(IER,SRNAME,NREC,FLAG)
!
      RETURN
    END  SUBROUTINE


 SUBROUTINE CT(F,TNOW,YNOW,YPNOW,WORK,CFLAG)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!
!  If you are not familiar with the code CT and how it is used in
!  conjunction with SETUP to solve initial value problems, you should study
!  the document file rksuite.doc carefully before attempting to use the code.
!  The following "Brief Reminder" is intended only to remind you of the
!  meaning, type, and size requirements of the arguments.
!
!  NAME DECLARED IN AN EXTERNAL STATEMENT IN THE CALLING PROGRAM:
!
!     F         - name of the subroutine for evaluating the differential
!                 equations.
!
!  The subroutine F must have the form
!
!  SUBROUTINE F(T,Y,YP)
!  REAL(rd) T,Y(*),YP(*)
!     Using the input values of the independent variable T and the solution
!     components Y(*), for each L = 1,2,...,NEQ evaluate the differential
!     equation for the derivative of the Lth solution component and place the
!     value in YP(L).  Do not alter the input values of T and Y(*).
!  RETURN
!  END
!
!  OUTPUT VARIABLES
!
!     TNOW      - REAL(rd)
!                 Current value of the independent variable.
!     YNOW(*)   - REAL(rd) array of length NEQ
!                 Approximation to the true solution at TNOW.
!     YPNOW(*)  - REAL(rd) array of length NEQ
!                 Approximation to the first derivative of the
!                 true solution at TNOW.
!
!  WORKSPACE
!
!     WORK(*)   - REAL(rd) array as used in SETUP
!                 Do not alter the contents of this array.
!
!  OUTPUT VARIABLE
!
!     CFLAG     - INTEGER(id)
!
!                       SUCCESS.  A STEP WAS TAKEN TO TNOW.
!                 = 1 - Complete success.
!
!                       "SOFT" FAILURES
!                 = 2 - Warning:  You have obtained an answer by integrating
!                       to TEND (TNOW = TEND).  You have done this at least
!                       100 times, and monitoring of the computation reveals
!                       that this way of getting output has degraded the
!                       efficiency of the code. If you really need answers at
!                       so many specific points, it would be more efficient to
!                       get them with INTRP.  (If METHOD = 3, you would need
!                       to change METHOD and restart from TNOW, YNOW(*) by a
!                       call to SETUP.)  If you wish to continue as you are,
!                       you may.
!                 = 3 - Warning:  A considerable amount of work has been
!                       expended. To continue the integration, just call
!                       CT again.
!                 = 4 - Warning:  It appears that this problem is "stiff".
!                       You really should change to another code that is
!                       intended for such problems, but if you insist, you 
!                       can continue with CT by calling it again.
!
!                       "HARD" FAILURES
!                 = 5 - You are asking for too much accuracy. You cannot
!                       continue integrating this problem.
!                 = 6 - The global error assessment may not be reliable beyond
!                       the current point in the integration.  You cannot
!                       continue integrating this problem.
!
!                       "CATASTROPHIC" FAILURES
!                 = 911 - The nature of the catastrophe is reported on
!                         the standard output channel. Unless special
!                         provision was made in advance (see rksuite.doc),
!                         the computation then comes to a STOP.
!
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
use kindtype
implicit none

!     .. Scalar Arguments ..
      REAL(rd)  TNOW
      INTEGER(id)           CFLAG
!     .. Array Arguments ..
      REAL(rd)  WORK(*), YNOW(*), YPNOW(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block for General Workspace Pointers ..
      INTEGER(id)           PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      COMMON /RKCOM3/   PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      SAVE   /RKCOM3/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Global Error Assessment ..
      REAL(rd)  MAXERR, LOCMAX
      INTEGER(id)           GNFCN, PRZSTG, PRZY, PRZYP, PRZERS, PRZERR,&
     &                  PRZYNU
      LOGICAL(lgt)           ERASON, ERASFL
      COMMON /RKCOM6/   MAXERR, LOCMAX, GNFCN, PRZSTG, PRZY, PRZYP,&
     &                  PRZERS, PRZERR, PRZYNU, ERASON, ERASFL
      SAVE   /RKCOM6/
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      CHARACTER*6       SRNAME
      PARAMETER         (SRNAME='CT')
      LOGICAL(lgt)           ASK, TELL
      PARAMETER         (ASK=.TRUE.,TELL=.FALSE.)
      INTEGER(id)           MINUS1, MINUS2
      PARAMETER         (MINUS1=-1_id,MINUS2=-2_id)
      INTEGER(id)           MAXFCN
      PARAMETER         (MAXFCN=100000000_id) !@new
      REAL(rd)  ZERO, PT1, PT9, ONE, TWO, HUNDRD
      PARAMETER         (ZERO=0.0_rd,PT1=0.1_rd,PT9=0.9_rd,ONE=1.0_rd,&
     &                  TWO=2.0_rd,HUNDRD=100.0_rd)
!     .. Local Scalars ..
      REAL(rd)  ALPHA, BETA, ERR, ERROLD, HAVG, HMIN, HTRY, TAU,&
     &                  TEMP1, TEMP2, YPNORM
      INTEGER(id)           IER, JFLSTP, L, NREC, NTEND, POINT, STATE, YNEW,&
     &                  YPOLD
      LOGICAL(lgt)           CHKEFF, FAILED, MAIN, PHASE1, PHASE2, PHASE3,&
     &                  TOOMCH
!     .. External Subroutines ..
 !     EXTERNAL          RKMSG, RKSIT, STEP, STIFF, TRUERR
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, MAX, MIN, SIGN
!     .. Save statement ..
      SAVE              JFLSTP, NTEND, ERROLD, HAVG, PHASE2, YNEW,&
     &                  YPOLD, CHKEFF
!     .. Executable Statements ..
!
      IER = 1_id
      NREC = 0_id
!
!  Is it permissible to call CT?
!
      CALL RKSIT(ASK,'SETUP',STATE)
      IF (STATE.EQ.911_id) THEN
         IER = 912_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** A catastrophic error has already been detected elsewhere.'
         GO TO 180
      END IF
      IF (STATE.EQ.MINUS1) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** You have not called SETUP, so you cannot use CT.'
         GO TO 180
      END IF
      IF (UTASK) THEN
         CALL RKSIT(ASK,'UT',STATE)
         IF (STATE.NE.MINUS2) THEN
            IER = 911_id
            NREC = 2_id
            WRITE (REC,'(A/A)')&
     &' ** You have called CT after you specified in SETUP that ',&
     &' ** you were going to use UT. This is not permitted.'
            UTASK = .FALSE.
            GO TO 180
         END IF
      END IF
      CALL RKSIT(ASK,SRNAME,STATE)
      IF (STATE.EQ.5_id .OR. STATE.EQ.6_id) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A/A)')&
     &' ** CT has already returned with a flag value of 5 or 6.',&
     &' ** You cannot continue integrating this problem. You must ',&
     &' ** call SETUP to start another problem.'
         GO TO 180
      END IF
!
      IF (FIRST) THEN
!
!  First call in an integration -- initialize everything.
!
         CHKEFF = .FALSE.
         NTEND = 0_id
         JFLSTP = 0_id
!
!  A scratch area of WORK(*) starting at PRSCR is used to hold two
!  arrays in this subroutine: the higher order approximate solution at
!  the end of a step and the approximate derivative of the solution at
!  the end of the last step. To make this clear, local pointers YNEW and
!  YPOLD are used.
         YNEW = PRSCR
         YPOLD = PRSCR
!
!  For this first step T was initialized to TSTRT in SETUP and the
!  starting values YSTART(*) were loaded into the area of WORK(*) reserved
!  for the current solution approximation starting at location PRY. The
!  derivative is now computed and stored in WORK(*) starting at PRYP.
!  Subsequently these arrays are copied to the output vectors YNOW(*)
!  and YPNOW(*).
         CALL F(T,WORK(PRY),WORK(PRYP))
         NFCN = NFCN + 1_id
         DO 20 L = 1_id, NEQN
            YNOW(L) = WORK(PRY-1_id+L)
            YPNOW(L) = WORK(PRYP-1_id+L)
   20    CONTINUE
!
!  Set dependent variables for error assessment.
         IF (ERASON) THEN
            DO 40 L = 1_id, NEQN
               WORK(PRZY-1_id+L) = YNOW(L)
               WORK(PRZYP-1_id+L) = YPNOW(L)
   40       CONTINUE
         END IF
!
!  The weights for the control of the error depend on the size of the
!  solution at the beginning and at the end of the step. On the first
!  step we do not have all this information. Whilst determining the
!  initial step size we initialize the weight vector to the larger of
!  abs(Y(L)) and the threshold for this component.
         DO 60 L = 1_id, NEQN
            WORK(PRWT-1_id+L) = MAX(ABS(YNOW(L)),WORK(PRTHRS-1_id+L))
   60    CONTINUE
!
!  If HSTRT is equal to zero, the code is to find an on-scale initial step
!  size H.  CT has an elaborate scheme of three phases for finding such an H,
!  and some preparations are made earlier.  In SETUP an upper bound is placed
!  on H that reflects the scale of the independent variable. When UTASK is
!  .TRUE., UT refines this bound using the first output point.  Here in CT
!  PHASE1 applies a rule of thumb based on the error control, the order of the
!  the formula, and the size of the initial slope to get a crude approximation
!  to an on-scale H.  PHASE2 may reduce H in the course of taking the first
!  step.  PHASE3 repeatedly adjusts H and retakes the first step until H is
!  on scale.
!
!  A guess for the magnitude of the first step size H can be provided to SETUP
!  as HSTART.  If it is too big or too small, it is ignored and the automatic
!  determination of an on-scale initial step size is activated.  If it is
!  acceptable, H is set to HSTART in SETUP.  Even when H is supplied to CT,
!  PHASE3 of the scheme for finding an on-scale initial step size is made
!  active so that the code can deal with a bad guess.
!
         PHASE1 = HSTRT .EQ. ZERO
         PHASE2 = PHASE1
         PHASE3 = .TRUE.
         IF (PHASE1) THEN
            H = ABS(H)
            YPNORM = ZERO
            DO 80 L = 1_id, NEQN
               IF (ABS(YNOW(L)).NE.ZERO) THEN
                  YPNORM = MAX(YPNORM,ABS(YPNOW(L))/WORK(PRWT-1_id+L))
               END IF
   80       CONTINUE
            TAU = TOLR**EXPON
            IF (H*YPNORM.GT.TAU) H = TAU/YPNORM
            HMIN = MAX(TINY,TOOSML*MAX(ABS(TSTRT),ABS(TND)))
            H = DIR*MAX(H,HMIN)
            PHASE1 = .FALSE.
         END IF
!
      ELSE
!
! Continuation call
!
         IF (LAST) THEN
            IER = 911_id
            NREC = 3_id
            WRITE (REC,'(A,D13.5,A/A/A)')&
     &' ** You have already reached TEND ( = ', TND, ').',&
     &' ** To integrate further with the same problem you must ',&
     &' ** call the routine RESET with a new value of TEND.'
            GO TO 180
         END IF
      END IF
!
!  Begin computation of a step here.
!
      FAILED = .FALSE.
!
  100 CONTINUE
      H = SIGN(ABS(H),DIR)
!
!  Reduce the step size if necessary so that the code will not step
!  past TND.  "Look ahead" to prevent unnecessarily small step sizes.
      LAST = DIR*((T+H)-TND) .GE. ZERO
      IF (LAST) THEN
         H = TND - T
      ELSE IF (DIR*((T+TWO*H)-TND).GE.ZERO) THEN
         H = (TND-T)/TWO
      END IF
!
!  When the integrator is at T and attempts a step of H, the function
!  defining the differential equations will be evaluated at a number of
!  arguments between T and T+H.  If H is too small, these arguments cannot
!  be clearly distinguished in the precision available.
!
      HMIN = MAX(TINY,TOOSML*MAX(ABS(T),ABS(T+H)))
      IF (ABS(H).LT.HMIN) THEN
         IER = 5_id
         NREC = 3_id
         WRITE (REC,'(A/A,D13.5,A,D13.5/A)')&
     &' ** In order to satisfy your error requirements CT would ',&
     &' ** have to use a step size of ',H,' at TNOW = ',T,&
     &' ** This is too small for the machine precision.'
         GO TO 180
      END IF
!
!  Monitor the impact of output on the efficiency of the integration.
!
      IF (CHKEFF) THEN
         NTEND = NTEND + 1_id
         IF (NTEND.GE.100_id .AND. NTEND.GE.OKSTP/3_id) THEN
            IER = 2_id
            NREC = 6_id
            WRITE (REC,'(A/A/A/A/A/A)')&
     &' ** More than 100 output points have been obtained by ',&
     &' ** integrating to TEND.  They have been sufficiently close ',&
     &' ** to one another that the efficiency of the integration has ',&
     &' ** been degraded. It would probably be (much) more efficient ',&
     &' ** to obtain output by interpolating with INTRP (after ',&
     &' ** changing to METHOD=2 if you are using METHOD = 3).'
            NTEND = 0_id
            GO TO 180
         END IF
      END IF
!
!  Check for stiffness and for too much work.  Stiffness can be
!  checked only after a successful step.
!
      IF (.NOT.FAILED) THEN
!
!  Check for too much work.
         TOOMCH = NFCN .GT. MAXFCN
         IF (TOOMCH) THEN
            IER = 3_id
            NREC = 3_id
           WRITE (REC,'(A,I9,A/A/A)')&
     &' ** Approximately ',MAXFCN,' function evaluations have been ',&
     &' ** used to compute the solution since the integration ',&
     &' ** started or since this message was last printed.'
!
!  After this warning message, NFCN is reset to permit the integration
!  to continue.  The total number of function evaluations in the primary
!  integration is SVNFCN + NFCN.
            SVNFCN = SVNFCN + NFCN
            NFCN = 0_id
         END IF
!
!  Check for stiffness.  NREC is passed on to STIFF because when
!  TOOMCH = .TRUE. and stiffness is diagnosed, the message about too
!  much work is augmented inside STIFF to explain that it is due to
!  stiffness.
         CALL STIFF(F,HAVG,JFLSTP,TOOMCH,MAXFCN,WORK,IER,NREC)
!
         IF (IER.NE.1_id) GO TO 180
      END IF
!
!  Take a step.  Whilst finding an on-scale H (PHASE2 = .TRUE.), the input
!  value of H might be reduced (repeatedly), but it will not be reduced
!  below HMIN.  The local error is estimated, a weight vector is formed,
!  and a weighted maximum norm, ERR, of the local error is returned.
!  The variable MAIN is input as .TRUE. to tell STEP that this is the
!  primary, or "main", integration.
!
!  H resides in the common block /RKCOM2/ which is used by both CT and STEP;
!  since it may be changed inside STEP, a local copy is made to ensure
!  portability of the code.
!
      MAIN = .TRUE.
      HTRY = H
      CALL STEP(F,NEQN,T,WORK(PRY),WORK(PRYP),WORK(PRSTGS),TOLR,HTRY,&
     &          WORK(PRWT),WORK(YNEW),WORK(PRERST),ERR,MAIN,HMIN,&
     &          WORK(PRTHRS),PHASE2)
      H = HTRY
!
!  Compare the norm of the local error to the tolerance.
!
      IF (ERR.GT.TOLR) THEN
!
!  Failed step.  Reduce the step size and try again.
!
!  First step:  Terminate PHASE3 of the search for an on-scale step size.
!               The step size is not on scale, so ERR may not be accurate;
!               reduce H by a fixed factor.  Failed attempts to take the
!               first step are not counted.
!  Later step:  Use ERR to compute an "optimal" reduction of H.  More than
!               one failure indicates a difficulty with the problem and an
!               ERR that may not be accurate, so reduce H by a fixed factor.
!
         IF (FIRST) THEN
            PHASE3 = .FALSE.
            ALPHA = RS1
         ELSE
            FLSTP = FLSTP + 1_id
            JFLSTP = JFLSTP + 1_id
            IF (FAILED) THEN
               ALPHA = RS1
            ELSE
               ALPHA = SAFETY*(TOLR/ERR)**EXPON
               ALPHA = MAX(ALPHA,RS1)
            END IF
         END IF
         H = ALPHA*H
         FAILED = .TRUE.
         GO TO 100
      END IF
!
!  Successful step.
!
!  Predict a step size appropriate for the next step.  After the first
!  step the prediction can be refined using an idea of H.A. Watts that
!  takes account of how well the prediction worked on the previous step.
      BETA = (ERR/TOLR)**EXPON
      IF (.NOT.FIRST) THEN
         TEMP1 = (ERR**EXPON)/H
         TEMP2 = (ERROLD**EXPON)/HOLD
         IF (TEMP1.LT.TEMP2*HUNDRD .AND. TEMP2.LT.TEMP1*HUNDRD) THEN
            BETA = BETA*(TEMP1/TEMP2)
         END IF
      END IF
      ALPHA = RS3
      IF (SAFETY.LT.BETA*ALPHA) ALPHA = SAFETY/BETA
!
!  On the first step a search is made for an on-scale step size.  PHASE2
!  of the scheme comes to an end here because a step size has been found
!  that is both successful and has a credible local error estimate. Except
!  in the special case that the first step is also the last, the step is
!  repeated in PHASE3 as long as an increase greater than RS2 appears
!  possible.  An increase as big as RS3 is permitted.  A step failure
!  terminates PHASE3.
!
      IF (FIRST) THEN
         PHASE2 = .FALSE.
         PHASE3 = PHASE3 .AND. .NOT. LAST .AND. (ALPHA.GT.RS2)
         IF (PHASE3) THEN
            H = ALPHA*H
            GO TO 100
         END IF
      END IF
!
!  After getting on scale, step size changes are more restricted.
      ALPHA = MIN(ALPHA,RS)
      IF (FAILED) ALPHA = MIN(ALPHA,ONE)
      ALPHA = MAX(ALPHA,RS1)
      HOLD = H
      H = ALPHA*H
!
!  For the diagnosis of stiffness, an average accepted step size, HAVG,
!  must be computed and SAVEd.
      IF (FIRST) THEN
         HAVG = HOLD
      ELSE
         HAVG = PT9*HAVG + PT1*HOLD
      END IF
!
      FIRST = .FALSE.
      ERROLD = ERR
      TOLD = T
!
!  Take care that T is set to precisely TND when the end of the
!  integration is reached.
      IF (LAST) THEN
         T = TND
      ELSE
         T = T + HOLD
      END IF
!
!  Increment counter on accepted steps.  Note that successful steps
!  that are repeated whilst getting on scale are not counted.
      OKSTP = OKSTP + 1_id
!
!  Advance the current solution and its derivative.  (Stored in WORK(*)
!  with the first location being PRY and PRYP, respectively.)  Update the
!  previous solution and its derivative.  (Stored in WORK(*) with the first
!  location being PRYOLD and YPOLD, respectively.)  Note that the previous
!  derivative will overwrite YNEW(*).
!
      DO 120 L = 1_id, NEQN
         WORK(PRYOLD-1_id+L) = WORK(PRY-1_id+L)
         WORK(PRY-1_id+L) = WORK(YNEW-1_id+L)
         WORK(YPOLD-1_id+L) = WORK(PRYP-1_id+L)
  120 CONTINUE
!
      IF (FSAL) THEN
!
!  When FSAL = .TRUE., YP(*) is the last stage of the step.
         POINT = PRSTGS + (LSTSTG-1_id)*NEQN
         DO 140 L = 1_id, NEQN
            WORK(PRYP-1_id+L) = WORK(POINT-1_id+L)
  140    CONTINUE
      ELSE
!
!  Call F to evaluate YP(*).
         CALL F(T,WORK(PRY),WORK(PRYP))
         NFCN = NFCN + 1_id
      END IF
!
!  If global error assessment is desired, advance the secondary
!  integration from TOLD to T.
!
      IF (ERASON) THEN
         CALL TRUERR(F,NEQN,WORK(PRY),TOLR,WORK(PRWT),WORK(PRZY),&
     &               WORK(PRZYP),WORK(PRZERR),WORK(PRZYNU),WORK(PRZERS),&
     &               WORK(PRZSTG),IER)
         IF (IER.EQ.6_id) THEN
!
!  The global error estimating procedure has broken down. Treat it as a
!  failed step. The solution and derivative are reset to their values at
!  the beginning of the step since the last valid error assessment refers
!  to them.
            OKSTP = OKSTP - 1_id
            ERASFL = .TRUE.
            LAST = .FALSE.
            T = TOLD
            H = HOLD
            DO 160 L = 1_id, NEQN
               WORK(PRY-1_id+L) = WORK(PRYOLD-1_id+L)
               WORK(PRYP-1_id+L) = WORK(YPOLD-1_id+L)
  160       CONTINUE
            IF (OKSTP.GT.1_id) THEN
               NREC = 2_id
               WRITE (REC,'(A/A,D13.5,A)')&
     &' ** The global error assessment may not be reliable for T past ',&
     &' ** TNOW = ',T,'.  The integration is being terminated.'
            ELSE
               NREC = 2_id
               WRITE (REC,'(A/A)')&
     &' ** The global error assessment algorithm failed at the start',&
     &' ** the integration.  The integration is being terminated.'
            END IF
            GO TO 180
         END IF
      END IF
!
!
!  Exit point for CT
!
  180 CONTINUE
!
!  Set the output variables and flag that interpolation is permitted
!
      IF (IER.LT.911_id) THEN
         TNOW = T
         LAST = TNOW .EQ. TND
         CHKEFF = LAST
         DO 200 L = 1_id, NEQN
            YNOW(L) = WORK(PRY-1_id+L)
            YPNOW(L) = WORK(PRYP-1_id+L)
  200    CONTINUE
         IF (IER.EQ.1_id) THEN
            STATE = MINUS2
            CALL RKSIT(TELL,'INTRP',STATE)
         END IF
      END IF
!
!  Call RKMSG to report what happened and set CFLAG.
!
      CALL RKMSG(IER,SRNAME,NREC,CFLAG)
!
      RETURN
      END  SUBROUTINE


SUBROUTINE INTRP(TWANT,REQEST,NWANT,YWANT,YPWANT,F,WORK,WRKINT,&
     &                 LENINT)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!
!  If you are not familiar with the code INTRP and how it is used in
!  conjunction with CT to solve initial value problems, you should study the
!  document file rksuite.doc carefully before attempting to use the code. The
!  following "Brief Reminder" is intended only to remind you of the meaning,
!  type, and size requirements of the arguments.
!
!  When integrating with METHOD = 1 or 2, answers may be obtained inexpensively
!  between steps by interpolation. INTRP is called after a step by CT from a
!  previous value of T, called TOLD below, to the current value of T to get
!  an answer at TWANT. You can specify any value of TWANT you wish, but
!  specifying a value outside the interval [TOLD,T] is likely to yield
!  answers with unsatisfactory accuracy.
!
!  INPUT VARIABLE
!
!     TWANT     - REAL(rd)
!                 The value of the independent variable where a solution
!                 is desired.
!
!  The interpolant is to be evaluated at TWANT to approximate the solution
!  and/or its first derivative there.  There are three cases:
!
!  INPUT VARIABLE
!
!     REQEST    - CHARACTER*(*)
!                 Only the first character of REQEST is significant.
!                 REQEST(1:1) = `S' or `s'- compute approximate `S'olution 
!                                           only.
!                             = `D' or `d'- compute approximate first 
!                                           `D'erivative of the solution only.
!                             = `B' or `b'- compute `B'oth approximate solution
!                                           and first derivative.
!                 Constraint: REQEST(1:1) must be `S',`s',`D',`d',`B' or `b'.
!
!  If you intend to interpolate at many points, you should arrange for the
!  the interesting components to be the first ones because the code
!  approximates only the first NWANT components.
!
!  INPUT VARIABLE
!
!     NWANT     - INTEGER(id)
!                 Only the first NWANT components of the answer are to be
!                 computed.
!                 Constraint:  NEQ >= NWANT >= 1
!
!  OUTPUT VARIABLES
!
!     YWANT(*)  - REAL(rd) array of length NWANT
!                 Approximation to the first NWANT components of the true
!                 solution at TWANT when REQESTed.
!     YPWANT(*) - REAL(rd) array of length NWANT
!                 Approximation to the first NWANT components of the first
!                 derivative of the true solution at TWANT when REQESTed.
!
!  NAME DECLARED IN AN EXTERNAL STATEMENT IN THE PROGRAM CALLING INTRP:
!
!     F         - name of the subroutine for evaluating the differential
!                 equations as provided to CT.
!
!  WORKSPACE
!
!     WORK(*)   - REAL(rd) array as used in SETUP and CT
!                 Do not alter the contents of this array.
!
!     WRKINT(*) - REAL(rd) array of length LENINT
!                 Do not alter the contents of this array.
!
!     LENINT    - INTEGER(id)
!                 Length of WRKINT. If
!                 METHOD = 1, LENINT must be at least 1
!                        = 2, LENINT must be at least NEQ+MAX(NEQ,5*NWANT)
!                        = 3--CANNOT BE USED WITH THIS SUBROUTINE
!
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  TWANT
      INTEGER(id)           LENINT, NWANT
      CHARACTER*(*)     REQEST
!     .. Array Arguments ..
      REAL(rd)  WORK(*), WRKINT(*), YPWANT(*), YWANT(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block for General Workspace Pointers ..
      INTEGER(id)           PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      COMMON /RKCOM3/   PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      SAVE   /RKCOM3/
!     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      CHARACTER*6       SRNAME
      PARAMETER         (SRNAME='INTRP')
      LOGICAL(lgt)           ASK
      INTEGER(id)           PLUS1, MINUS1, MINUS2
      PARAMETER         (ASK=.TRUE.,PLUS1=1_id,MINUS1=-1_id,MINUS2=-2_id)
!     .. Local Scalars ..
      INTEGER(id)           FLAG, ICHK, IER, NREC, NWNTSV, STARTP, STATE,&
     &                  STATE1
      LOGICAL(lgt)           ININTP, LEGALR
      CHARACTER         REQST1
!     .. External Subroutines ..
 !     EXTERNAL          EVALI, FORMI, RKMSG, RKSIT
!     .. Intrinsic Functions ..
      INTRINSIC         MAX
!     .. Save statement ..
      SAVE              NWNTSV, ININTP, STARTP
!     .. Data statements ..
      DATA              NWNTSV/MINUS1/
!     .. Executable Statements ..
!
      IER = 1_id
      NREC = 0_id
!
!  Is it permissible to call INTRP?
!
      CALL RKSIT(ASK,'CT',STATE)
      IF (STATE.EQ.911_id) THEN
         IER = 912_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** A catastrophic error has already been detected elsewhere.'
         GO TO 20
      END IF
      IF (UTASK) THEN
         CALL RKSIT(ASK,'UT',STATE1)
         IF (STATE1.NE.MINUS2) THEN
            IER = 911_id
            NREC = 2_id
            WRITE (REC,'(A/A)')&
     &' ** You have called INTRP after you specified to SETUP ',&
     &' ** that you were going to use UT. This is not permitted.'
            GO TO 20
         END IF
      END IF
      IF (STATE.EQ.MINUS1) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** You have not called CT, so you cannot use INTRP.'
         GO TO 20
      END IF
      IF (STATE.GT.PLUS1) THEN
         IER = 911_id
         NREC = 2_id
         WRITE (REC,'(A/A)')&
     &' ** CT has returned with a flag value greater than 1.',&
     &' ** You cannot call INTRP in this circumstance.'
         GO TO 20
      END IF
!
!  Check input
!
      REQST1 = REQEST(1:1)
      LEGALR = REQST1 .EQ. 'S' .OR. REQST1 .EQ. 's' .OR.&
     &         REQST1 .EQ. 'D' .OR. REQST1 .EQ. 'd' .OR.&
     &         REQST1 .EQ. 'B' .OR. REQST1 .EQ. 'b'
      IF (.NOT.LEGALR) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A/A,A,A/A)')&
     &' ** You have set the first character of ',&
     &' ** REQEST to be ''',REQST1,'''. It must be one of ',&
     &' ** ''S'',''s'',''D'',''d'',''B'' or ''b''.'
         GO TO 20
      END IF
!
      IF (NWANT.GT.NEQN) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A,I6,A/A,I6,A/A)')&
     &' ** You have specified the value of NWANT to be ',NWANT,'. This',&
     &' ** is greater than ',NEQN,', which is the number of equations ',&
     &' ** in the system being integrated.'
         GO TO 20
      ELSE IF (NWANT.LT.1_id) THEN
         IER = 911_id
         NREC = 3_id
         WRITE (REC,'(A,I6,A/A/A)')&
     &' ** You have specified the value of NWANT to be ',NWANT,', but ',&
     &' ** this is less than 1. You cannot interpolate a zero or ',&
     &' ** negative number of components.'
         GO TO 20
      END IF
!
      IF (METHD.EQ.1_id) THEN
         IF (LENINT.LT.1_id) THEN
            IER = 911_id
            NREC = 2_id
            WRITE (REC,'(A,I6,A/A)')&
     &' ** You have specified LENINT to be ',LENINT,'.',&
     &' ** This is too small. LENINT must be at least 1.'
            GO TO 20
         END IF
         STARTP = 1_id
      ELSE IF (METHD.EQ.2_id) THEN
         ICHK = NEQN + MAX(NEQN,5_id*NWANT)
         IF (LENINT.LT.ICHK) THEN
            IER = 911_id
            NREC = 3_id
            WRITE (REC,'(A,I6,A/A/A,I6,A)')&
     &' ** You have specified LENINT to be ',LENINT,'. This is too',&
     &' ** small. NINT must be at least NEQ + MAX(NEQ, 5*NWANT) ',&
     &' ** which is ', ICHK,'.'
            GO TO 20
         END IF
         STARTP = NEQN + 1_id
      ELSE IF (METHD.EQ.3_id) THEN
         IER = 911_id
         NREC = 5_id
         WRITE (REC,'(A/A/A/A/A)')&
     &' ** You have been using CT with METHOD = 3 to integrate your  ',&
     &' ** equations. You have just called INTRP, but interpolation  ',&
     &' ** is not available for this METHOD. Either use METHOD = 2,  ',&
     &' ** for which interpolation is available, or use RESET to make',&
     &' ** CT step exactly to the points where you want output.'
         GO TO 20
      END IF
!
!  Has the interpolant been initialised for this step?
!
      CALL RKSIT(ASK,SRNAME,STATE)
      ININTP = STATE .NE. MINUS2
!
!  Some initialization must be done before interpolation is possible.
!  To reduce the overhead, the interpolating polynomial is formed for
!  the first NWANT components.  In the unusual circumstance that NWANT
!  is changed while still interpolating within the span of the current
!  step, the scheme must be reinitialized to accomodate the additional
!  components.
!
      IF (.NOT.ININTP .OR. NWANT.NE.NWNTSV) THEN
!
!  At present the derivative of the solution at the previous step, YPOLD(*),
!  is stored in the scratch array area starting at PRSCR. In the case of
!  METHD = 1 we can overwrite the stages.
!
         IF (METHD.EQ.1_id) THEN
            CALL FORMI(F,NEQN,NWANT,WORK(PRY),WORK(PRYP),WORK(PRYOLD),&
     &                 WORK(PRSCR),WORK(PRSTGS),.NOT.ININTP,&
     &                 WORK(PRSTGS),WORK(PRSTGS))
         ELSE
            CALL FORMI(F,NEQN,NWANT,WORK(PRY),WORK(PRYP),WORK(PRYOLD),&
     &                 WORK(PRSCR),WORK(PRSTGS),.NOT.ININTP,WRKINT,&
     &                 WRKINT(STARTP))
         END IF
!
!  Set markers to show that interpolation has been initialized for
!  NWANT components.
         NWNTSV = NWANT
         ININTP = .TRUE.
      END IF
!
!  The actual evaluation of the interpolating polynomial and/or its first
!  derivative is done in EVALI.
!
      IF (METHD.EQ.1) THEN
         CALL EVALI(WORK(PRY),WORK(PRYP),WORK(PRSTGS),TWANT,REQEST,&
     &              NWANT,YWANT,YPWANT)
      ELSE
         CALL EVALI(WORK(PRY),WORK(PRYP),WRKINT(STARTP),TWANT,REQEST,&
     &              NWANT,YWANT,YPWANT)
      END IF
!
   20 CONTINUE
!
      CALL RKMSG(IER,SRNAME,NREC,FLAG)
!
      RETURN
      END  SUBROUTINE

SUBROUTINE RESET(TENDNU)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!
!  If you are not familiar with the code RESET and how it is used in
!  conjunction with CT to solve initial value problems, you should study the
!  document file rksuite.doc carefully before attempting to use the code. The
!  following "Brief Reminder" is intended only to remind you of the meaning,
!  type, and size requirements of the arguments.
!
!  The integration using CT proceeds from TSTART in the direction of TEND, and
!  is now at TNOW.  To reset TEND to a new value TENDNU, just call RESET with
!  TENDNU as the argument.  You must continue integrating in the same
!  direction, so the sign of (TENDNU - TNOW) must be the same as the sign of
!  (TEND - TSTART). To change direction you must restart by a call to SETUP.
!
!  INPUT VARIABLE
!
!     TENDNU    - REAL(rd)
!                 The new value of TEND.
!                 Constraint: TENDNU and TNOW must be clearly distinguishable
!                 in the precision used.  The sign of (TENDNU - TNOW) must be
!                 the same as the sign of (TEND - TSTART).
!
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  TENDNU
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      CHARACTER*6       SRNAME
      PARAMETER         (SRNAME='RESET')
      LOGICAL(lgt)           ASK
      INTEGER(id)           MINUS1, MINUS2
      PARAMETER         (ASK=.TRUE.,MINUS1=-1_id,MINUS2=-2_id)
      REAL(rd)  ZERO
      PARAMETER         (ZERO=0.0_rd)
!     .. Local Scalars ..
      REAL(rd)  HMIN, TDIFF
      INTEGER(id)           FLAG, IER, NREC, STATE, STATE1
!     .. External Subroutines ..
 !     EXTERNAL          RKMSG, RKSIT
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, MAX
!     .. Executable Statements ..
      IER = 1_id
      NREC = 0_id
!
!  Is it permissible to call RESET?
!
      CALL RKSIT(ASK,'CT',STATE)
      IF (STATE.EQ.911_id) THEN
         IER = 912_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** A catastrophic error has already been detected elsewhere.'
         GO TO 20
      END IF
      IF (UTASK) THEN
         CALL RKSIT(ASK,'UT',STATE1)
         IF (STATE1.NE.MINUS2) THEN
            IER = 911_id
            NREC = 2_id
            WRITE (REC,'(A/A)')&
     &' ** You have called RESET after you specified to SETUP that ',&
     &' ** you were going to use UT. This is not permitted.'
            GO TO 20
         END IF
      END IF
      IF (STATE.EQ.MINUS1) THEN
         IER = 911_id
         NREC = 1_id
         WRITE (REC,'(A)')&
     &' ** You have not called CT, so you cannot use RESET.'
         GO TO 20
      END IF
      IF (STATE.EQ.5_id .OR. STATE.EQ.6_id) THEN
         IER = 911_id
         NREC = 2_id
         WRITE (REC,'(A,I1,A/A)')&
     &' ** CT has returned with CFLAG =  ',STATE,'.',&
     &' ** You cannot call RESET in this circumstance.'
         GO TO 20
      END IF
!
!  Check value of TENDNU
!
      IF (DIR.GT.ZERO .AND. TENDNU.LE.T) THEN
         IER = 911_id
         NREC = 4_id
         WRITE (REC,'(A/A,D13.5/A,D13.5,A/A)')&
     &' ** Integration is proceeding in the positive direction. The ',&
     &' ** current value for the independent variable is ',T,&
     &' ** and you have set TENDNU = ',TENDNU,'.  TENDNU must be ',&
     &' ** greater than T.'
      ELSE IF (DIR.LT.ZERO .AND. TENDNU.GE.T) THEN
         IER = 911_id
         NREC = 4_id
         WRITE (REC,'(A/A,D13.5/A,D13.5,A/A)')&
     &' ** Integration is proceeding in the negative direction. The ',&
     &' ** current value for the independent variable is ',T,&
     &' ** and you have set TENDNU = ',TENDNU,'.  TENDNU must be ',&
     &' ** less than T.'
      ELSE
         HMIN = MAX(TINY,TOOSML*MAX(ABS(T),ABS(TENDNU)))
         TDIFF = ABS(TENDNU-T)
         IF (TDIFF.LT.HMIN) THEN
            IER = 911_id
            NREC = 4_id
            WRITE (REC,'(A,D13.5,A/A,D13.5,A/A/A,D13.5,A)')&
     &' ** The current value of the independent variable T is ',T,'.',&
     &' ** The TENDNU you supplied has ABS(TENDNU-T) = ',TDIFF,'.',&
     &' ** For the METHOD and the precision of the computer being ',&
     &' ** used, this difference must be at least ',HMIN,'.'
         END IF
      END IF
      IF (IER.EQ.911_id) GO TO 20
!
      TND = TENDNU
      LAST = .FALSE.
!
   20 CONTINUE
!
      CALL RKMSG(IER,SRNAME,NREC,FLAG)
!
      RETURN
      END  SUBROUTINE

SUBROUTINE MCONST(METHOD)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:   Sets machine-dependent global quantities
!
!  Common:    Initializes:    /RKCOM7/ OUTCH, MCHEPS, DWARF, RNDOFF,
!                                      SQRRMC, CUBRMC, TINY
!             Reads:          none
!             Alters:         none
!
!  Comments:
!  =========
!  OUTCH, MCHEPS, DWARF are pure environmental parameters with values
!  obtained from a call to ENVIRN. The other quantities set depend on
!  the environmental parameters, the implementation, and, possibly,
!  METHOD. At present the METHODs implemented in the RK suite do not
!  influence the values of these quantities.
!  OUTCH  - Standard output channel
!  MCHEPS - Largest positive number such that 1.0_rd + MCHEPS = 1.0_rd
!  DWARF  - Smallest positive number
!  RNDOFF - 10 times MCHEPS
!  SQRRC - Square root of MCHEPS
!  CUBRMC - Cube root of MCHEPS
!  TINY   - Square root of DWARF
!
USE kindtype
implicit none
!     .. Scalar Arguments ..
      INTEGER(id)           METHOD
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Parameters ..
      REAL(rd)  TEN, THIRD
      PARAMETER         (TEN=10.0_rd,THIRD=1.0_rd/3.0_rd)
!     .. External Subroutines ..
!      EXTERNAL          ENVIRN
!     .. Intrinsic Functions ..
      INTRINSIC         SQRT
!     .. Executable Statements ..
!
      CALL ENVIRN(OUTCH,MCHEPS,DWARF)
!
      RNDOFF = TEN*MCHEPS
      SQRRMC = SQRT(MCHEPS)
      CUBRMC = MCHEPS**THIRD
      TINY = SQRT(DWARF)
!
      RETURN
      END  SUBROUTINE
      SUBROUTINE ENVIRN(OUTCH,MCHEPS,DWARF)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!
!  The RK suite requires some environmental parameters that are provided by
!  this subroutine.  The values provided with the distribution codes are those
!  appropriate to the IEEE standard.  They must be altered, if necessary, to
!  those appropriate to the computing system you are using before calling the 
!  codes of the suite.  
!
!        ================================================================
!        ================================================================
!        TO MAKE SURE THAT THESE MACHINE AND INSTALLATION DEPENDENT 
!        QUANTITIES ARE SPECIFIED PROPERLY, THE DISTRIBUTION VERSION 
!        WRITES A MESSAGE ABOUT THE MATTER TO THE STANDARD OUTPUT CHANNEL
!        AND TERMINATES THE RUN.  THE VALUES PROVIDED IN THE DISTRIBUTION
!        VERSION SHOULD BE ALTERED, IF NECESSARY, AND THE "WRITE" AND 
!        "STOP" STATEMENTS COMMENTED OUT.
!        ================================================================
!        ================================================================
!
!  OUTPUT VARIABLES 
!
!     OUTCH     - INTEGER(id)
!                 Standard output channel
!     MCHEPS    - REAL(rd)
!                 MCHEPS is the largest positive number such that
!                 1.0_rd + MCHEPS = 1.0_rd. 
!     DWARF     - REAL(rd)
!                 DWARF is the smallest positive number.
!
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
!
!     .. Scalar Arguments ..  
use kindtype
implicit none
      INTEGER(id)           OUTCH
      REAL(rd)  DWARF, MCHEPS
!     .. Executable Statements ..      
!
!  The following six statements are to be Commented out after verification that
!  the machine and installation dependent quantities are specified correctly.
!  If you pass copies of RKSUITE on to others, please give them the whole
!  distribution version of RKSUITE, and in particular, give them a version 
!  of ENVIRN that does not have the following six statements Commented out.
!      WRITE(*,*) ' Before using RKSUITE, you must verify that the  '
!      WRITE(*,*) ' machine- and installation-dependent quantities  '
!      WRITE(*,*) ' specified in the subroutine ENVIRN are correct, '
!      WRITE(*,*) ' and then Comment these WRITE statements and the '
!      WRITE(*,*) ' STOP statement out of ENVIRN.                   '
!      STOP
!
!  The following values are appropriate to IEEE arithmetic with the typical
!  standard output channel.
!
      OUTCH = 6
      MCHEPS = epsilon(1.0_rd)

      DWARF = tiny(1.0_rd)

!      
!------------------------------------------------------------------------------
!  If you have the routines D1MACH and I1MACH on your system, you could
!  replace the preceding statements by the following ones to obtain the 
!  appropriate machine dependent numbers. The routines D1MACH and I1MACH 
!  are public domain software.  They are available from NETLIB.
!      .. Scalar Arguments ..  
!      INTEGER(id)           OUTCH
!      REAL(rd)  DWARF, MCHEPS
!      .. External Functions ..
!      INTEGER(id)           I1MACH
!      REAL(rd)  D1MACH
!      .. Executable Statements ..
!
!      OUTCH = I1MACH(2)
!      MCHEPS = D1MACH(3)
!      DWARF = D1MACH(1)
!
!  If you have the NAG Fortran Library available on your system, you could 
!  replace the preceding statements by the following ones to obtain the 
!  appropriate machine dependent numbers.
!
!      .. Scalar Arguments ..  
!      INTEGER(id)           OUTCH
!      REAL(rd)  DWARF, MCHEPS
!      .. External Functions ..
!      REAL(rd)  X02AJF, X02AMF
!      .. Executable Statements ..
!
!      CALL X04AAF(0,OUTCH)
!      MCHEPS = X02AJF()
!      DWARF = X02AMF()
!
!  If you have the IMSL MATH/LIBRARY available on your system, you could
!  replace the preceding statements by the following ones to obtain the
!  appropriate machine dependent numbers.
!
!      .. Scalar Arguments ..  
!      INTEGER(id)           OUTCH
!      REAL(rd)  DWARF, MCHEPS
!      .. External Functions ..
!      REAL(rd)  DMACH
!      .. Executable Statements ..
!
!      CALL UMACH(2,OUTCH)
!      MCHEPS = DMACH(4)
!      DWARF = DMACH(1)
!------------------------------------------------------------------------------
!
      RETURN
      END  SUBROUTINE
      SUBROUTINE CONST(METHOD,VECSTG,REQSTG,LINTPL)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!*************************************************
!
!  Purpose:   Set formula definitions and formula characteristics for
!             selected method. Return storage requirements for the
!             selected method.
!
!  Input:     METHOD
!  Output:    VECSTG, REQSTG, LINTPL
!
!  Common:    Initializes:    /RKCOM4/ A(*,*), B(*), C(*), BHAT(*), R(*),
!                                      E(*), PTR(*), NSTAGE, METHD, INTP, MINTP
!                             /RKCOM5/ TOOSML, COST, SAFETY, EXPON, STBRAD,
!                                      TANANG, RS, RS1, RS2, RS3, RS4, ORDER,
!                                      LSTSTG, MAXTRY, NSEC, FSAL
!             Reads:          /RKCOM7/ RNDOFF
!             Alters:         none
!
!  Comments:
!  =========
!  Runge-Kutta formula pairs are described by a set of coefficients
!  and by the setting of a number of parameters that describe the
!  characteristics of the pair.  The input variable METHD indicates
!  which of the three pairs available is to be set. In the case of
!  METHD = 2 additional coefficients are defined that make interpolation
!  of the results possible and provide an additional error estimator.
!  VECSTG is the number of columns of workspace required to compute the
!  stages of a METHD. For interpolation purposes the routine returns via
!  COMMON the logical variable INTP indicating whether interpolation is
!  possible and via the call list:
!  REQSTG - whether the stages are required to form the
!           interpolant (set .FALSE. if INTP=.FALSE.)
!  LINTPL - the number of extra columns of storage required for use
!           with UT (set 0 if INTP=.FALSE.)
!
!  Quantities set in common blocks:
!  METHD - copy of METHOD
!  A, B, C, BHAT - coefficients of the selected method
!  R      - extra coefficents for interpolation with METHD = 2
!  E      - extra coefficients for additional local error estimate
!           with METHD = 2
!  PTR    - vector of pointers indicating how individual stages are to
!           be stored.  With it zero coefficients of the formulas can
!           be exploited to reduce the storage required
!  NSTAGE - number of stages for the specified METHD
!  INTP   - indicates whether there is an associated interpolant
!           (depending on the method, the user may have to supply
!           extra workspace)
!  MINTP  - the degree of the interpolating polynomial, if one exists
!  FSAL   - indicates whether the last stage of a step can be used as
!           the first stage of the following step
!  LSTSTG - pointer to location of last stage for use with FSAL=.TRUE.
!  ORDER  - the lower order of the pair of Runge-Kutta formulas that
!           constitute a METHD
!  TANANG, 
!  STBRAD - the stability region of the formula used to advance
!           the integration is approximated by a sector in the left half
!           complex plane.  TANANG is the tangent of the interior angle
!           of the sector and STBRAD is the radius of the sector.
!  COST   - cost of a successful step in function evaluations
!  MAXTRY - limit on the number of iterations in the stiffness check. As
!           set, no more than 24 function evaluations are made in the check.
!  NSEC   - each step of size H in the primary integration corresponds to
!           NSEC steps of size H/NSEC in the secondary integration when
!           global error assessment is done.
!  EXPON  - used to adjust the step size; this code implements an error
!           per step control for which EXPON = 1/(ORDER + 1).
!  SAFETY - quantity used in selecting the step size
!  TOOSML - quantity used to determine when a step size is too small for
!           the precision available
!  RS, RS1,
!  RS2, RS3,
!  RS4    - quantities used in determining the maximum and minimum change
!           change in step size (set independently of METHD)
!
!  Further comments on SAFETY:
! ============================
!  The code estimates the largest step size that will yield the specified
!  accuracy.  About half the time this step size would result in a local
!  error that is a little too large, and the step would be rejected.  For
!  this reason a SAFETY factor is used to reduce the "optimal" value to one
!  more likely to succeed.  Unfortunately, there is some art in choosing this
!  value. The more expensive a failed step, the smaller one is inclined to 
!  take this factor. However, a small factor means that if the prediction were
!  good, more accuracy than desired would be obtained and the behavior of the
!  error would then be irregular.  The more stringent the tolerance, the better
!  the prediction, except near limiting precision. Thus the general range of 
!  tolerances expected influences the choice of SAFETY.
USE kindtype
implicit none
!     .. Scalar Arguments ..
      INTEGER(id)           LINTPL, METHOD, VECSTG
      LOGICAL(lgt)           REQSTG
!     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Parameters ..
      REAL(rd)  ONE, ZERO, TWO, FIFTY, FIVEPC
      PARAMETER         (ONE=1.0_rd,ZERO=0.0_rd,TWO=2.0_rd,FIFTY=50.0_rd,&
     &                  FIVEPC=0.05_rd)
!     .. Local Scalars ..
      REAL(rd)  CDIFF, DIFF
      INTEGER(id)           I, J
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, DBLE, INT, MAX, MIN
!     .. Executable Statements ..
!
      METHD = METHOD
!
      GO TO (20,40,100) METHD
!
!  METHD = 1.
!    This pair is from "A 3(2) Pair of Runge-Kutta Formulas" by P. Bogacki
!    and L.F. Shampine, Appl. Math. Lett., 2, pp. 321-325, 1989.  The authors
!    are grateful to P. Bogacki for his assistance in implementing the pair.
!
   20 CONTINUE
      NSTAGE = 4_id
      FSAL = .TRUE.
      ORDER = 2_id
      TANANG = 8.9_rd
      STBRAD = 2.3_rd
      SAFETY = 0.8_rd
      INTP = .TRUE.
      MINTP = 3_id
      REQSTG = .FALSE.
      LINTPL = 2_id
      NSEC = 3_id
!
      PTR(1) = 0_id
      PTR(2) = 1_id
      PTR(3) = 2_id
      PTR(4) = 3_id
!
      A(2,1) = 1._rd/2._rd
      A(3,1) = 0._rd
      A(3,2) = 3._rd/4._rd
      A(4,1) = 2._rd/9._rd
      A(4,2) = 1._rd/3._rd
      A(4,3) = 4._rd/9._rd
!
!  The coefficients BHAT(*) refer to the formula used to advance the
!  integration, here the one of order 3.  The coefficients B(*) refer
!  to the other formula, here the one of order 2. For this pair, BHAT(*)
!  is not needed since FSAL = .TRUE.
!
      B(1) = 7._rd/24._rd
      B(2) = 1._rd/4._rd
      B(3) = 1._rd/3._rd
      B(4) = 1._rd/8._rd
!
      C(1) = 0._rd
      C(2) = 1._rd/2._rd
      C(3) = 3._rd/4._rd
      C(4) = 1._rd
!
      GO TO 120
!
!  METHD = 2
!    This pair is from "An Efficient Runge-Kutta (4,5) Pair" by P. Bogacki
!    and L.F. Shampine, Rept. 89-20, Math. Dept., Southern Methodist
!    University, Dallas, Texas, USA, 1989.  The authors are grateful to
!    P. Bogacki for his assistance in implementing the pair.  Shampine and
!    Bogacki subsequently modified the formula to enhance the reliability of
!    the pair.  The original fourth order formula is used in an estimate of
!    the local error.  If the step fails, the computation is broken off.  If
!    the step is acceptable, the first evaluation of the next step is done,
!    i.e., the pair is implemented as FSAL and the local error of the step
!    is again estimated with a fourth order formula using the additional data.
!    The step must succeed with both estimators to be accepted.  When the
!    second estimate is formed, it is used for the subsequent adjustment of
!    the step size because it is of higher quality.  The two fourth order
!    formulas are well matched to leading order, and only exceptionally do
!    the estimators disagree -- problems with discontinuous coefficients are
!    handled more reliably by using two estimators as is global error
!    estimation.
!
   40 CONTINUE
      NSTAGE = 8_id
      FSAL = .TRUE.
      ORDER = 4_id
      TANANG = 5.2_rd
      STBRAD = 3.9_rd
      SAFETY = 0.8_rd
      INTP = .TRUE.
      REQSTG = .TRUE.
      MINTP = 6_id
      LINTPL = 6_id
      NSEC = 2_id
!
      PTR(1) = 0_id
      PTR(2) = 1_id
      PTR(3) = 2_id
      PTR(4) = 3_id
      PTR(5) = 4_id
      PTR(6) = 5_id
      PTR(7) = 6_id
      PTR(8) = 7_id
!
      A(2,1) = 1._rd/6._rd
      A(3,1) = 2._rd/27._rd
      A(3,2) = 4._rd/27._rd
      A(4,1) = 183._rd/1372._rd
      A(4,2) = -162._rd/343._rd
      A(4,3) = 1053._rd/1372._rd
      A(5,1) = 68._rd/297._rd
      A(5,2) = -4._rd/11._rd
      A(5,3) = 42._rd/143._rd
      A(5,4) = 1960._rd/3861._rd
      A(6,1) = 597._rd/22528._rd
      A(6,2) = 81._rd/352._rd
      A(6,3) = 63099._rd/585728._rd
      A(6,4) = 58653._rd/366080._rd
      A(6,5) = 4617._rd/20480._rd
      A(7,1) = 174197._rd/959244._rd
      A(7,2) = -30942._rd/79937._rd
      A(7,3) = 8152137._rd/19744439._rd
      A(7,4) = 666106._rd/1039181._rd
      A(7,5) = -29421._rd/29068._rd
      A(7,6) = 482048._rd/414219._rd
      A(8,1) = 587._rd/8064._rd
      A(8,2) = 0._rd
      A(8,3) = 4440339._rd/15491840._rd
      A(8,4) = 24353._rd/124800._rd
      A(8,5) = 387._rd/44800._rd
      A(8,6) = 2152._rd/5985._rd
      A(8,7) = 7267._rd/94080._rd
!
!  The coefficients B(*) refer to the formula of order 4.
!
      B(1) = 2479._rd/34992._rd
      B(2) = 0._rd
      B(3) = 123._rd/416._rd
      B(4) = 612941._rd/3411720._rd
      B(5) = 43._rd/1440._rd
      B(6) = 2272._rd/6561._rd
      B(7) = 79937._rd/1113912._rd
      B(8) = 3293._rd/556956._rd
!
!  The coefficients E(*) refer to an estimate of the local error based on
!  the first formula of order 4.  It is the difference of the fifth order
!  result, here located in A(8,*), and the fourth order result.  By
!  construction both E(2) and E(7) are zero.
!
      E(1) = -3._rd/1280._rd
      E(2) = 0._rd
      E(3) = 6561._rd/632320._rd
      E(4) = -343._rd/20800._rd
      E(5) = 243._rd/12800._rd
      E(6) = -1._rd/95._rd
      E(7) = 0._rd
!
      C(1) = 0._rd
      C(2) = 1._rd/6._rd
      C(3) = 2._rd/9._rd
      C(4) = 3._rd/7._rd
      C(5) = 2._rd/3._rd
      C(6) = 3._rd/4._rd
      C(7) = 1._rd
      C(8) = 1._rd
!
!  To do interpolation with this pair, some extra stages have to be computed.
!  The following additional A(*,*) and C(*) coefficients are for this purpose.
!  In addition there is an array R(*,*) that plays a role for interpolation
!  analogous to that of BHAT(*) for the basic step.
!
      C(9) = 1._rd/2._rd
      C(10) = 5._rd/6._rd
      C(11) = 1._rd/9._rd
!
      A(9,1) = 455._rd/6144._rd
      A(10,1) = -837888343715._rd/13176988637184._rd
      A(11,1) = 98719073263._rd/1551965184000._rd
      A(9,2) = 0._rd
      A(10,2) = 30409415._rd/52955362._rd
      A(11,2) = 1307._rd/123552._rd
      A(9,3) = 10256301._rd/35409920._rd
      A(10,3) = -48321525963._rd/759168069632._rd
      A(11,3) = 4632066559387._rd/70181753241600._rd
      A(9,4) = 2307361._rd/17971200._rd
      A(10,4) = 8530738453321._rd/197654829557760._rd
      A(11,4) = 7828594302389._rd/382182512025600._rd
      A(9,5) = -387._rd/102400._rd
      A(10,5) = 1361640523001._rd/1626788720640._rd
      A(11,5) = 40763687._rd/11070259200._rd
      A(9,6) = 73._rd/5130._rd
      A(10,6) = -13143060689._rd/38604458898._rd
      A(11,6) = 34872732407._rd/224610586200._rd
      A(9,7) = -7267._rd/215040._rd
      A(10,7) = 18700221969._rd/379584034816._rd
      A(11,7) = -2561897._rd/30105600._rd
      A(9,8) = 1._rd/32._rd
      A(10,8) = -5831595._rd/847285792._rd
      A(11,8) = 1._rd/10._rd
      A(10,9) = -5183640._rd/26477681._rd
      A(11,9) = -1._rd/10._rd
      A(11,10) = -1403317093._rd/11371610250._rd
!
      DO 60 I = 1_id, 11_id
         R(I,1) = 0._rd
   60 CONTINUE
      DO 80 I = 1_id, 6_id
         R(2,I) = 0._rd
   80 CONTINUE
      R(1,6) = -12134338393._rd/1050809760._rd
      R(1,5) = -1620741229._rd/50038560._rd
      R(1,4) = -2048058893._rd/59875200._rd
      R(1,3) = -87098480009._rd/5254048800._rd
      R(1,2) = -11513270273._rd/3502699200._rd
!
      R(3,6) = -33197340367._rd/1218433216._rd
      R(3,5) = -539868024987._rd/6092166080._rd
      R(3,4) = -39991188681._rd/374902528._rd
      R(3,3) = -69509738227._rd/1218433216._rd
      R(3,2) = -29327744613._rd/2436866432._rd
!
      R(4,6) = -284800997201._rd/19905339168._rd
      R(4,5) = -7896875450471._rd/165877826400._rd
      R(4,4) = -333945812879._rd/5671036800._rd
      R(4,3) = -16209923456237._rd/497633479200._rd
      R(4,2) = -2382590741699._rd/331755652800._rd
!
      R(5,6) = -540919._rd/741312._rd
      R(5,5) = -103626067._rd/43243200._rd
      R(5,4) = -633779._rd/211200._rd
      R(5,3) = -32406787._rd/18532800._rd
      R(5,2) = -36591193._rd/86486400._rd
!
      R(6,6) = 7157998304._rd/374350977._rd
      R(6,5) = 30405842464._rd/623918295._rd
      R(6,4) = 183022264._rd/5332635._rd
      R(6,3) = -3357024032._rd/1871754885._rd
      R(6,2) = -611586736._rd/89131185._rd
!
      R(7,6) = -138073._rd/9408._rd
      R(7,5) = -719433._rd/15680._rd
      R(7,4) = -1620541._rd/31360._rd
      R(7,3) = -385151._rd/15680._rd
      R(7,2) = -65403._rd/15680._rd
!
      R(8,6) = 1245._rd/64._rd
      R(8,5) = 3991._rd/64._rd
      R(8,4) = 4715._rd/64._rd
      R(8,3) = 2501._rd/64._rd
      R(8,2) = 149._rd/16._rd
      R(8,1) = 1._rd
!
      R(9,6) = 55._rd/3._rd
      R(9,5) = 71._rd
      R(9,4) = 103._rd
      R(9,3) = 199._rd/3._rd
      R(9,2) = 16.0_rd
!
      R(10,6) = -1774004627._rd/75810735._rd
      R(10,5) = -1774004627._rd/25270245._rd
      R(10,4) = -26477681._rd/359975._rd
      R(10,3) = -11411880511._rd/379053675._rd
      R(10,2) = -423642896._rd/126351225._rd
!
      R(11,6) = 35._rd
      R(11,5) = 105._rd
      R(11,4) = 117._rd
      R(11,3) = 59._rd
      R(11,2) = 12._rd
!
      GO TO 120
!
!  METHD = 3
!    This pair is from "High Order Embedded Runge-Kutta Formulae" by P.J.
!    Prince and J.R. Dormand, J. Comp. Appl. Math.,7, pp. 67-75, 1981.  The
!    authors are grateful to P. Prince and J. Dormand for their assistance in
!    implementing the pair.
!
  100 CONTINUE
      NSTAGE = 13_id
      FSAL = .FALSE.
      ORDER = 7_id
      TANANG = 11.0_rd
      STBRAD = 5.2_rd
      SAFETY = 0.8_rd
      INTP = .FALSE.
      REQSTG = .FALSE.
      MINTP = 0_id
      LINTPL = 0_id
      NSEC = 2_id
!
      PTR(1) = 0_id
      PTR(2) = 1_id
      PTR(3) = 2_id
      PTR(4) = 1_id
      PTR(5) = 3_id
      PTR(6) = 2_id
      PTR(7) = 4_id
      PTR(8) = 5_id
      PTR(9) = 6_id
      PTR(10) = 7_id
      PTR(11) = 8_id
      PTR(12) = 9_id
      PTR(13) = 1_id
!
      A(2,1) = 5.55555555555555555555555555556e-2_rd
      A(3,1) = 2.08333333333333333333333333333e-2_rd
      A(3,2) = 6.25e-2_rd
      A(4,1) = 3.125e-2_rd
      A(4,2) = 0._rd
      A(4,3) = 9.375e-2_rd
      A(5,1) = 3.125e-1_rd
      A(5,2) = 0._rd
      A(5,3) = -1.171875_rd
      A(5,4) = 1.171875_rd
      A(6,1) = 3.75e-2_rd
      A(6,2) = 0._rd
      A(6,3) = 0._rd
      A(6,4) = 1.875e-1_rd
      A(6,5) = 1.5e-1_rd
      A(7,1) = 4.79101371111111111111111111111e-2_rd
      A(7,2) = 0._rd
      A(7,3) = 0.0_rd
      A(7,4) = 1.12248712777777777777777777778e-1_rd
      A(7,5) = -2.55056737777777777777777777778e-2_rd
      A(7,6) = 1.28468238888888888888888888889e-2_rd
      A(8,1) = 1.6917989787292281181431107136e-2_rd
      A(8,2) = 0._rd
      A(8,3) = 0._rd
      A(8,4) = 3.87848278486043169526545744159e-1_rd
      A(8,5) = 3.59773698515003278967008896348e-2_rd
      A(8,6) = 1.96970214215666060156715256072e-1_rd
      A(8,7) = -1.72713852340501838761392997002e-1_rd
      A(9,1) = 6.90957533591923006485645489846e-2_rd
      A(9,2) = 0._rd
      A(9,3) = 0._rd
      A(9,4) = -6.34247976728854151882807874972e-1_rd
      A(9,5) = -1.61197575224604080366876923982e-1_rd
      A(9,6) = 1.38650309458825255419866950133e-1_rd
      A(9,7) = 9.4092861403575626972423968413e-1_rd
      A(9,8) = 2.11636326481943981855372117132e-1_rd
      A(10,1) = 1.83556996839045385489806023537e-1_rd
      A(10,2) = 0._rd
      A(10,3) = 0._rd
      A(10,4) = -2.46876808431559245274431575997_rd
      A(10,5) = -2.91286887816300456388002572804e-1_rd
      A(10,6) = -2.6473020233117375688439799466e-2_rd
      A(10,7) = 2.84783876419280044916451825422_rd
      A(10,8) = 2.81387331469849792539403641827e-1_rd
      A(10,9) = 1.23744899863314657627030212664e-1_rd
      A(11,1) = -1.21542481739588805916051052503_rd
      A(11,2) = 0._rd
      A(11,3) = 0._rd
      A(11,4) = 1.66726086659457724322804132886D1
      A(11,5) = 9.15741828416817960595718650451e-1_rd
      A(11,6) = -6.05660580435747094755450554309_rd
      A(11,7) = -1.60035735941561781118417064101D1
      A(11,8) = 1.4849303086297662557545391898D1
      A(11,9) = -1.33715757352898493182930413962D1
      A(11,10) = 5.13418264817963793317325361166_rd
      A(12,1) = 2.58860916438264283815730932232e-1_rd
      A(12,2) = 0._rd
      A(12,3) = 0._rd
      A(12,4) = -4.77448578548920511231011750971_rd
      A(12,5) = -4.3509301377703250944070041181e-1_rd
      A(12,6) = -3.04948333207224150956051286631_rd
      A(12,7) = 5.57792003993609911742367663447_rd
      A(12,8) = 6.15583158986104009733868912669_rd
      A(12,9) = -5.06210458673693837007740643391_rd
      A(12,10) = 2.19392617318067906127491429047_rd
      A(12,11) = 1.34627998659334941535726237887e-1_rd
      A(13,1) = 8.22427599626507477963168204773e-1_rd
      A(13,2) = 0._rd
      A(13,3) = 0._rd
      A(13,4) = -1.16586732572776642839765530355D1
      A(13,5) = -7.57622116690936195881116154088e-1_rd
      A(13,6) = 7.13973588159581527978269282765e-1_rd
      A(13,7) = 1.20757749868900567395661704486D1
      A(13,8) = -2.12765911392040265639082085897_rd
      A(13,9) = 1.99016620704895541832807169835_rd
      A(13,10) = -2.34286471544040292660294691857e-1_rd
      A(13,11) = 1.7589857770794226507310510589e-1_rd
      A(13,12) = 0._rd
!
!  The coefficients BHAT(*) refer to the formula used to advance the
!  integration, here the one of order 8.  The coefficients B(*) refer
!  to the other formula, here the one of order 7.
!
      BHAT(1) = 4.17474911415302462220859284685e-2_rd
      BHAT(2) = 0._rd
      BHAT(3) = 0._rd
      BHAT(4) = 0._rd
      BHAT(5) = 0._rd
      BHAT(6) = -5.54523286112393089615218946547e-2_rd
      BHAT(7) = 2.39312807201180097046747354249e-1_rd
      BHAT(8) = 7.0351066940344302305804641089e-1_rd
      BHAT(9) = -7.59759613814460929884487677085e-1_rd
      BHAT(10) = 6.60563030922286341461378594838e-1_rd
      BHAT(11) = 1.58187482510123335529614838601e-1_rd
      BHAT(12) = -2.38109538752862804471863555306e-1_rd
      BHAT(13) = 2.5e-1_rd
!
      B(1) = 2.9553213676353496981964883112e-2_rd
      B(2) = 0._rd
      B(3) = 0._rd
      B(4) = 0._rd
      B(5) = 0._rd
      B(6) = -8.28606276487797039766805612689e-1_rd
      B(7) = 3.11240900051118327929913751627e-1_rd
      B(8) = 2.46734519059988698196468570407_rd
      B(9) = -2.54694165184190873912738007542_rd
      B(10) = 1.44354858367677524030187495069_rd
      B(11) = 7.94155958811272872713019541622e-2_rd
      B(12) = 4.44444444444444444444444444445e-2_rd
      B(13) = 0._rd
!
      C(1) = 0._rd
      C(2) = 5.55555555555555555555555555556e-2_rd
      C(3) = 8.33333333333333333333333333334e-2_rd
      C(4) = 1.25e-1_rd
      C(5) = 3.125e-1_rd
      C(6) = 3.75e-1_rd
      C(7) = 1.475e-1_rd
      C(8) = 4.65e-1_rd
      C(9) = 5.64865451382259575398358501426e-1_rd
      C(10) = 6.5e-1_rd
      C(11) = 9.24656277640504446745013574318e-1_rd
      C(12) = 1._rd
      C(13) = C(12)
!
      GO TO 120
!
!  The definitions of all pairs come here for the calculation of
!  LSTSTG, RS1, RS2, RS3, RS4, COST, MAXTRY, EXPON, TOOSML, and VECSTG.
!
  120 CONTINUE
      LSTSTG = PTR(NSTAGE)
      IF (FSAL) THEN
         COST = DBLE(NSTAGE-1_id)
      ELSE
         COST = DBLE(NSTAGE)
      END IF
!
!  MAXTRY - limit on the number of iterations of a computation made in
!  diagnosing stiffness.  There are at most Q = 3 function calls per
!  iteration. MAXTRY is determined so that  Q*MAXTRY <= 5% of the cost of
!  50 steps and 1 <= MAXTRY <= 8. This limits the number of calls to FCN
!  in each diagnosis of stiffness to 24 calls.
!
      MAXTRY = MIN(8_id,MAX(1_id,INT(FIVEPC*COST*FIFTY)))
!
      EXPON = ONE/(ORDER+ONE)
!
!     In calculating CDIFF it is assumed that there will be a non-zero
!     difference |C(I) - C(J)| less than one. If C(I) = C(J) for any I not
!     equal to J, they should be made precisely equal by assignment.
!
      CDIFF = ONE
      DO 160 I = 1_id, NSTAGE - 1_id
         DO 140 J = I + 1_id, NSTAGE
            DIFF = ABS(C(I)-C(J))
            IF (DIFF.NE.ZERO) CDIFF = MIN(CDIFF,DIFF)
  140    CONTINUE
  160 CONTINUE
      TOOSML = RNDOFF/CDIFF
!
!  Determine the number of columns needed in STAGES(1:NEQ,*) (this will be
!  at most NSTAGE-1 since the first stage is held in a separate array).
!  The PTR array contains the column positions of the stages.
!
      VECSTG = 0_id
      DO 180 I = 2_id, NSTAGE
         VECSTG = MAX(PTR(I),VECSTG)
  180 CONTINUE
!
      RS = TWO
      RS1 = ONE/RS
      RS2 = RS**2_id
      RS3 = RS**3_id
      RS4 = ONE/RS3
!
      RETURN
      END SUBROUTINE
    


  SUBROUTINE FORMI(F,NEQ,NWANT,Y,YP,YOLD,YPOLD,STAGES,CALSTG,&
     &                 XSTAGE,P)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:    Forms an interpolating polynomial for use with
!              METHDs 1 or 2.
!
!  Input:      NEQ, NWANT, T, Y(*), YP(*), HOLD, YOLD(*), YPOLD(*),
!              STAGES(NEQ,*), CALSTG
!  Output:     P(*), XSTAGE(NEQ)
!  External:   F
!
!  Common:     Initializes:    none
!              Reads:          /RKCOM4/ A(*,*), C(*), R(*), METHD, MINTP
!                              /RKCOM2/ T, TOLD, HOLD
!              Alters:         /RKCOM2/ NFCN
!
!  Comments:
!  =========
!  The integration has reached T with a step HOLD from TOLD = T-HOLD.
!  Y(*),YP(*) and YOLD(*),YPOLD(*) approximate the solution and its
!  derivative at T and TOLD respectively.  STAGES(NEQ,*) holds the stages
!  computed in taking this step. In the case of METHD = 2 it is necessary 
!  to compute some more stages in this subroutine. CALSTG indicates whether
!  or not the extra stages need to be computed. A(*,*) and C(*) are used in 
!  computing these stages. The extra stages are stored in STAGES(NEQ,*) and 
!  XSTAGE(*).  The coefficients of the interpolating polynomials for the first
!  NWANT components of the solution are returned in the array P(*). The 
!  polynomial is of degree MINTP = 3 for METHD = 1 and of degree MINTP = 6 
!  for METHD = 2. The vector R(*) is used for workspace when METHD = 2.
USE kindtype
implicit none
!     .. Scalar Arguments ..
      INTEGER(id)           NEQ, NWANT
      LOGICAL(lgt)           CALSTG
!     .. Array Arguments ..
      REAL(rd)  P(*), STAGES(NEQ,*), XSTAGE(*), Y(*), YOLD(*),&
     &                  YP(*), YPOLD(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!     .. Local Scalars ..
      REAL(rd)  D1, D2, D3, D4, HYP, HYPOLD
      INTEGER(id)           I, J, K, L
!     .. Executable Statements ..
!
      IF (METHD.EQ.1_id) THEN
!
!  METHD = 1.  Use the cubic Hermite interpolant that is is fully
!  specified by the values and slopes at the two ends of the step.
!
         DO 20 L = 1_id, NWANT
            D1 = Y(L) - YOLD(L)
            HYP = HOLD*YP(L)
            HYPOLD = HOLD*YPOLD(L)
            D2 = HYP - D1
            D3 = D1 - HYPOLD
            D4 = D2 - D3
            P(L) = D2 + D4
            P(NWANT+L) = D4
   20    CONTINUE
!
      ELSE
!
!  METHD = 2.
!
         IF (CALSTG) THEN
!
!  Compute the extra stages needed for interpolation using the facts that
!       1. Stage 1 is YPOLD(*).
!       2. Stage i (i>1) is stored in STAGES(1:NEQ,i).
!       3. This pair is FSAL, i.e. STAGES(1:NEQ,7)=YP(1:NEQ), which frees
!          up STAGES(1:NEQ,7) for use by stage 9.
!       4. XSTAGE(1:NEQ) is used for stage 10.
!       5. The coefficient of stage 2 in the interpolant is always 0, so
!          STAGES(1:NEQ,1) is used for stage 11.
!  The vector P(1:NEQ) is used as workspace for computing the stages.
!
            DO 180 I = 9_id, 11_id
               DO 40 L = 1_id, NEQ
                  P(L) = A(I,1)*YPOLD(L)
   40          CONTINUE
               DO 140 J = 2_id, I - 1_id
                  IF (J.LE.7) THEN
                     DO 60 L = 1_id, NEQ
                        P(L) = P(L) + A(I,J)*STAGES(L,J-1_id)
   60                CONTINUE
                  ELSE IF (J.EQ.8_id) THEN
                     DO 80 L = 1_id, NEQ
                        P(L) = P(L) + A(I,J)*YP(L)
   80                CONTINUE
                  ELSE IF (J.EQ.9) THEN
                     DO 100 L = 1_id, NEQ
                        P(L) = P(L) + A(I,J)*STAGES(L,7_id)
  100                CONTINUE
                  ELSE IF (J.EQ.10) THEN
                     DO 120 L = 1_id, NEQ
                        P(L) = P(L) + A(I,J)*XSTAGE(L)
  120                CONTINUE
                  END IF
  140          CONTINUE
               DO 160 L = 1_id, NEQ
                  P(L) = YOLD(L) + HOLD*P(L)
  160          CONTINUE
               IF (I.EQ.9_id) THEN
                  CALL F(TOLD+C(I)*HOLD,P,STAGES(1,7))
                  NFCN = NFCN + 1_id
               ELSE IF (I.EQ.10_id) THEN
                  CALL F(TOLD+C(I)*HOLD,P,XSTAGE)
                  NFCN = NFCN + 1_id
               ELSE
                  CALL F(TOLD+C(I)*HOLD,P,STAGES(1,1))
                  NFCN = NFCN + 1_id
               END IF
  180       CONTINUE
         END IF
!
!  Form the coefficients of the interpolating polynomial in its shifted
!  and scaled form.  The transformation from the form in which the
!  polynomial is derived can be somewhat ill-conditioned.  The terms 
!  are grouped so as to minimize the errors of the transformation.
!
!  Coefficient of SIGMA**6
         K = 4_id*NWANT
         DO 200 L = 1_id, NWANT
            P(K+L) = R(5,6)*STAGES(L,4) +&
     &               ((R(10,6)*XSTAGE(L)+R(8,6)*YP(L))+&
     &               (R(7,6)*STAGES(L,6)+R(6,6)*STAGES(L,5))) +&
     &               ((R(4,6)*STAGES(L,3)+R(9,6)*STAGES(L,7))+&
     &               (R(3,6)*STAGES(L,2)+R(11,6)*STAGES(L,1))+&
     &               R(1,6)*YPOLD(L))
  200    CONTINUE
!
!  Coefficient of SIGMA**5
         K = 3_id*NWANT
         DO 220 L = 1_id, NWANT
            P(K+L) = (R(10,5)*XSTAGE(L)+R(9,5)*STAGES(L,7)) +&
     &               ((R(7,5)*STAGES(L,6)+R(6,5)*STAGES(L,5))+&
     &               R(5,5)*STAGES(L,4)) + ((R(4,5)*STAGES(L,3)+&
     &               R(8,5)*YP(L))+(R(3,5)*STAGES(L,2)+R(11,5)*&
     &               STAGES(L,1))+R(1,5)*YPOLD(L))
  220    CONTINUE
!
!  Coefficient of SIGMA**4
         K = 2_id*NWANT
         DO 240 L = 1_id, NWANT
            P(K+L) = ((R(4,4)*STAGES(L,3)+R(8,4)*YP(L))+&
     &               (R(7,4)*STAGES(L,6)+R(6,4)*STAGES(L,5))+&
     &               R(5,4)*STAGES(L,4)) + ((R(10,4)*XSTAGE(L)+&
     &               R(9,4)*STAGES(L,7))+(R(3,4)*STAGES(L,2)+&
     &               R(11,4)*STAGES(L,1))+R(1,4)*YPOLD(L))
  240    CONTINUE
!
!  Coefficient of SIGMA**3
         K = NWANT
         DO 260 L = 1_id, NWANT
            P(K+L) = R(5,3)*STAGES(L,4) + R(6,3)*STAGES(L,5) +&
     &               ((R(3,3)*STAGES(L,2)+R(9,3)*STAGES(L,7))+&
     &               (R(10,3)*XSTAGE(L)+R(8,3)*YP(L))+R(1,3)*&
     &               YPOLD(L))+((R(4,3)*STAGES(L,3)+R(11,3)*&
     &               STAGES(L,1))+R(7,3)*STAGES(L,6))
  260    CONTINUE
!
!  Coefficient of SIGMA**2
!
         DO 280 L = 1_id, NWANT
            P(L) = R(5,2)*STAGES(L,4) + ((R(6,2)*STAGES(L,5)+&
     &             R(8,2)*YP(L))+R(1,2)*YPOLD(L)) +&
     &             ((R(3,2)*STAGES(L,2)+R(9,2)*STAGES(L,7))+&
     &             R(10,2)*XSTAGE(L)) + ((R(4,2)*STAGES(L,3)+&
     &             R(11,2)*STAGES(L,1))+R(7,2)*STAGES(L,6))
  280    CONTINUE
!
!  Scale all the coefficients by the step size.
!
         DO 300 L = 1_id, NWANT*(MINTP-1)
            P(L) = HOLD*P(L)
  300    CONTINUE
!
      END IF
!
      RETURN
      END SUBROUTINE



      SUBROUTINE EVALI(Y,YP,P,TWANT,REQEST,NWANT,YWANT,YPWANT)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!*************************************************
!
!  Purpose:    Evaluation of an interpolating polynomial and/or its
!              first derivative.
!
!  Input:      Y(*), YP(*), P(NWANT,*), TWANT, REQEST, NWANT
!  Output:     YWANT(*), YPWANT(*)
!
!  Common:     Initializes:    none
!              Reads:          /RKCOM2/ HOLD, T
!                              /RKCOM4/ MINTP
!              Alters:         none
!
!  Comments:
!  =========
!  The interpolant is evaluated at TWANT to approximate the solution,
!  YWANT, and/or its first derivative there, YPWANT. Only the first
!  NWANT components of the answer are computed. There are three cases
!  that are indicated by the first character of REQEST:
!    REQEST(1:1) = `S' or `s'- compute approximate `S'olution only.
!                = `D' or `d'- compute approximate first `D'erivative
!                              of the solution only.
!                = `B' or `b'- compute `B'oth approximate solution and
!                              first derivative.
!  The coefficents of the polynomial are contained in Y(*), YP(*) and
!  P(NWANT,*).
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  TWANT
      INTEGER(id)           NWANT
      CHARACTER*(*)     REQEST
!     .. Array Arguments ..
      REAL(rd)  P(NWANT,*), Y(*), YP(*), YPWANT(*), YWANT(*)
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!     .. Local Scalars ..
      REAL(rd)  SIGMA
      INTEGER(id)           K, L
      CHARACTER         REQST1
!     .. Executable Statements ..
!
!  Evaluate the interpolating polynomial of degree MINTP in terms of the
!  shifted and scaled independent variable SIGMA.
!
      SIGMA = (TWANT-T)/HOLD
!
      REQST1 = REQEST(1:1)
      IF (REQST1.EQ.'S' .OR. REQST1.EQ.'s' .OR. &
     &    REQST1.EQ.'B' .OR. REQST1.EQ.'b') THEN
!
         DO 20 L = 1_id, NWANT
            YWANT(L) = P(L,MINTP-1)*SIGMA
   20    CONTINUE
         DO 60 K = MINTP - 2_id, 1_id, -1_id
            DO 40 L = 1_id, NWANT
               YWANT(L) = (YWANT(L)+P(L,K))*SIGMA
   40       CONTINUE
   60    CONTINUE
         DO 80 L = 1_id, NWANT
            YWANT(L) = (YWANT(L)+HOLD*YP(L))*SIGMA + Y(L)
   80    CONTINUE
      END IF
!
!  Evaluate the derivative of the interpolating polynomial.
!
      IF (REQST1.EQ.'D' .OR. REQST1.EQ.'d' .OR. &
     &    REQST1.EQ.'B' .OR. REQST1.EQ.'b') THEN
!
!  The derivative of the interpolating polynomial with respect to TWANT 
!  is the derivative with respect to S divided by HOLD.
!
         DO 100 L = 1_id, NWANT
            YPWANT(L) = MINTP*P(L,MINTP-1)*SIGMA
  100    CONTINUE
         DO 140 K = MINTP - 1_id, 2_id, -1_id
            DO 120 L = 1_id, NWANT
               YPWANT(L) = (YPWANT(L)+K*P(L,K-1_id))*SIGMA
  120       CONTINUE
  140    CONTINUE
         DO 160 L = 1_id, NWANT
            YPWANT(L) = (YPWANT(L)+HOLD*YP(L))/HOLD
  160    CONTINUE
      END IF
!
      RETURN
      END SUBROUTINE


      SUBROUTINE RKMSG(IER,SRNAME,NREC,FLAG)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:      To process error messages and terminate the program
!                in the event of a "catastrophic" failure.
!
!  Input:        IER, SRNAME, NREC
!  Output:       FLAG
!
!  Common:       Initializes:    none
!                Reads:          /RKCOM7/ OUTCH
!                                /RKCOM8/ MSG, UTASK
!                                /RKCOM9/ REC
!                Alters:         none
!
!  Comments:
!  =========
!  The output variable FLAG is assigned the value of the input variable IER.
!
!  IER = -2  reports a successful call of the subroutine SRNAME and
!            indicates that special action is to be taken elsewhere
!            in the suite.  FLAG is set and a return is effected.
!
!  IER = 1   reports a successful call of the subroutine SRNAME.  FLAG
!            is set and a return is effected.
!
!  1 < IER < 911 and MSG = .TRUE.: a message of NREC records contained in
!            the array REC(*) is written to the standard output channel, 
!            OUTCH.  FLAG is set and a return is effected.
!
!  IER = 911 reports a "catastrophic" error was detected in SRNAME.  A
!            message is written to OUTCH regardless of the value of MSG and
!            normally the execution of the program is terminated.  The
!            execution is not terminated if the error is the result of an
!            indirect call to CT, RESET, or INTRP through UT (UTASK = .TRUE.).
!            Termination can be prevented by using the subroutine SOFTFL.
!
!  IER = 912 reports that a "catastrophic" error was detected earlier and
!            termination was prevented, but the user has failed to take
!            appropriate remedial action.  Execution is terminated.
use kindtype
implicit none
!     .. Scalar Arguments ..
      INTEGER(id)           FLAG, IER, NREC
      CHARACTER*(*)     SRNAME
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Common Block for Integrator Options ..
      LOGICAL(lgt)           MSG, UTASK
      COMMON /RKCOM8/   MSG, UTASK
      SAVE   /RKCOM8/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      INTEGER(id)           PLUS1
      LOGICAL(lgt)           ASK, TELL
      PARAMETER         (PLUS1=1,ASK=.TRUE.,TELL=.FALSE.)
!     .. Local Scalars ..
      INTEGER(id)           I
      LOGICAL(lgt)           BADERR, OK, ON, UTCALL
!     .. External Subroutines ..
 !     EXTERNAL          CHKFL, RKSIT, SOFTFL
!     .. Executable Statements ..
!
!  Check where the call came from - if it is an indirect call from UT,
!  the run is not STOPped.
      UTCALL = (SRNAME.EQ.'RESET' .OR. SRNAME.EQ.'CT' .OR.&
     &          SRNAME.EQ.'INTRP') .AND. UTASK
!
!  Check if can continue with integrator.
      OK = (SRNAME.EQ.'CT' .OR. SRNAME.EQ.'UT') .AND.&
     &     (IER.EQ.2_id .OR. IER.EQ.3_id .OR. IER.EQ.4_id)
!
!  Check if program termination has been overridden.
      CALL SOFTFL(ASK,ON)
!
      IF ((MSG.AND.IER.GT.PLUS1) .OR. IER.GE.911_id) THEN
         WRITE (OUTCH,'(/A)') ' **'
         WRITE (OUTCH,'(A)') (REC(I),I=1_id,NREC)
         IF (IER.GE.911_id) THEN
            WRITE (OUTCH,'(A/A,A,A/A/)') &
     &' **',&
     &' ** Catastrophic error detected in ', SRNAME, '.', &
     &' **'
            IF ((.NOT.(UTCALL.OR.ON).AND.IER.EQ.911_id) .OR.&
     &          IER.EQ.912_id) THEN
               WRITE (OUTCH,'(A/A/A)') &
     &' **',&
     &' ** Execution of your program is being terminated.',&
     &' **'
               STOP
            END IF
         ELSE IF (OK) THEN
            WRITE (OUTCH,'(A/A,A,A,I2,A/A/A)') &
     &' **',&
     &' ** Warning from routine ', SRNAME, ' with flag set ',IER, '.',&
     &' ** You can continue integrating this problem.',&
     &' **'
         ELSE
            WRITE (OUTCH,'(A/A,A,A,I2,A/A/A)') &
     &' **',&
     &' ** Warning from routine ', SRNAME, ' with flag set ',IER, '.',&
     &' ** You cannot continue integrating this problem.', &
     &' **'
         END IF
      END IF
      DO 20 I = NREC + 1_id, 10_id
         REC(I) = ' '
   20 CONTINUE
      FLAG = IER
!
!  TELL RKSIT the status of the routine associated with SRNAME
      CALL RKSIT(TELL,SRNAME,FLAG)
!
!  Indicate that a catastrophic error has been detected
      BADERR = FLAG .GE. 911_id
      CALL CHKFL(TELL,BADERR)
!
      RETURN
!
      END SUBROUTINE


      SUBROUTINE RKSIT(ASK,SRNAME,STATE)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:      To save or enquire about the status of each
!                subprogram in the suite.
!
!  Input:        ASK, SRNAME
!  Input/output: STATE
!
!
!  Comments:
!  =========
!  SRNAME indicates which routine is of interest in the call to RKSIT.
!
!  If ASK=.FALSE., then the value of STATE (which as used is the error
!  flag) for the routine SRNAME is saved internally. This value of STATE
!  is usually positive.  There are two exceptions:
!     1. SRNAME='SETUP' and STATE=-1 indicates a completely new problem,
!        so all SAVEd states are cleared.
!     2. STATE=-2 is used by some routines in the suite to indicate
!        circumstances which require special action.
!
!  If ASK=.TRUE., then RKSIT first checks to see if there were any
!  catastrophic errors, that is, a SAVEd state has value 911. This should
!  happen only when the user has overridden program termination in the event
!  of catastrophic failures from routines in the package but has failed to
!  take appropriate action. If this is the case, then RKSIT returns a value
!  of STATE = 911 which forces a termination of execution inside RKMSG. If 
!  no catastrophic errors are flagged, then STATE returns the saved state 
!  value for the routine specified by SRNAME.
use kindtype
implicit none
!     .. Scalar Arguments ..
      INTEGER(id)           STATE
      LOGICAL(lgt)           ASK
      CHARACTER*(*)     SRNAME
!     .. Parameters ..
      INTEGER(id)           STATES, MINUS1
      PARAMETER         (STATES=7_id,MINUS1=-1_id)
!     .. Local Scalars ..
      INTEGER(id)           I, NAME
!     .. Local Arrays ..
      INTEGER(id)           SVSTA(STATES)
!     .. Save statement ..
      SAVE              SVSTA
!     .. Data statements ..
      DATA              SVSTA/STATES*MINUS1/
!     .. Executable Statements ..
!
      IF (SRNAME.EQ.'SETUP') THEN
         NAME = 1_id
      ELSE IF (SRNAME.EQ.'UT') THEN
         NAME = 2_id
      ELSE IF (SRNAME.EQ.'STAT') THEN
         NAME = 3_id
      ELSE IF (SRNAME.EQ.'GLBERR') THEN
         NAME = 4_id
      ELSE IF (SRNAME.EQ.'CT') THEN
         NAME = 5_id
      ELSE IF (SRNAME.EQ.'INTRP') THEN
         NAME = 6_id
      ELSE IF (SRNAME.EQ.'RESET') THEN
         NAME = 7_id
      ELSE
         NAME = 0_id
      END IF
!
!  (Re)initialize if SETUP is telling RKSIT to do so.
      IF (.NOT.ASK .AND. NAME.EQ.1_id .AND. STATE.EQ.MINUS1) THEN
         DO 20 I = 1_id, STATES
            SVSTA(I) = MINUS1
   20    CONTINUE
         GO TO 60
      END IF
!
!  Check for 911 on exit from a previous call.
      IF (ASK) THEN
         DO 40 I = 1_id, STATES
            IF (SVSTA(I).EQ.911_id) THEN
               STATE = 911_id
               GO TO 60
            END IF
   40    CONTINUE
      END IF
!
      IF (ASK) THEN
         STATE = SVSTA(NAME)
      ELSE
         SVSTA(NAME) = STATE
      END IF
!
   60 CONTINUE
!
      RETURN
      END SUBROUTINE



      SUBROUTINE TRUERR(F,NEQ,Y,TOL,WEIGHT,ZY,ZYP,ZERROR,ZYNEW,ZERRES,&
     &                  ZSTAGE,IER)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:      Compute a running RMS measure of the true (global) error
!                for a general Runge-Kutta pair.
!
!
!  Input:        NEQ, Y(*), TOL, WEIGHT(*),
!  Input/output: ZY(*), ZYP(*), ZERROR(*)
!  Workspace:    ZYNEW(*), ZERRES(*), ZSTAGE(NEQ,*)
!  Output:       IER
!  External:     F
!
!  Common:       Initializes:    none
!                Reads:          /RKCOM2/ T, HOLD
!                                /RKCOM5/ TOOSML, ORDER, NSEC
!                                /RKCOM7/ TINY
!                Alters:         /RKCOM6/ MAXERR, LOCMAX, GNFCN
!
!  Comments:
!  =========
!  A secondary integration is performed using a fraction of the step size 
!  of the primary integration. ZY(*) and ZYP(*) are the approximate solution
!  and first derivative of this secondary integration. ZERRES(*) contains the 
!  error estimates for the secondary integration. ZYNEW(*) and ZSTAGE(*,*) are
!  workspace for taking a step. The error assessment is computed using the
!  difference of the primary and secondary solutions at the primary
!  integration points as an estimate of the true error there.  The weights 
!  used are those of the error test of the primary integration. This error 
!  assessment is maintained in the vector ZERROR(*).  MAXERR and LOCMAX 
!  contain the maximum contribution to the assessment and its location,
!  respectively.  The number of calls to F is counted by GNFCN.
use kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  TOL
      INTEGER(id)           IER, NEQ
!     .. Array Arguments ..
      REAL(rd)  WEIGHT(*), Y(*), ZERRES(*), ZERROR(*),&
     &                  ZSTAGE(NEQ,*), ZY(*), ZYNEW(*), ZYP(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Global Error Assessment ..
      REAL(rd)  MAXERR, LOCMAX
      INTEGER(id)           GNFCN, PRZSTG, PRZY, PRZYP, PRZERS, PRZERR,&
     &                  PRZYNU
      LOGICAL(lgt)           ERASON, ERASFL
      COMMON /RKCOM6/   MAXERR, LOCMAX, GNFCN, PRZSTG, PRZY, PRZYP,&
     &                  PRZERS, PRZERR, PRZYNU, ERASON, ERASFL
      SAVE   /RKCOM6/
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Parameters ..
      REAL(rd)  PT1, TEN, DUMMY
      PARAMETER         (PT1=0.1_rd,TEN=10.0_rd,DUMMY=1.0_rd)
!     .. Local Scalars ..
      REAL(rd)  DIFF, ERRMAX, HMIN, HSEC, MXERLC, TSEC, ZLERR,&
     &                  ZTEST1, ZTEST2
      INTEGER(id)           ISTEP, L, LEVEL
      LOGICAL(lgt)           LDUMMY, MAIN
!     .. Local Arrays ..
      REAL(rd)  DUMARR(1)
!     .. External Subroutines ..
!      EXTERNAL          STEP
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, DBLE, MAX
!     .. Executable Statements ..
      TSEC = T - HOLD
      HSEC = HOLD/DBLE(NSEC)
      HMIN = MAX(TINY,TOOSML*MAX(ABS(TSEC),ABS(T)))
      IF (ABS(HSEC).LT.HMIN) THEN
         IER = 6_id
         GO TO 120
      END IF
      ZTEST1 = TOL/DBLE(NSEC)
      ZTEST2 = TOL/TEN
      LEVEL = 0_id
!
!  The subroutine STEP is used to take a step.  In its use in the primary
!  integration provision is made for getting on scale in the first step.
!  In this situation only the subroutine might reduce the step size.  By
!  setting MAIN = .FALSE., the subroutine will take a step of the size input.
!  In this use of the subroutine, all items of the call list appearing after
!  MAIN are dummy variables.
!
!  Perform secondary integration.
      MAIN = .FALSE.
      LDUMMY = .FALSE.
      DO 60 ISTEP = 1_id, NSEC
!
!  Take a step.
         CALL STEP(F,NEQ,TSEC,ZY,ZYP,ZSTAGE,ZTEST1,HSEC,WEIGHT,ZYNEW,&
     &             ZERRES,ZLERR,MAIN,DUMMY,DUMARR,LDUMMY)
!
!  The primary integration is using a step size of HUSED and the secondary
!  integration is using the smaller step size HSEC = HUSED/NSEC.  If steps
!  of this size were taken from the same starting point and the asymptotic
!  behavior were evident, the smaller step size would result in a local error
!  that is considerably smaller, namely by a factor of 1/(NSEC**(ORDER+1)).
!  If the two approximate solutions are close and TOLR is neither too large nor
!  too small, this should be approximately true.  The step size is chosen in
!  the primary integration so that the local error ERR is no larger than TOLR.
!  The local error, ZLERR, of the secondary integration is compared to TOLR in
!  an attempt to diagnose a secondary integration that is not rather more
!  accurate than the primary integration.
!
         IF (ZLERR.GE.ZTEST1) THEN
            LEVEL = 2_id
         ELSE IF (ZLERR.GT.ZTEST2) THEN
            LEVEL = LEVEL + 1_id
         END IF
         IF (LEVEL.GE.2) THEN
            IER = 6_id
            GO TO 120
         END IF
!
!  Advance TSEC and the dependent variables ZY(*) and ZYP(*).
         TSEC = T - DBLE(NSEC-ISTEP)*HSEC
         DO 20 L = 1_id, NEQ
            ZY(L) = ZYNEW(L)
   20    CONTINUE
!
         IF (FSAL) THEN
!
!  When FSAL = .TRUE., the derivative ZYP(*) is the last stage of the step.
            DO 40 L = 1_id, NEQ
               ZYP(L) = ZSTAGE(L,LSTSTG)
   40       CONTINUE
         ELSE
!
!  Call F to evaluate ZYP(*).
            CALL F(TSEC,ZY,ZYP)
            GNFCN = GNFCN + 1_id
         END IF
!
   60 CONTINUE
!
!  Update the maximum error seen, MAXERR, and its location, LOCMAX.
!  Use local variables ERRMAX and MXERLC.
!
      ERRMAX = MAXERR
      MXERLC = LOCMAX
      DO 80 L = 1_id, NEQ
         DIFF = ABS(ZY(L)-Y(L))/WEIGHT(L)
         IF (DIFF.GT.ERRMAX) THEN
            ERRMAX = DIFF
            MXERLC = T
         END IF
   80 CONTINUE
!
!  If the global error is greater than 0.1_rd, the solutions have diverged so
!  far that comparing them may not provide a reliable estimate of the global
!  error. The test is made before ZERROR(*) and MAXERR, LCMXER are updated so
!  that on a failure, they refer to the last reliable results.
!
      IF (ERRMAX.GT.PT1) THEN
         IER = 6_id
         GO TO 120
      ELSE
         MAXERR = ERRMAX
         LOCMAX = MXERLC
         DO 100 L = 1_id, NEQ
            DIFF = ABS(ZY(L)-Y(L))/WEIGHT(L)
            ZERROR(L) = ZERROR(L) + DIFF**2
  100    CONTINUE
         IER = 1_id
      END IF
!
!  Exit point for TRUERR
  120 CONTINUE
!
      RETURN
      END SUBROUTINE



      SUBROUTINE STEP(F,NEQ,TNOW,Y,YP,STAGES,TOL,HTRY,WEIGHT,YNEW,&
     &                ERREST,ERR,MAIN,HMIN,THRES,PHASE2)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:      To compute a step of an explicit Runge-Kutta
!                method and estimate the local error of the step.
!
!  Input:        NEQ, TNOW, Y(*), YP(*), TOL, MAIN, HMIN, THRES(*)
!  Input/output: HTRY, PHASE2, LAST, WEIGHT(*)
!  Output:       STAGES(NEQ,*), YNEW(*), ERREST(*), ERR
!
!  Common:       Initializes:    none
!                Reads:          /RKCOM1/ TND
!                                /RKCOM2/ LAST
!                                /RKCOM4/ A, B, C, BHAT, PTR, NSTAGE, METHD
!                                /RKCOM5/ FSAL
!                Alters:         /RKCOM2/ NFCN, LAST
!                                /RKCOM6/ GNFCN
!
!  Comments:
!  =========
!  From an approximate solution Y(*) at TNOW and first derivative there,
!  YP(*) = F(TNOW,Y,YP), a step is taken to get an approximation YNEW(*)
!  at TNOW + HTRY. The Runge-Kutta method and how it is used are defined
!  by A, B, C, BHAT, PTR, NSTAGE, METHD and FSAL. Intermediate stages
!  of the method are stored in the array STAGES(NEQ,*). The error in
!  each solution component is estimated and returned in ERREST(*). A
!  weighted maximum norm of the local error, ERR, is formed. For some
!  methods an intermediate error estimate can be computed before completion
!  of the step (see routine STEPB); if the estimate is greater than the
!  specified tolerance TOL, the computation of the step is terminated.
!
!  When global error estimation is desired, two integrations are done.
!  The usual integration is referred to as the "primary", or "main",
!  integration (MAIN=.TRUE.).  For global error estimation another,
!  "secondary" integration (MAIN=.FALSE.) is carried out with a smaller
!  step size.  The weight vector WEIGHT(*) used in computing ERR is
!  determined by the main integration.  Thus this argument is output when
!  MAIN = .TRUE. and input when MAIN = .FALSE..
!
!  When taking the first step in an integration, the logical variable
!  PHASE2 may be input as .TRUE. and if the first step is the whole of
!  the range of integration, then LAST will be .TRUE.. When PHASE2=.TRUE.,
!  the first three stages are monitored to help assure that the step
!  size H is small enough for the integration to be stable and for the
!  estimate of the error of the step to be credible. Calls are made to
!  the subroutine STEPA for this purpose. If necessary, H will be
!  reduced in STEPA (and LAST altered accordingly) and the step retried
!  in STEP until an acceptable value is found.
!
!  In the primary integration the number of calls to F is counted by
!  NFCN, and in the secondary integration, by GNFCN.
use kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  ERR, HMIN, HTRY, TNOW, TOL
      INTEGER(id)           NEQ
      LOGICAL(lgt)           MAIN, PHASE2
!     .. Array Arguments ..
      REAL(rd)  ERREST(*), STAGES(NEQ,*), THRES(*), WEIGHT(*),&
     &                  Y(*), YNEW(*), YP(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Global Error Assessment ..
      REAL(rd)  MAXERR, LOCMAX
      INTEGER(id)           GNFCN, PRZSTG, PRZY, PRZYP, PRZERS, PRZERR,&
     &                  PRZYNU
      LOGICAL(lgt)           ERASON, ERASFL
      COMMON /RKCOM6/   MAXERR, LOCMAX, GNFCN, PRZSTG, PRZY, PRZYP,&
     &                  PRZERS, PRZERR, PRZYNU, ERASON, ERASFL
      SAVE   /RKCOM6/
!     .. Parameters ..
      REAL(rd)  ZERO, HALF, ONE
      PARAMETER         (ZERO=0.0_rd,HALF=0.5_rd,ONE=1.0_rd)
!     .. Local Scalars ..
      REAL(rd)  AVGY, TSTG
      INTEGER(id)           I, J, L
      LOGICAL(lgt)           CUTBAK
!     .. External Subroutines ..
 !     EXTERNAL          STEPA, STEPB
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, MAX, SIGN
!     .. Executable Statements ..
!
!  Many of the following loops over L = 1, NEQ have constant array values
!  inside. The code is written with clarity in mind.  Any optimizing
!  compiler will identify these occurrences and take appropriate action.
!  A check for zero multipliers has been included so as to prevent
!  needless computation resulting from the storing of zero coefficients
!  in the arrays for the sake of clarity.  The array ERREST(*) is used
!  for working storage in this computation.
!
   20 CONTINUE

      IF (MAIN) THEN
         IF (PHASE2) THEN
!
!  Initialize weights for measuring the local error.
            DO 40 L = 1_id, NEQ
               WEIGHT(L) = MAX(THRES(L),ABS(Y(L)))
   40       CONTINUE
         END IF
      END IF
!

      DO 140 I = 2_id, NSTAGE
       
         DO 100 J = 1_id, I - 1_id
            IF (J.EQ.1) THEN
               DO 60 L = 1_id, NEQ
                  ERREST(L) = A(I,1)*YP(L)

   60          CONTINUE
            ELSE
               IF (A(I,J).NE.ZERO) THEN
                  DO 80 L = 1_id, NEQ
                     ERREST(L) = ERREST(L) + A(I,J)*STAGES(L,PTR(J))
   80             CONTINUE
               END IF
            END IF
  100    CONTINUE
         DO 120 L = 1_id, NEQ
            YNEW(L) = Y(L) + HTRY*ERREST(L)
  120    CONTINUE

!
!  METHD = 2 is special in that an estimate of the local error can be
!  formed before the step is completed.  If the step is a failure,
!  return immediately.  Otherwise, complete the step and compute a more
!  accurate error estimate.
         IF (METHD.EQ.2_id .AND. I.EQ.7_id) THEN
            CALL STEPB(NEQ,Y,YP,HTRY,YNEW,STAGES,THRES,ERR,MAIN,WEIGHT)
            IF (ERR.GT.TOL) RETURN
         END IF
!
         TSTG = TNOW + C(I)*HTRY
         IF (MAIN .AND. LAST .AND. C(I).EQ.ONE) TSTG = TND
         CALL F(TSTG,YNEW,STAGES(1_id,PTR(I)))
!
!  Increment the counter for the number of function evaluations
!  depending on whether the primary or secondary integration is taking
!  place.
         IF (MAIN) THEN
            NFCN = NFCN + 1_id
         ELSE
            GNFCN = GNFCN + 1_id
         END IF
!
!----------------------------------------------------------------------
!  When PHASE2 is .TRUE. we are in the second phase of the automatic
!  selection of the initial step size.  The results of the first three
!  stages are monitored in the subroutine STEPA for evidence that H is
!  too large -- instability and/or an unreliable estimate of the error
!  of the step is then possible.  When the subroutine believes H to be
!  too large, it returns CUTBAK = .TRUE. and a suitably reduced H for
!  another try.
!
         IF (MAIN) THEN
            IF (PHASE2) THEN
               IF (I.LE.3_id .AND. ABS(HTRY).GT.HMIN) THEN
                  CALL STEPA(TNOW,Y,YP,TSTG,YNEW,STAGES(1_id,PTR(I)),&
     &                       HTRY,WEIGHT,CUTBAK)

                  IF (CUTBAK) THEN
                     LAST = .FALSE.
!
!  Make sure that STEPA does not reduce the step size below the
!  minimum. If it does, reset H to HMIN and deactivate PHASE2.
                     IF (ABS(HTRY).LE.HMIN) THEN
                        HTRY = SIGN(HMIN,HTRY)
                        PHASE2 = .FALSE.
                     END IF
                     GO TO 20
                  END IF
               END IF
            END IF
         END IF
!----------------------------------------------------------------------
!
  140 CONTINUE
!
!  Some formulas are constructed so that the last stage represents
!  the result of the step (FSAL=.TRUE.), hence if the step is acceptable,
!  it will be the first stage for the next step. When FSAL=.FALSE., we
!  have to complete the computation of the step.
!
      IF (.NOT.FSAL) THEN
         DO 200 I = 1_id, NSTAGE
            IF (I.EQ.1_id) THEN
               DO 160 L = 1_id, NEQ
                  ERREST(L) = BHAT(1)*YP(L)
  160          CONTINUE
            ELSE
               IF (BHAT(I).NE.ZERO) THEN
                  DO 180 L = 1_id, NEQ
                     ERREST(L) = ERREST(L) + BHAT(I)*STAGES(L,PTR(I))
  180             CONTINUE
               END IF
            END IF
  200    CONTINUE
         DO 220 L = 1, NEQ
            YNEW(L) = Y(L) + HTRY*ERREST(L)
  220    CONTINUE
      END IF
!
!  Form an estimate of the error in the lower order formula by comparing
!  it to the higher order formula of the pair. ERREST(*) has been used
!  as working storage above.  The higher order approximation has been
!  formed as YNEW(*) = Y(*) + HTRY*ERREST(*) where ERREST(*) is a linear
!  combination of the stages of the formula. The lower order result also
!  has the form Y(*) plus HTRY times a different linear combination of
!  the stages. Hence, this different linear combination of stages for
!  the lower order formula can just be subtracted from the combination
!  stored in ERREST(*) to produce the errors. The result is then
!  multiplied by HTRY to obtain the error estimate.
!
      DO 280 I = 1_id, NSTAGE
         IF (I.EQ.1_id .AND. B(1).NE.ZERO) THEN
            DO 240 L = 1_id, NEQ
               ERREST(L) = ERREST(L) - B(1)*YP(L)
  240       CONTINUE
         ELSE
            IF (B(I).NE.ZERO) THEN
               DO 260 L = 1_id, NEQ
                  ERREST(L) = ERREST(L) - B(I)*STAGES(L,PTR(I))
  260          CONTINUE
            END IF
         END IF
  280 CONTINUE
      DO 300 L = 1_id, NEQ
         ERREST(L) = HTRY*ERREST(L)
  300 CONTINUE
!
!  The error in a solution component is measured relative to a weight
!  that is the larger of a threshold and the size of the solution over
!  the step.  Using the magnitude of a solution component at both ends
!  of the step in the definition of "size" increases the robustness of
!  the test. When global error estimation is specified, the weight
!  vector WEIGHT(*) is defined by the primary integration and is then
!  used in the secondary integration.
!
      IF (MAIN) THEN
         DO 320 L = 1_id, NEQcontrol
            AVGY = HALF*(ABS(Y(L))+ABS(YNEW(L)))
            WEIGHT(L) = MAX(AVGY,THRES(L))
  320    CONTINUE
      END IF
!
      ERR = ZERO
      DO 340 L = 1_id,NEQcontrol
         ERR = MAX(ERR,ABS(ERREST(L)/WEIGHT(L)))
  340 CONTINUE   

      RETURN
      END SUBROUTINE


      SUBROUTINE STEPA(TNOW,Y,YP,TSTG,YSTG,YPSTG,HTRY,WEIGHT,CUTBAK)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:      To calculate an "on-scale" step size for phase 2 of
!                the initial step size computation.
!
!  Input:        TNOW, Y(*), YP(*), TSTG, YSTG(*), YPSTG(*)
!  Input/output: HTRY, WEIGHT
!  Output:       CUTBAK
!
!  Common:       Initializes:    none
!                Reads:          /RKCOM1/ TND, NEQ
!                                /RKCOM5/ STBRAD, RS1, RS4
!                                /RKCOM7/ RNDOFF
!                Alters:         none
!
!  Comments:
!  =========
!  This subroutine is used during the first three stages of the first step.
!  A Lipschitz constant L for the differential equation in autonomous form
!  is approximated, and the product abs(HTRY)*L is compared to an approximate
!  radius, STBRAD, of the stability region of the method. The step size is 
!  reduced as necessary, within a range specified by the step size control 
!  parameters RS1 and RS4, to assure stability and give some confidence in 
!  the error estimator.  If HTRY is reduced, CUTBAK is set .TRUE..
!
!  Y(*) and YP(*) contain the solution and its derivative at TNOW and
!  similarly YSTG(*) and YPSTG(*) contain approximations at TSTG.
!
!  Normally the weights used in the control of the error depend on the
!  size of the solution at the beginning and at the end of the step, but
!  at this time we do not have a solution at the end of the step.  Each
!  stage YSTG(*) of the Runge - Kutta process represents a low order
!  approximation to the solution at TSTG.  Because the initial value of
!  WEIGHT(*) provided in the first phase of the scheme is based only on
!  the solution at T and THRES(*), it is continually updated in STEPA to
!  account for the size of the solution throughout the step as revealed
!  by the intermediate stages YSTG(*). Inside this subroutine only, the
!  differential equation is converted to autonomous form. After the
!  conversion, the end of the interval of integration, TND, is used
!  to define a suitable weight for the independent variable.
use kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  HTRY, TNOW, TSTG
      LOGICAL(lgt)           CUTBAK
!     .. Array Arguments ..
      REAL(rd)  WEIGHT(*), Y(*), YP(*), YPSTG(*), YSTG(*)
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Parameters ..
      REAL(rd)  ZERO
      PARAMETER         (ZERO=0.0_rd)
!     .. Local Scalars ..
      REAL(rd)  ARGDIF, FDIFF, SCL, TDIFF, TWT, WT, YNRM, YSTGNM
      INTEGER(id)           L
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, MAX, MIN
!     .. Executable Statements ..
!
!  Update the weights to account for the current intermediate solution
!  approximation YSTG(*).  Compute the sizes of Y(*) and YSTG(*) in the
!  new norm.  The size of the Lipschitz constant is assessed by a difference
!  in the arguments Y(*), YSTG(*) and a difference in the function evaluated
!  at these arguments.
!
      YNRM = ZERO
      YSTGNM = ZERO
      ARGDIF = ZERO
      FDIFF = ZERO
      DO 20 L = 1_id, neqcontrol
         WT = MAX(WEIGHT(L),ABS(YSTG(L)))
         WEIGHT(L) = WT
         YNRM = MAX(YNRM,ABS(Y(L))/WT)
         YSTGNM = MAX(YSTGNM,ABS(YSTG(L))/WT)
         ARGDIF = MAX(ARGDIF,ABS(YSTG(L)-Y(L))/WT)
         FDIFF = MAX(FDIFF,ABS(YPSTG(L)-YP(L))/WT)
   20 CONTINUE
!
!  The transformation of the equation to autonomous form is done
!  implicitly.  The difference of the arguments must take into account
!  the difference between the values of the independent variable T and
!  TSTG. The difference of the corresponding component of the function
!  is zero because of the way the standard transformation is done.
!
      TDIFF = TSTG - TNOW
      TWT = ABS(TND-TNOW)
      YNRM = MAX(YNRM,ABS(TNOW)/TWT)
      YSTGNM = MAX(YSTGNM,ABS(TSTG)/TWT)
      ARGDIF = MAX(ARGDIF,ABS(TDIFF)/TWT)
!
!  The ratio FDIFF/ARGDIF is a lower bound for, and an approximation to, a
!  Lipschitz constant L for the differential equation written in autonomous
!  form.  First we must ask if the difference ARGDIF is significant in the 
!  precision available.  If it appears to be, we insist that abs(HTRY)*L be 
!  less than an approximate radius, STBRAD, of the stability region of the
!  method.  This is more stringent than necessary for stability, possibly a
!  lot more stringent, but the aim is to get an HTRY small enough that the
!  error estimate for the step is credible.  The reduction is required to be
!  at least as much as the step control parameter RS1. It is necessary to 
!  limit the reduction of HTRY at any one time because we may be misled in 
!  the size of the reduction that is appropriate due to nonlinearity of the 
!  differential equation and to inaccurate weights caused by HTRY much too 
!  large.  The reduction is not permitted to be more than the step control 
!  parameter RS4.
!
      CUTBAK = .FALSE.
      IF (ARGDIF.GT.RNDOFF*MAX(YNRM,YSTGNM)) THEN
         IF ((ABS(HTRY)*FDIFF).GT.(STBRAD*ARGDIF)) THEN
            SCL = (STBRAD*ARGDIF)/(ABS(HTRY)*FDIFF)
            SCL = MIN(SCL,RS1)
            SCL = MAX(SCL,RS4)
            HTRY = SCL*HTRY
            CUTBAK = .TRUE.
         END IF
      END IF
!
      RETURN
      END SUBROUTINE



      SUBROUTINE STEPB(NEQ,Y,YP,H,YNEW,STAGES,THRES,ERR,MAIN,WEIGHT)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:      To compute an error estimate for METHD = 2 prior
!                to completing the step.
!
!  Input:        NEQ, Y(*), YP(*), H, STAGES(NEQ,*), THRES(*), MAIN,
!                WEIGHT(*)
!  Output:       ERR
!
!  Common:       Initializes:    none
!                Reads:          /RKCOM4/ E, PTR
!                Alters:         none
!
!  Comments:
!  =========
!  If global error assessment is taking place, then MAIN = .FALSE. and
!  the weight vector generated by the primary integration is used.  The
!  error estimate is a linear combination (with coefficients in E(*))
!  of the stages stored in STAGES(*,*) (located by PTR(*)).
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  ERR, H
      INTEGER(id)           NEQ
      LOGICAL(lgt)           MAIN
!     .. Array Arguments ..
      REAL(rd)  STAGES(NEQ,*), THRES(*), WEIGHT(*), Y(*),&
     &                  YNEW(*), YP(*)
!     .. Common Block to hold Formula Definitions ..
      REAL(rd)  A(13,13), B(13), C(13), BHAT(13), R(11,6),&
     &                  E(7)
      INTEGER(id)           PTR(13), NSTAGE, METHD, MINTP
      LOGICAL(lgt)           INTP
      COMMON /RKCOM4/   A, B, C, BHAT, R, E, PTR, NSTAGE, METHD,&
     &                  MINTP, INTP
      SAVE   /RKCOM4/
!     .. Parameters ..
      REAL(rd)  ZERO, HALF
      PARAMETER         (ZERO=0.0_rd,HALF=0.5_rd)
!     .. Local Scalars ..
      REAL(rd)  AVGY, SUM, WT
      INTEGER(id)           INDEX, L
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, MAX
!     .. Executable Statements ..
!

      ERR = ZERO

      DO 40 L = 1_id, neqcontrol
!
!  Estimate the local error of component L. The coding makes use of
!  E(2) = 0.0_rd and E(7) = 0.0_rd.
!
         SUM = E(1)*YP(L)
         DO 20 INDEX = 3_id, 6_id
            SUM = SUM + E(INDEX)*STAGES(L,PTR(INDEX))
   20    CONTINUE
!
!  The local error is H*SUM.  A weighted maximum norm of SUM is formed 
!  and then the factor of H is taken into account.
!  
         IF (MAIN) THEN
            AVGY = HALF*(ABS(Y(L))+ABS(YNEW(L)))
            WT = MAX(AVGY,THRES(L))
         ELSE
            WT = WEIGHT(L)
         END IF
!
     !    if((abs(sum/wt)).gt.1.0_rd) write(*,*) 'stepb     ',l,abs(sum/wt)

         ERR = MAX(ERR,ABS(SUM/WT))
   40 CONTINUE
      ERR = ABS(H)*ERR

      RETURN
    END SUBROUTINE



      SUBROUTINE STIFF(F,HAVG,JFLSTP,TOOMCH,MAXFCN,WORK,IER,NREC)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:      Diagnose stiffness.  This depends on two things: whether
!                the step size is being restricted on grounds of stability
!                and whether the integration to TND can be completed in no
!                more than MAXFCN function evaluations.
!
!  Input:        HAVG, TOOMCH, MAXFCN, WORK(*)
!  Input/output: JFLSTP
!  Output:       IER, NREC
!  Workspace:    WORK(*)
!  External:     F
!
!  Common:       Initializes:    /RKCOM9/ REC
!                Reads:          /RKCOM1/ TND, NEQN
!                                /RKCOM2/ T, H, NFCN, SVNFCN, OKSTP
!                                /RKCOM3/ PRY, PRYP, PRTHRS, PRWT, PRSCR,
!                                         PRSTGS, PRYOLD
!                                /RKCOM5/ COST
!                Alters:         /RKCOM2/ NFCN
!                                /RKCOM9/ REC
use kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  HAVG
      INTEGER(id)           IER, JFLSTP, MAXFCN, NREC
      LOGICAL(lgt)           TOOMCH
!     .. Array Arguments ..
      REAL(rd)  WORK(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Common Block for General Workspace Pointers ..
      INTEGER(id)           PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      COMMON /RKCOM3/   PRTHRS, PRERST, PRWT, PRYOLD, PRSCR, PRY, PRYP,&
     &                  PRSTGS, PRINTP, LNINTP
      SAVE   /RKCOM3/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Error Message ..
      CHARACTER*80      REC(10)
      COMMON /RKCOM9/   REC
      SAVE   /RKCOM9/
!     .. Parameters ..
      REAL(rd)  HALF
      PARAMETER         (HALF=0.5_rd)
!     .. Local Scalars ..
      REAL(rd)  AVGY, XTRAWK
      INTEGER(id)           L
      LOGICAL(lgt)           LOTSFL, STIF, UNSURE
!     .. External Subroutines ..
 !     EXTERNAL          STIFFA
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, DBLE, MAX, MOD
!     .. Executable Statements ..
!
      IF (MOD(OKSTP-10_id,40_id).EQ.0_id) THEN
         LOTSFL = JFLSTP .GE. 10_id
         JFLSTP = 0_id
      ELSE
         LOTSFL = .FALSE.
      END IF
!
!  If either too much work has been done or there are lots of failed steps,
!  test for stiffness.
!
      IF (TOOMCH .OR. LOTSFL) THEN
!
!  Regenerate weight vector
         DO 20 L = 1_id, NEQN
            AVGY = HALF*(ABS(WORK(PRY-1_id+L))+ABS(WORK(PRYOLD-1_id+L)))
            WORK(PRWT-1_id+L) = MAX(AVGY,WORK(PRTHRS-1_id+L))
   20    CONTINUE
!
!  STIFFA determines whether the problem is STIFF. In some circumstances it
!  is UNSURE.  The decision depends on two things: whether the step size is
!  being restricted on grounds of stability and whether the integration to
!  TND can be completed in no more than MAXFCN function evaluations.  The
!  last four arguments of STIFFA are vectors of length NEQN used for working
!  storage.  Some storage in WORK(*) reserved for the stages (there are a
!  minimum of three such vectors reserved for the METHDs implemented) and
!  the scratch vector starting at PRSCR are used for this purpose.
!
         CALL STIFFA(F,T,WORK(PRY),H,HAVG,TND,MAXFCN,WORK(PRWT),&
     &               WORK(PRYP),WORK(PRERST),UNSURE,STIF,WORK(PRSTGS),&
     &               WORK(PRSTGS+NEQN),WORK(PRSTGS+2_id*NEQN),WORK(PRSCR))
         IF (.NOT.UNSURE) THEN
            IF (STIF) THEN
!
!  Predict how much eXTRA WorK will be needed to reach TND.
               XTRAWK = (COST*ABS((TND-T)/HAVG))/DBLE(SVNFCN+NFCN)
               IER = 4_id
               WRITE (REC(NREC+1),'(A)')&
     &' ** Your problem has been diagnosed as stiff.  If the '
               WRITE (REC(NREC+2),'(A,D13.5)')&
     &' ** situation persists, it will cost roughly ', XTRAWK
               WRITE (REC(NREC+3),'(A)')&
     &' ** times as much to reach TEND as it has cost to reach TNOW.'
               WRITE (REC(NREC+4),'(A)')&
     &' ** You should probably change to a code intended for '
               WRITE (REC(NREC+5),'(A)')&
     &' ** stiff problems. '
               NREC = NREC + 5_id
            END IF
         END IF
      END IF
!
      RETURN
      END SUBROUTINE



      SUBROUTINE STIFFA(F,X,Y,HNOW,HAVG,XEND,MAXFCN,WT,FXY,V0,UNSURE,&
     &                  STIF,V1,V2,V3,VTEMP)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  External:     F
!  Input:        X, Y(*), HNOW, HAVG, XEND, MAXFCN, WT(*), FXY(*)
!  Input/Output  V0(*)
!  Output:       UNSURE, STIF
!  Workspace:    V1(*), V2(*), V3(*), VTEMP(*)
!
!  Common:       Initializes:    none
!                Reads:          /RKCOM1/ TND, NEQN
!                                /RKCOM5/ COST, STBRAD, TANANG
!                                /RKCOM7/ SQRRMC, CUBRMC
!                Alters:         none
!
!  STIFFA diagnoses stiffness for an explicit Runge-Kutta code.  When it
!  is called, either many step failures have been observed, or a lot of
!  work has been done.
!
!  The NEQ equations of the problem are defined by the subroutine F(X,Y,YP).
!  When STIFFA is called, the integration has reached X where the approximate
!  solution is Y(*).  The vector FXY(*) is defined by a call of F(X,Y,FXY).
!  It is an input argument because it is usually available from the integrator.
!
!  The last successful step was of size HNOW, and an average step size is
!  HAVG.  A weighted norm is used to measure the local error with the error
!  in solution component L divided by the positive weight WT(L) provided in 
!  the vector WT(*).
!
!  Explicit Runge - Kutta codes estimate the local error of Y(*) by
!  forming the difference of two approximate solutions.  This difference
!  must be provided in the vector V0(*).  When this difference is too
!  small to be significant, STIFFA will replace it with a "random" vector.
!
!  STIF is set .TRUE. when the average step size appears to be restricted
!  on grounds of stability.  In certain cases the variable UNSURE is set 
!  .TRUE.; the value of STIF is then not defined.
!
!  The stability region of the explicit Runge-Kutta formula is described
!  by quantities TANANG and STBRAD that are communicated by the setup routine
!  via COMMON.  Stability regions often change sharply near the imaginary
!  axis so that it is difficult to classify the stiffness of a problem with
!  eigenvalues of a local Jacobian that are "near" the imaginary axis.  For
!  this reason,  we consider only points Z in the upper left half complex
!  plane for which TAN( IMAG(Z)/( - RE(Z))) <= TANANG. Eigenvalues outside
!  this region are one reason for the code being UNSURE.  The stability
!  region is approximated by the intersection of a disk with this sector.
!  The radius of this disk is called STBRAD.
!
!  Working storage must be provided via the four vectors V1(*),V2(*),
!  V3(*),VTEMP(*).  These vectors must be of length at least NEQ.
use kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  HAVG, HNOW, X, XEND
      INTEGER(id)           MAXFCN
      LOGICAL(lgt)           STIF, UNSURE
!     .. Array Arguments ..
      REAL(rd)  FXY(*), V0(*), V1(*), V2(*), V3(*), VTEMP(*),&
     &                  WT(*), Y(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Formula Characterisitcs ..
      REAL(rd)  TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4
      INTEGER(id)           ORDER, LSTSTG, MAXTRY, NSEC
      LOGICAL(lgt)           FSAL
      COMMON /RKCOM5/   TOOSML, COST, SAFETY, EXPON, STBRAD, TANANG,&
     &                  RS, RS1, RS2, RS3, RS4, ORDER, LSTSTG, MAXTRY,&
     &                  NSEC, FSAL
      SAVE   /RKCOM5/
!     .. Common Block for Environment Parameters ..
      REAL(rd)  MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY
      INTEGER(id)           OUTCH
      COMMON /RKCOM7/   MCHEPS, DWARF, RNDOFF, SQRRMC, CUBRMC, TINY,&
     &                  OUTCH
      SAVE   /RKCOM7/
!     .. Parameters ..
      REAL(rd)  LARGE
      PARAMETER         (LARGE=1.0e+10_rd)
      REAL(rd)  ZERO, P001, P9, ONE, TWO, FIVE, FIFTH
      PARAMETER         (ZERO=0.0_rd,P001=0.001_rd,P9=0.9_rd,ONE=1.0_rd,&
     &                  TWO=2.0_rd,FIVE=5.0_rd,FIFTH=0.2_rd)
!     .. Local Scalars ..
      REAL(rd)  ALPHA1, ALPHA2, BETA1, BETA2, D1, D2, DET1,&
     &                  DET2, DIST, RES2, RHO, RHO2, ROLD, SCALE, V0NRM,&
     &                  V0V0, V0V1, V0V2, V1V1, V1V2, V1V3, V2V2, V2V3,&
     &                  V3NRM, V3V3, XTRFCN, YNRM
      INTEGER(id)           L, NTRY
      LOGICAL(lgt)           ROOTRE
!     .. Local Arrays ..
      REAL(rd)  R1(2), R2(2), ROOT1(2), ROOT2(2)
!     .. External Functions ..
 !     REAL(rd)  DOTPRD
 !     EXTERNAL          DOTPRD
!     .. External Subroutines ..
   !   EXTERNAL          STIFFB, STIFFC, STIFFD
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, MIN, SQRT
!     .. Executable Statements ..
!
!  If the current step size differs substantially from the average,
!  the problem is not stiff.
!
      IF (ABS(HNOW/HAVG).GT.FIVE .OR. ABS(HNOW/HAVG).LT.FIFTH) THEN
         STIF = .FALSE.
         UNSURE = .FALSE.
         RETURN
      ELSE
         UNSURE = .TRUE.
      END IF
!
!  The average step size is used to predict the cost in function evaluations
!  of finishing the integration to XEND.  If this cost is no more than MAXFCN,
!  the problem is declared not stiff: If the step size is being restricted on
!  grounds of stability, it will stay close to HAVG.  The prediction will
!  then be good, but the cost is too low to consider the problem stiff.  If
!  the step size is not close to HAVG, the problem is not stiff.  Either way
!  there is no point to testing for a step size restriction due to stability.
!
      XTRFCN = COST*ABS((XEND-X)/HAVG)
      IF (XTRFCN.LE.MAXFCN) THEN
         STIF = .FALSE.
         UNSURE = .FALSE.
         RETURN
      ELSE
         UNSURE = .TRUE.
      END IF
!
!  There have been many step failures or a lot of work has been done.  Now 
!  we must determine if this is due to the stability characteristics of the
!  formula.  This is done by calculating the dominant eigenvalues of the
!  local Jacobian and then testing whether HAVG corresponds to being on the
!  boundary of the stability region.
!
!  The size of Y(*) provides scale information needed to approximate
!  the Jacobian by differences.
!
      YNRM = SQRT(DOTPRD(Y,Y,WT,NEQN))
      SCALE = YNRM*SQRRMC
      IF (SCALE.EQ.ZERO) THEN
!
!  Degenerate case.  Y(*) is (almost) the zero vector so the scale is not 
!  defined.  The input vector V0(*) is the difference between Y(*) and a 
!  lower order approximation to the solution that is within the error 
!  tolerance.  When Y(*) vanishes, V0(*) is itself an acceptable approximate
!  solution, so we take SCALE from it, if this is possible.
!
         YNRM = SQRT(DOTPRD(V0,V0,WT,NEQN))
         SCALE = YNRM*SQRRMC
         IF (SCALE.EQ.ZERO) THEN
            UNSURE = .TRUE.
            RETURN
         END IF
      END IF
!
      V0V0 = DOTPRD(V0,V0,WT,NEQN)
      IF (V0V0.EQ.ZERO) THEN
!
!  Degenerate case.  V0(*) is (almost) the zero vector so cannot
!  be used to define a direction for an increment to Y(*).  Try a
!  "random" direction.
!
         DO 20 L = 1_id, NEQN
            V0(L) = ONE
   20    CONTINUE
         V0V0 = DOTPRD(V0,V0,WT,NEQN)
      END IF
      V0NRM = SQRT(V0V0)
      DO 40 L = 1_id, NEQN
         V0(L) = V0(L)/V0NRM
   40 CONTINUE
      V0V0 = ONE
!
!  Use a nonlinear power method to estimate the two dominant eigenvalues.
!  V0(*) is often very rich in the two associated eigenvectors.  For this 
!  reason the computation is organized with the expectation that a minimal 
!  number of iterations will suffice.  Indeed, it is necessary to recognize 
!  a kind of degeneracy when there is a dominant real eigenvalue.  The
!  subroutine STIFFB does this.  In the first try, NTRY = 1, a Rayleigh 
!  quotient for such an eigenvalue is initialized as ROLD.  After each 
!  iteration, REROOT computes a new Rayleigh quotient and tests whether the
!  two approximations agree to one tenth of one per cent and the eigenvalue,
!  eigenvector pair satisfy a stringent test on the residual.  ROOTRE = .TRUE.
!  signals that a single dominant real root has been found.
!
      NTRY = 1_id
   60 CONTINUE
!
      CALL STIFFD(V0,HAVG,X,Y,F,FXY,WT,SCALE,V0V0,V1,V1V1,VTEMP)
!
!  The quantity SQRT(V1V1/V0V0) is a lower bound for the product of HAVG
!  and a Lipschitz constant.  If it should be LARGE, stiffness is not
!  restricting the step size to the stability region.  The principle is
!  clear enough, but the real reason for this test is to recognize an
!  extremely inaccurate computation of V1V1 due to finite precision
!  arithmetic in certain degenerate circumstances.
!
      IF (SQRT(V1V1).GT.LARGE*SQRT(V0V0)) THEN
         UNSURE = .TRUE.
         RETURN
      END IF
!
      V0V1 = DOTPRD(V0,V1,WT,NEQN)
      IF (NTRY.EQ.1_id) THEN
         ROLD = V0V1/V0V0
!
!  This is the first Rayleigh quotient approximating the product of HAVG
!  and a dominant real eigenvalue.  If it should be very small, the
!  problem is not stiff.  It is important to test for this possibility so
!  as to prevent underflow and degeneracies in the subsequent iteration.
!
         IF (ABS(ROLD).LT.CUBRMC) THEN
            UNSURE = .FALSE.
            STIF = .FALSE.
            RETURN
         END IF
      ELSE
         CALL STIFFB(V1V1,V0V1,V0V0,ROLD,RHO,ROOT1,ROOT2,ROOTRE)
         IF (ROOTRE) GO TO 100
      END IF
      CALL STIFFD(V1,HAVG,X,Y,F,FXY,WT,SCALE,V1V1,V2,V2V2,VTEMP)
      V0V2 = DOTPRD(V0,V2,WT,NEQN)
      V1V2 = DOTPRD(V1,V2,WT,NEQN)
      CALL STIFFB(V2V2,V1V2,V1V1,ROLD,RHO,ROOT1,ROOT2,ROOTRE)
      IF (ROOTRE) GO TO 100
!
!  Fit a quadratic in the eigenvalue to the three successive iterates
!  V0(*),V1(*),V2(*) of the power method to get a first approximation to
!  a pair of eigenvalues.  A test made earlier in STIFFB implies that
!  the quantity DET1 here will not be too small.
!
      DET1 = V0V0*V1V1 - V0V1**2
      ALPHA1 = (-V0V0*V1V2+V0V1*V0V2)/DET1
      BETA1 = (V0V1*V1V2-V1V1*V0V2)/DET1
!
!  Iterate again to get V3, test again for degeneracy, and then fit a
!  quadratic to V1(*),V2(*),V3(*) to get a second approximation to a pair
!  of eigenvalues.
!
      CALL STIFFD(V2,HAVG,X,Y,F,FXY,WT,SCALE,V2V2,V3,V3V3,VTEMP)
      V1V3 = DOTPRD(V1,V3,WT,NEQN)
      V2V3 = DOTPRD(V2,V3,WT,NEQN)
      CALL STIFFB(V3V3,V2V3,V2V2,ROLD,RHO,ROOT1,ROOT2,ROOTRE)
      IF (ROOTRE) GO TO 100
      DET2 = V1V1*V2V2 - V1V2**2
      ALPHA2 = (-V1V1*V2V3+V1V2*V1V3)/DET2
      BETA2 = (V1V2*V2V3-V2V2*V1V3)/DET2
!
!  First test the residual of the quadratic fit to see if we might
!  have determined a pair of eigenvalues.
!
      RES2 = ABS(V3V3+V2V2*ALPHA2**2+V1V1*BETA2**2+TWO*V2V3*ALPHA2+&
     &       TWO*V1V3*BETA2+TWO*V1V2*ALPHA2*BETA2)
      IF (RES2.LE.V3V3*P001**2) THEN
!
!  Calculate the two approximate pairs of eigenvalues.
!
         CALL STIFFC(ALPHA1,BETA1,R1,R2)
         CALL STIFFC(ALPHA2,BETA2,ROOT1,ROOT2)
!
!  The test for convergence is done on the larger root of the second
!  approximation.  It is complicated by the fact that one pair of roots 
!  might be real and the other complex.  First calculate the spectral 
!  radius RHO of HAVG*J as the magnitude of ROOT1.  Then see if one of 
!  the roots R1,R2 is within one per cent of ROOT1.  A subdominant root 
!  may be very poorly approximated if its magnitude is much smaller than 
!  RHO -- this does not matter in our use of these eigenvalues.
!
         RHO = SQRT(ROOT1(1)**2+ROOT1(2)**2)
         D1 = (ROOT1(1)-R1(1))**2 + (ROOT1(2)-R1(2))**2
         D2 = (ROOT1(1)-R2(1))**2 + (ROOT1(2)-R2(2))**2
         DIST = SQRT(MIN(D1,D2))
         IF (DIST.LE.P001*RHO) GO TO 100
      END IF
!
!  Do not have convergence yet.  Because the iterations are cheap, and
!  because the convergence criterion is stringent, we are willing to try
!  a few iterations.
!
      IF (NTRY.LT.MAXTRY) THEN
         NTRY = NTRY + 1_id
         V3NRM = SQRT(V3V3)
         DO 80 L = 1_id, NEQN
            V0(L) = V3(L)/V3NRM
   80    CONTINUE
         V0V0 = ONE
         GO TO 60
      ELSE
         UNSURE = .TRUE.
         RETURN
      END IF
!
!                        **************
!
!  We now have the dominant eigenvalues.  Decide if the average step
!  size is being restricted on grounds of stability.  Check the real
!  parts of the eigenvalues.  First see if the dominant eigenvalue is
!  in the left half plane -- there won't be a stability restriction
!  unless it is. If there is another eigenvalue of comparable magnitude
!  with a positive real part, the problem is not stiff. If the dominant
!  eigenvalue is too close to the imaginary axis, we cannot diagnose
!  stiffness.
!
  100 CONTINUE
      IF (ROOT1(1).GT.ZERO) THEN
         STIF = .FALSE.
         UNSURE = .FALSE.
         RETURN
      END IF
      RHO2 = SQRT(ROOT2(1)**2+ROOT2(2)**2)
      IF (RHO2.GE.P9*RHO .AND. ROOT2(1).GT.ZERO) THEN
         STIF = .FALSE.
         UNSURE = .FALSE.
         RETURN
      END IF
      IF (ABS(ROOT1(2)).GT.ABS(ROOT1(1))*TANANG) THEN
         UNSURE = .TRUE.
         RETURN
      END IF
!
!  If the average step size corresponds to being well within the
!  stability region, the step size is not being restricted because
!  of stability.
!
      STIF = RHO .GE. P9*STBRAD
      UNSURE = .FALSE.
      RETURN
      END SUBROUTINE



      SUBROUTINE STIFFB(V1V1,V0V1,V0V0,ROLD,RHO,ROOT1,ROOT2,ROOTRE)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Input:        V1V1, V0V1, V0V0
!  Input/output: ROLD
!  Output:       RHO, ROOT1(*),ROOT2(*),ROOTRE
!
!  Decide if the iteration has degenerated because of a strongly
!  dominant real eigenvalue.  Have just computed the latest iterate.
!  V1V1 is its dot product with itself, V0V1 is the dot product
!  of the previous iterate with the current one, and V0V0 is the
!  dot product of the previous iterate with itself.  ROLD is a
!  previous Rayleigh quotient approximating a dominant real
!  eigenvalue.  It must be computed directly the first time the
!  subroutine is called.  It is updated each call to STIFFB, hence
!  is available for subsequent calls.
!
!  If there is a strongly dominant real eigenvalue, ROOTRE is set
!  .TRUE., ROOT1(*) returns the eigenvalue, RHO returns the magnitude
!  of the eigenvalue, and ROOT2(*) is set to zero.
USE kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  RHO, ROLD, V0V0, V0V1, V1V1
      LOGICAL(lgt)           ROOTRE
!     .. Array Arguments ..
      REAL(rd)  ROOT1(2), ROOT2(2)
!     .. Parameters ..
      REAL(rd)  ZERO, P001
      PARAMETER         (ZERO=0.0_rd,P001=0.001_rd)
!     .. Local Scalars ..
      REAL(rd)  DET, R, RES
!     .. Intrinsic Functions ..
      INTRINSIC         ABS
!     .. Executable Statements ..
!
      R = V0V1/V0V0
      RHO = ABS(R)
      DET = V0V0*V1V1 - V0V1**2
      RES = ABS(DET/V0V0)
      ROOTRE = DET .EQ. ZERO .OR. (RES.LE.V1V1*P001**2 .AND.&
     &         ABS(R-ROLD).LE.P001*RHO)
      IF (ROOTRE) THEN
         ROOT1(1) = R
         ROOT1(2) = ZERO
         ROOT2(1) = ZERO
         ROOT2(2) = ZERO
      END IF
      ROLD = R
!
      RETURN
      END SUBROUTINE



      SUBROUTINE STIFFC(ALPHA,BETA,R1,R2)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Input:  ALPHA, BETA
!  Output: R1(*), R2(*)
!
!  This subroutine computes the two complex roots R1 and R2 of
!  the quadratic equation X**2 + ALPHA*X + BETA = 0.  The magnitude
!  of R1 is greater than or equal to the magnitude of R2. R1 and R2 are
!  returned as vectors of two components with the first being the real
!  part of the complex number and the second being the imaginary part.
use kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  ALPHA, BETA
!     .. Array Arguments ..
      REAL(rd)  R1(2), R2(2)
!     .. Parameters ..
      REAL(rd)  ZERO, TWO
      PARAMETER         (ZERO=0.0_rd,TWO=2.0_rd)
!     .. Local Scalars ..
      REAL(rd)  DISC, SQDISC, TEMP
!     .. Intrinsic Functions ..
      INTRINSIC         ABS, SQRT
!     .. Executable Statements ..
      TEMP = ALPHA/TWO
      DISC = TEMP**2 - BETA
      IF (DISC.EQ.ZERO) THEN
!
!  Double root.
!
         R1(1) = -TEMP
         R1(2) = ZERO
         R2(1) = R1(1)
         R2(2) = R1(2)
         RETURN
      END IF
!
      SQDISC = SQRT(ABS(DISC))
      IF (DISC.LT.ZERO) THEN
!
!  Complex conjugate roots.
!
         R1(1) = -TEMP
         R1(2) = SQDISC
         R2(1) = R1(1)
         R2(2) = -R1(2)
      ELSE
!
!  Real pair of roots.  Calculate the bigger one in R1(1).
!
         IF (TEMP.GT.ZERO) THEN
            R1(1) = -TEMP - SQDISC
         ELSE
            R1(1) = -TEMP + SQDISC
         END IF
         R1(2) = ZERO
         R2(1) = BETA/R1(1)
         R2(2) = ZERO
      END IF
!
      RETURN
      END SUBROUTINE


      SUBROUTINE STIFFD(V,HAVG,X,Y,F,FXY,WT,SCALE,VDOTV,Z,ZDOTZ,VTEMP)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  External:     F
!  Input:        V(*), HAVG, X, Y(*), FXY(*), WT(*), SCALE, VDOTV,
!  Output:       Z(*), ZDOTZ
!  Workspace:    VTEMP(*)
!
!  For an input vector V(*) of length NEQ, this subroutine computes a vector
!  Z(*) that approximates the product HAVG*J*V where HAVG is an input scalar 
!  and J is the Jacobian matrix of a function F evaluated at the input 
!  arguments (X,Y(*)).  This function is defined by a subroutine of the form
!  F(T,U,F) that when given T and U(*), returns the value of the function in 
!  F(*).  The input vector FXY(*) is defined by F(X,Y,FXY).  Scaling is a 
!  delicate matter.  A weighted Euclidean norm is used with the (positive) 
!  weights provided in WT(*).  The input scalar SCALE is the square root of 
!  the unit roundoff times the norm of Y(*).  The square of the norm of the
!  input vector V(*) is input as VDOTV.  The routine outputs the square of
!  the norm of the output vector Z(*) as ZDOTZ.  The subroutine calls the
!  REAL(rd) FUNCTION DOTPRD(U,V,WT,NEQ) to compute the dot (inner)
!  product.  The vector VTEMP(*) is used for working storage.
use kindtype
implicit none
!     .. Scalar Arguments ..
      REAL(rd)  HAVG, SCALE, VDOTV, X, ZDOTZ
!     .. Array Arguments ..
      REAL(rd)  FXY(*), V(*), VTEMP(*), WT(*), Y(*), Z(*)
!     .. Subroutine Arguments ..
      EXTERNAL          F
!     .. Common Block for Problem Definition ..
      REAL(rd)  TSTRT, TND, DIR, HSTRT, TOLR
      INTEGER(id)           NEQN
      COMMON /RKCOM1/   TSTRT, TND, DIR, HSTRT, TOLR, NEQN
      SAVE   /RKCOM1/
!     .. Common Block to hold Problem Status ..
      REAL(rd)  T, H, TOLD, HOLD
      INTEGER(id)           NFCN, SVNFCN, OKSTP, FLSTP
      LOGICAL(lgt)           FIRST, LAST
      COMMON /RKCOM2/   T, H, TOLD, HOLD, NFCN, SVNFCN, OKSTP, FLSTP,&
     &                  FIRST, LAST
      SAVE   /RKCOM2/
!     .. Local Scalars ..
      REAL(rd)  TEMP1, TEMP2
      INTEGER(id)           L
!     .. External Functions ..
  !    REAL(rd)  DOTPRD
 !               DOTPRD
!     .. Intrinsic Functions ..
      INTRINSIC         SQRT
!     .. Executable Statements ..
!
!  Scale V(*) so that it can be used as an increment to Y(*)
!  for an accurate difference approximation to the Jacobian.
!
      TEMP1 = SCALE/SQRT(VDOTV)
      DO 20 L = 1_id, NEQN
         VTEMP(L) = Y(L) + TEMP1*V(L)
   20 CONTINUE
!
      CALL F(X,VTEMP,Z)
      NFCN = NFCN + 1_id
!
!  Form the difference approximation.  At the same time undo
!  the scaling of V(*) and introduce the factor of HAVG.
!
      TEMP2 = HAVG/TEMP1
      DO 40 L = 1_id, NEQN
         Z(L) = TEMP2*(Z(L)-FXY(L))
   40 CONTINUE
!
      ZDOTZ = DOTPRD(Z,Z,WT,NEQN)
!
      RETURN
      END SUBROUTINE


      REAL(rd) FUNCTION DOTPRD(U,V,WT,NEQ)
!************************************************
!**** NOT A DESIGNATED USER-CALLABLE ROUTINE ****
!************************************************
!
!  Purpose:   To compute a weighted Euclidean dot (inner) product of
!             two vectors.
!
!  Input:     U(*), V(*), WT(*), NEQ
!  Output:    the result DOTPRD is returned via the subprogram name
!
!  Comments:
!  =========
!  The vectors U(*), V(*), and WT(*) are of length NEQ. The components
!  of WT(*) are weights that must be non-zero.
use kindtype
implicit none
!     .. Scalar Arguments ..
      INTEGER(id)           NEQ
!     .. Array Arguments ..
      REAL(rd)  U(*), V(*), WT(*)
!     .. Parameters ..
      REAL(rd)  ZERO
      PARAMETER         (ZERO=0.0_rd)
!     .. Local Scalars ..
      REAL(rd)  SUM
      INTEGER(id)           L
!     .. Executable Statements ..
!
      SUM = ZERO
      DO 20 L = 1, NEQ
         SUM = SUM + (U(L)/WT(L))*(V(L)/WT(L))
   20 CONTINUE
!
      DOTPRD = SUM
!
      RETURN
      END FUNCTION


      SUBROUTINE SOFTFL(ASK,ON)
!
!  Purpose:      To prevent a program STOP after a "catastrophic"
!                failure when using a routine from RKSUITE.
!
!  Input:        ASK
!  Input/output: ON
!
!  Comments:
!  =========
!  When a "catastrophic" failure is detected, the default action of
!  RKSUITE is to write an explanation to the standard output channel,
!  OUTCH, and STOP.  This subroutine can be used to prevent the STOP and
!  so allow the main program to continue.  To do this, you call SOFTFL with
!  ASK = .FALSE. and ON = .TRUE.  You must then call the subroutine CHKFL
!  after every call to a user-callable routine in RKSUITE to check whether
!  a catastrophic error occurred and take appropriate action if it did.  Of
!  course, you may call SETUP at any time to start a new problem, but calling
!  any other user-callable routine in RKSUITE after a catastrophic error will
!  lead to a STOP (even when "soft failure" has been set "on").
!
!  When ON is set by a call to SOFTFL with ASK = .FALSE., the value of ON
!  is SAVEd.  The subroutine RKMSG in RKSUITE calls SOFTFL with ASK = .TRUE.
!  to find out the SAVEd value of ON.
use kindtype
implicit none
!     .. Scalar Arguments ..
      LOGICAL(lgt)           ASK, ON
!     .. Local Scalars ..
      LOGICAL(lgt)           SOFT
!     .. Save statement ..
      SAVE              SOFT
!     .. Data statements ..
      DATA              SOFT/.FALSE./
!     .. Executable Statements ..
!
      IF (ASK) THEN
         ON = SOFT
      ELSE
         SOFT = ON
      END IF
!
      RETURN
      END SUBROUTINE


      SUBROUTINE CHKFL(ASK,ERROR)
!
!  Purpose:      Enquiry routine used in conjunction with SOFTFL.
!                Reports whether a "catastrophic" error was detected.
!
!  Input:        ASK
!  Input/output: ERROR
!
!  Comments:
!  =========
!  When a "catastrophic" failure is detected, the default action of
!  RKSUITE is to write an explanation to the standard output channel,
!  OUTCH, and STOP.  SOFTFL can be used to prevent the STOP and so
!  allow the main program to continue.  It is then necessary to call
!  CHKFL with ASK = .TRUE. after every call to a user-callable routine 
!  in RKSUITE to check whether a catastrophic error occurred and take 
!  appropriate action if it did.  If there was a catastrophic error, 
!  ERROR is returned .TRUE.  Of course, you may call SETUP at any time 
!  to start a new problem, but calling any other user-callable routine 
!  in RKSUITE after a catastrophic error will lead to a STOP (even when
!  "soft failure" has been set "on").
!
!  When a catastrophic failure (IER = 911) is detected in one of
!  the routines in RKSUITE, it calls CHKFL with ASK = .FALSE. and
!  ERROR = .TRUE.  This value of ERROR is SAVEd.
use kindtype
implicit none
!     .. Scalar Arguments ..
      LOGICAL(lgt)           ASK, ERROR
!     .. Local Scalars ..
      LOGICAL(lgt)           SAVERR
!     .. Save statement ..
      SAVE              SAVERR
!     .. Data statements ..
      DATA              SAVERR/.FALSE./
!     .. Executable Statements ..
!
      IF (ASK) THEN
         ERROR = SAVERR
      ELSE
         SAVERR = ERROR
      END IF
!
      RETURN
      END subroutine


end module
